#!/home/tops/bin/python2.7
# -*- coding: utf-8 -*-
import os;
import sys;
import socket;
import json;
import commands;
import time;
import shutil;
import subprocess
from const import CConst;
from exception import CStrException,CFatalException;
from log import fatal, error, warning, info, debug;

class CEnv:
    @staticmethod
    def GetLocalHost():
        return socket.gethostname();

    @staticmethod
    def GetClusterName():

        path = "/bin/me"
        if os.path.exists("/bin/tj_show"):
            path = "/bin/tj_show"

        nRet, strClusterName = CProcess.RunCmd4NoException(path + ' 2>/dev/null | /bin/grep Local_cluster |cut -d \: -f 2 | tr -d " "')
        if nRet != 0 or not strClusterName:
            return "";
        else:
            return strClusterName;

    # 获取卷列表
    @staticmethod
    def GetVolumeList():
        result = []
        nRet, strOut = CProcess.RunCmd4NoException("/apsara/deploy/puadmin vol -ls")
        if nRet != 0:
            return result, strOut

        for each in strOut.split('\n'):
            if each.find("Name") != -1:
                result.append(each.split(":")[1].strip())

        return result, strOut


    # 如果传入的 strContent 大于 threshold，则写入文件 filename
    @staticmethod
    def WriteToFileIfNeeded(strContent, filename, threshold):
        if len(strContent) > threshold:
            open(filename, 'w').write(strContent);
            return "please check file ./data/result_data/" + filename.split("/")[-1];
        else:
            return strContent

    # mkdir for result, 并保存上一次运行的结果
    @staticmethod
    def SavedOldData(strDataPath, lsSavedPath, nSaveCount):
        try:
            # 保留一定数量的备份，防止磁盘空间过度使用
            backupPath = os.path.join(strDataPath, "backup/");
            if os.path.exists(backupPath):
                lsDir = []
                for x in os.listdir(backupPath):
                    currPath = os.path.join(backupPath, x);
                    if os.path.isdir(currPath):
                        lsDir.append(currPath);
                lsDir.sort();
                while len(lsDir) - nSaveCount >= 0:
                    info("backup has too many dirs, remove " + lsDir[0]);
                    shutil.rmtree(lsDir[0])
                    lsDir.pop(0);

            # 在backup 下创建新文件夹，备份数据
            ISOTIMEFORMAT='%Y-%m-%d_%X'
            bkdir = os.path.join(backupPath, time.strftime(ISOTIMEFORMAT, time.localtime()));
            os.makedirs(bkdir);

            for path in lsSavedPath:
                if os.path.exists(path) == True:
                    shutil.move(path, bkdir + "/");
                os.makedirs(path)

        except Exception as e:
            info("!!%s" %e);
            raise e

class CGlobal:
    @staticmethod
    def GetFunc(obj, strFunc):
        for strTmpFunc in dir(obj):
            if strFunc.lower() == strTmpFunc.lower():
                return getattr(obj, strTmpFunc)

        error("Can't found Func(%s)" %strFunc);
        raise CStrException("Can't found Func(%s)" %strFunc);

    @staticmethod
    def GetPyMod(strPyFile):
        #insert to directory into sys path
        strPyDir = os.path.dirname(strPyFile);
        if not strPyDir in sys.path:
            sys.path.insert(0, strPyDir);

        strImportPy = os.path.basename(strPyFile).split('.')[0]
        return __import__(strImportPy)

    @staticmethod
    def GetMonitorFunc(strPyFile):
        '''
        must eixst def GetDataList(mapArg)  --> list function
        '''
        oMod = CGlobal.GetPyMod(strPyFile);
        return getattr(oMod, 'GetDataList');

    @staticmethod
    def GetDataMapFunc(strPyFile):
        '''
        must eixst def GetDataList(mapArg)  --> list function
        '''
        oMod = CGlobal.GetPyMod(strPyFile);
        return getattr(oMod, 'GetDataMap');

    @staticmethod
    def GetCheckerFunc(strPyFile):
        '''
        must exist def CheckDataList(strParam, lsData)  --> check failed throw exception
        '''
        oMod = CGlobal.GetPyMod(strPyFile);
        return getattr(oMod, 'CheckDataList');

    @staticmethod
    def ParseJson(strFilePath, bIngnoreExist = False):
        info("load config:%s" %strFilePath, True);
        if bIngnoreExist and not os.path.isfile(strFilePath):
            return {};

        strJsonContent = open(strFilePath, 'r').read().strip();
        if not strJsonContent:
            return {};

        return json.loads(strJsonContent);

    @staticmethod
    def IsNeedAlarm(strLevel):
        return strLevel.lower() in CConst.M_CHECK_KEY_LIST;

class CProcess:
    NEW_PATH = ':'.join(['/usr/ali/bin', '/usr/ali/sbin',
                         '/sbin', '/usr/sbin', '/bin',
                         '/usr/bin', '/usr/local/bin',
                         '/usr/local/sbin',
                         os.environ['PATH']])
    @staticmethod
    def RunCmd4NoException(strCmd):
        debug(strCmd);
        nRet, strOutput = commands.getstatusoutput(strCmd)

        #python bug
        nRet = nRet >> 8;
        if 0 != nRet:
            error(strCmd);

        return nRet, strOutput;

    @staticmethod
    def RunCmd(strCmd):
        nRet, strOutput = CProcess.RunCmd4NoException(strCmd);
        if (0 != nRet):
            raise CStrException(strOutput)

        return strOutput;

    @staticmethod
    def ExecCommand(cmd, timeout=600, exception=True):
        nTryCount = timeout * 10
        proc = subprocess.Popen(cmd,
                                stdout = subprocess.PIPE,
                                stderr = subprocess.PIPE,
                                env = {"PATH": CProcess.NEW_PATH},
                                shell = True)
        while True:
            ret = subprocess.Popen.poll(proc)
            if ret is None:
                if nTryCount > 0 or timeout <= 0:
                    nTryCount -=1;
                    time.sleep(0.1)
                else:
                    err_msg = 'Command execution timeout. Command: %s, Timeout: %d'\
                              %(cmd, timeout)
                    error(err_msg)
                    if exception:
                        raise Exception(err_msg)
                    else:
                        return -1, err_msg, err_msg
            elif ret == 0:
                (stdoutdata, stderrdata) = proc.communicate()
                info("Command execution success. Command: %s, Return %d, Stdout: %s, Stderr: %s"
                     %(cmd, ret, stdoutdata, stderrdata))
                return 0, stdoutdata, stderrdata
            else:
                (stdoutdata, stderrdata) = proc.communicate()
                err_msg = 'Command execution failed. Command: %s, Return %d, Stdout: %s, Stderr: %s'\
                          %(cmd, ret, stdoutdata, stderrdata)
                error(err_msg)
                if exception:
                    raise Exception(err_msg)
                return ret, stdoutdata, stderrdata

class CStringHandle:
    @staticmethod
    def SplitMultiLine(strparam, splitter):
        listdata = strparam.split('\n');
        result = [];
        for each_line in listdata:
            if each_line != '':
                result.append(each_line.split(splitter)[1].strip());
        return result;

