#!/home/tops/bin/python2.7
# -*- coding: utf-8 -*-
import os
import sys
import json
import yaml
from common import CProcess
from log import info,warning,error


'''
Disks grouped by usage:
- system disks
- data disks

Data disks grouped by kernel/user space dirver:
- kernel space driver
- user space driver

Disks grouped by type:
- hdd
- sata_ssd
- nvme
- qlc
- optane
- aliflash # TODO
'''
class CDisk:
    DISK_TYPE_ALL = 'all'
    DISK_TYPE_HDD = 'hdd'
    DISK_TYPE_SSD = 'ssd'
    DISK_TYPE_SATA_SSD = 'sata_ssd'
    DISK_TYPE_NVME = 'nvme'
    DISK_TYPE_OPTANE = 'optane'
    DISK_TYPE_QLC = 'qlc'

    def __init__(self):
        self.m_strConfigDir = '/apsara/pangu_monitor/conf.global'
        self.m_strClusterTagConf = os.path.join(self.m_strConfigDir, 'tag.json')
        self.m_strPanguTagConf = os.path.join(self.m_strConfigDir,
                'services/pangu/tag.json')
        self.m_setAllDisks = set()
        self.m_setHddDisks = set()
        self.m_setSataSsdDisks = set()
        self.m_setOptaneDisks = set()
        self.m_setQlcDisks = set()
        self.m_setNvmeDisks = set()
        self.m_setSysDisks = set()
        self.m_setDataDisks = set()
        self.m_setKernelSpaceDisks = set()
        self.m_setUserSpaceDisks = set()

    def LoadConfJson(self, strConfPath):
        mapConf = {}
        try:
            mapConf = yaml.safe_load(open(strConfPath, 'r').read())
            info('Load config: %s' %(strConfPath))
        except Exception as e:
            error(str(e))
        return mapConf

    def GetAllDisks(self):
        if self.m_setAllDisks:
            return self.m_setAllDisks

        # Get block device form /sys/block
        lsAllDevices = []
        strCmd = "ls /sys/block | grep -e ^sd -e ^df -e ^nvme -e ^os"
        try:
            nRet, strOut, strErr = CProcess.ExecCommand(strCmd)
            if nRet == 0 and strOut:
                lsAllDevices = strOut.split()
        except Exception as e:
            error(str(e))

        # Get block device from /dev/
        lsAllDevs = []
        strCmd = "ls /dev | grep -e ^sd -e ^df -e ^nvme -e ^os"
        try:
            nRet, strOut, strErr = CProcess.ExecCommand(strCmd)
            if nRet == 0 and strOut:
                lsAllDevs = strOut.split()
        except Exception as e:
            error(str(e))

        self.m_setAllDisks = set(lsAllDevices) & set(lsAllDevs)
        return self.m_setAllDisks

    # Get mount device. If it's partition, strip the partion number
    def IsSsd(self, strDev):
        bRet = False
        try:
            strRotational = open('/sys/block/%s/queue/rotational'%(strDev), 'r').read().strip()
            if strRotational == '0':
                bRet = True
            elif strRotational == '1':
                bRet = False
            else:
                raise Exception('Invalid rotational value of %s'%(strDev))
        except Exception, e:
            error(str(e))
        return bRet

    def IsHdd(self, strDev):
        return not self.IsSsd(strDev)

    def IsSataSsd(self, strDev):
        return self.IsSsd(strDev) and strDev.startswith('sd')

    def IsNvmeSsd(self, strDev):
        return self.IsSsd(strDev) and strDev.startswith('nvme')

    def GetDevByFunc(self, strFunc):
        strCmd = """ls /sys/bus/pci/devices/%s/nvme"""%(strFunc)
        nRet, strOut, strErr = CProcess.ExecCommand(strCmd, exception=False)
        return strOut.strip('\n') + 'n1'

    def GetHddDisks(self):
        if self.m_setHddDisks:
            return self.m_setHddDisks

        setHddDisks = set()
        for disk in self.GetAllDisks():
            if self.IsHdd(disk):
                setHddDisks.add(disk)
        self.m_setHddDisks = setHddDisks
        return self.m_setHddDisks

    def GetSataSsdDisks(self):
        if self.m_setSataSsdDisks:
            return self.m_setSataSsdDisks

        setSataSsdDisks = set()
        for disk in self.GetAllDisks():
            if self.IsSataSsd(disk):
                setSataSsdDisks.add(disk)
        self.m_setSataSsdDisks= setSataSsdDisks
        return self.m_setSataSsdDisks

    def GetOptaneDisks(self):
        if self.m_setOptaneDisks:
            return self.m_setOptaneDisks

        lsDev = []
        strCmd = """lspci -mm -n -D | grep 0108 | grep 2701 | tr -d '"' | awk -F " " '{print $1}'"""
        try:
            nRet, strOut, strErr = CProcess.ExecCommand(strCmd)
            if nRet == 0 and strOut:
                lsFunc = strOut.split()
                for strFunc in lsFunc:
                    lsDev += [self.GetDevByFunc(strFunc)]
        except Exception as e:
            error(str(e))

        self.m_setOptaneDisks = set(lsDev)
        return self.m_setOptaneDisks

    def IsOptane(self, strDev):
        return strDev in self.GetOptaneDisks()

    def GetQlcDisks(self):
        if self.m_setQlcDisks:
            return self.m_setQlcDisks

        lsDev = []
        strCmd = """lspci -mm -n -D | grep 0108 | grep 4308 | tr -d '"' | awk -F " " '{print $1}'"""
        try:
            nRet, strOut, strErr = CProcess.ExecCommand(strCmd)
            if nRet == 0 and strOut:
                lsFunc = strOut.split()
                for strFunc in lsFunc:
                    lsDev += [self.GetDevByFunc(strFunc)]
        except Exception as e:
            error(str(e))

        self.m_setQlcDisks = set(lsDev)
        return self.m_setQlcDisks

    def IsQlc(self, strDev):
        return strDev in self.GetQlcDisks()

    # Nvme devices are not include qlc and optane
    def GetNvmeDisks(self):
        if self.m_setNvmeDisks:
            return self.m_setNvmeDisks

        lsDev = []
        strCmd = """lspci -mm -n -D | grep 0108 | grep -v 4308 | grep -v 2701 | tr -d '"' | awk -F " " '{print $1}'"""
        try:
            nRet, strOut, strErr = CProcess.ExecCommand(strCmd)
            if nRet == 0 and strOut:
                lsFunc = strOut.split()
                for strFunc in lsFunc:
                    lsDev += [self.GetDevByFunc(strFunc)]
        except Exception as e:
            error(str(e))

        self.m_setNvmeDisks = set(lsDev)
        return self.m_setNvmeDisks

    def IsNvme(self, strDev):
        return strDev in self.GetNvmeDisks()

    # Get device type
    def GetDevType(self, strDev):
        if self.IsHdd(strDev):
            return 'hdd'
        elif self.IsSataSsd(strDev):
            return 'sata_ssd'
        elif self.IsQlc(strDev):
            return 'qlc'
        elif self.IsOptane(strDev):
            return 'optane'
        elif self.IsNvme(strDev):
            return 'nvme'
        else:
            error('Unkown disk type: /dev/%s'%(strDev))

    def ExpandDiskType(self, setDiskType):
        mapDiskType = {}
        mapDiskType[CDisk.DISK_TYPE_ALL] = set([CDisk.DISK_TYPE_HDD, CDisk.DISK_TYPE_SSD])
        mapDiskType[CDisk.DISK_TYPE_SSD] = set([CDisk.DISK_TYPE_SATA_SSD,
                                                 CDisk.DISK_TYPE_NVME,
                                                 CDisk.DISK_TYPE_OPTANE,
                                                 CDisk.DISK_TYPE_QLC])

        if CDisk.DISK_TYPE_ALL in setDiskType:
            setDiskType = (setDiskType - set([CDisk.DISK_TYPE_ALL])) | mapDiskType[CDisk.DISK_TYPE_ALL]
        if CDisk.DISK_TYPE_SSD in setDiskType:
            setDiskType = (setDiskType - set([CDisk.DISK_TYPE_SSD])) | mapDiskType[CDisk.DISK_TYPE_SSD]

        return setDiskType

    def GetMountDev(self, strMountPoint):
        strDevice = ''
        strCmd = 'df %s| grep ^/dev'%(strMountPoint)
        nRet, strOut, strErr = CProcess.ExecCommand(strCmd, exception=False)
        if nRet == 0 and strOut:
            strDevice = strOut.split()[0].split('/')[-1]

        # Tell if the device is a partition
        #  $readlink -f /sys/class/block/sda1
        #  /sys/devices/pci0000:00/0000:00:11.5/ata1/host0/target0:0:0/0:0:0:0/block/sda/sda1
        #  $readlink -f /sys/class/block/sda
        #  /sys/devices/pci0000:00/0000:00:11.5/ata1/host0/target0:0:0/0:0:0:0/block/sda
        # nvme0p1
        if strDevice:
            if os.path.isdir('/sys/class/block') and os.path.isfile('/sys/class/block/%s'%(strDevice)):
                strCmd = 'basename $(readlink -f /sys/class/block/%s/..)'%(strDevice)
                try:
                    nRet, strOut, strErr = CProcess.ExecCommand(strCmd)
                    # Is a partition
                    if nRet ==0 and strOut != 'block':
                        strDevice = strOut
                except Exception as e:
                    error(str(e))
            else:
                if strDevice.startswith('nvme'):
                    strDevice = strDevice.split('p')[0]
                else:
                    strDevice = strDevice.strip('0123456789')

        return strDevice

    def GetSysDisks(self):
        if self.m_setSysDisks:
            return self.m_setSysDisks
        setSysDisks = set()
        for strPath in ['/', '/apsara', '/apsarapangu']:
            strDev = self.GetMountDev(strPath).strip()
            if strDev:
                setSysDisks.add(strDev)
        self.m_setSysDisks = setSysDisks
        return self.m_setSysDisks

    def GetDataDisks(self):
        if self.m_setDataDisks:
            return self.m_setDataDisks

        self.m_setDataDisks = self.GetAllDisks() - self.GetSysDisks()
        return self.m_setDataDisks

    # User space data disks and system disks
    # TODO Add user space system disk if it isn't  empty
    def GetUserSpaceDisks(self):
        if self.m_setUserSpaceDisks:
            return self.m_setUserSpaceDisks

        mapClusterTag = {}
        if os.path.isfile(self.m_strClusterTagConf):
            mapClusterTag = self.LoadConfJson(self.m_strClusterTagConf)

        mapPanguTag = {}
        if os.path.isfile(self.m_strPanguTagConf):
            mapPanguTag = self.LoadConfJson(self.m_strPanguTagConf)

        if not mapClusterTag and not mapPanguTag:
            warning('No cluster tag nor pangu tag found, no user space disk')
            return self.m_setUserSpaceDisks

        setDiskType = set()
        if mapPanguTag['Tags'].has_key('USSFS'):
            lsDiskType = mapPanguTag['Tags']['USSFS'].strip().lower().split(',')
            lsDiskType = map(str.strip, lsDiskType)
            setDiskType = self.ExpandDiskType(set(lsDiskType))
        elif mapClusterTag['Tags'].has_key('SPDK_FORMAT'):
            setDiskType = set(['nvme', 'qlc', 'optane'])

        setUserSpaceDisks = set()
        for disk in self.GetDataDisks():
            if self.GetDevType(disk) in setDiskType:
                setUserSpaceDisks.add(disk)
        self.m_setUserSpaceDisks = setUserSpaceDisks
        return self.m_setUserSpaceDisks

    # Include sys disks and data disks
    def GetKernelSpaceDisks(self):
        if self.m_setKernelSpaceDisks:
            return self.m_setKernelSpaceDisks

        self.m_setKernelSpaceDisks = self.GetAllDisks() - self.GetUserSpaceDisks()
        return self.m_setKernelSpaceDisks

    def GetAllNvmeSsd(self):
        setAllNvmeSsd = self.GetNvmeDisks() | self.GetQlcDisks() | self.GetOptaneDisks() -\
                self.GetUserSpaceDisks()
        return setAllNvmeSsd

    '''
    Get SATA SSD life time by smart attribute.
    1. try to get Media_Wearout_Indicator if it exists.
    2. try to get by attribute id and attribute name of "Unknown_SSD_Attribute".

    ATTR_ID     ATTR_NAME                   MANUFACTURE
    ----------------------------------------------------------------------------
    173         Media_Wearout_Indicator     Micron, Intel, Samsung
    177         Unknown(_SSD)_Attribute     Samsung
    202         Unknown(_SSD)_Attribute     Micron
    231         Unknown(_SSD)_Attribute     Seagate
    233         Unknown(_SSD)_Attribute     Intel
    ----------------------------------------------------------------------------
    '''
    def GetSataSsdLifeTime(self, strDev):
        strLifeTime = ""

        lsAttrIds = ["173", "177", "202", "231", "233"]
        strCmd = 'sudo smartctl -a /dev/%s' %(strDev)
        try:
            nRet, strOut, strErr = CProcess.ExecCommand(strCmd)
            if  strOut:
                lsOut = strOut.split("\n")
                for each in lsOut:
                    if each.find("Media_Wearout_Indicator") != -1:
                        strLifeTime = each.split()[3]
                        return strLifeTime

                for each in lsOut:
                    for attrId in lsAttrIds:
                        if (each.startswith(attrId) and \
                            ("Unknown_SSD_Attribute" in each or "Unknown_Attribute" in each)):
                            strLifeTime = each.split()[3]
                            return strLifeTime

                for each in lsOut:
                    # 适配华为 ssd，寿命描述字段为 SS Media used endurance indicator: 1%
                    if each.find("SS Media used endurance indicator") != -1:
                        strLifeTime = str(100 - int(each.replace("%", "").split(":")[1]))
                        return strLifeTime

            else:
                error('Failed to get smart attribute for device: %s, stdout: %s, stderr: %s'
                      %(strDev, strOut, strErr))
        except Exception as e:
            error(str(e))
        return strLifeTime

    '''
    NVMe SSD life time is recorded in percentage_used attribute. So percentage_used
    can be more than 100%.
    Smart log Example:
        Smart Log for NVME device:nvme0n1 namespace-id:ffffffff
        critical_warning                    : 0
        temperature                         : 43 C
        available_spare                     : 100%
        available_spare_threshold           : 0%
        percentage_used                     : 1%
    '''
    def GetNvmeSsdLifeTime(self, strDev):
        strLifeTime = ''
        strCmd = 'sudo nvme smart-log /dev/%s | grep percentage_used' %(strDev)
        try:
            nRet, strOut, strErr = CProcess.ExecCommand(strCmd)
            if nRet == 0 and strOut:
                nPercentageUsed = int(strOut.split(':')[-1].strip().strip('%'))
                strLifeTime = str(100 - nPercentageUsed)
            else:
                error('Failed to get smart log for device: %s, stdout: %s, stderr: %s'
                      %(strDev, strOut, strErr))
        except Exception as e:
            error(str(e))
        return strLifeTime

    def GetSsdLifeTime(self, strDev):
        if self.IsSataSsd(strDev):
            return self.GetSataSsdLifeTime(strDev)
        elif self.IsNvmeSsd(strDev):
            return self.GetNvmeSsdLifeTime(strDev)
        return ''

    def GetNvmeSsdDiskInfo(self, strDev):
        mapDiskInfo = {}
        mapDiskInfo['sn'] = 'Unkown'
        mapDiskInfo['model'] = 'Unkown'
        mapDiskInfo['fr'] = 'Unkown'
        strCmd = 'sudo nvme id-ctrl /dev/%s' %(strDev)
        try:
            nRet, strOut, strErr = CProcess.ExecCommand(strCmd)
            if nRet == 0 and strOut:
                lsDiskInfo = strOut.split('\n')
                #lsDiskInfo = ['sn      : PHLF749500FH2P0HGN', 'mn      : INTEL SSDPE2KX020T7', 'fr      : QDV101B1']
                for strInfo in lsDiskInfo:
                    if strInfo.startswith('sn '):
                        mapDiskInfo['sn'] = strInfo.split(':')[-1].strip()
                    elif strInfo.startswith('mn '):
                        mn = strInfo.split(':')[-1].strip()
                        mapDiskInfo['manufacture'] = mn.split()[0]
                        mapDiskInfo['model'] = mn.split()[-1]
                    elif strInfo.startswith('fr '):
                        mapDiskInfo['fr'] = strInfo.split(':')[-1].strip()
                error('Failed to send id_ctrl cmd to device: %s, stdout: %s, stderr: %s'
                      %(strDev, strOut, strErr))
        except Exception as e:
            error(str(e))
        mapDiskInfo['life'] = self.GetSsdLifeTime(strDev)
        return mapDiskInfo

    def GetDiskInfo(self):
        lsDiskInfo = []
        for disk in self.GetAllDisks():
            mapOneDiskInfo = {}
            mapOneDiskInfo['dev'] = disk
            mapOneDiskInfo['usage'] = 'system' if disk in self.GetSysDisks() else 'data'
            mapOneDiskInfo['mode'] = 'user_space' if disk in self.GetUserSpaceDisks() else 'kernel_space'
            mapOneDiskInfo['type'] = self.GetDevType(disk)
            lsDiskInfo.append(mapOneDiskInfo)
        return lsDiskInfo

    def GetAllSsdLifeTime(self):
        lsSsdLifeTime = []
        for disk in (self.GetSataSsdDisks() | self.GetAllNvmeSsd()):
            mapOneSsdLifeTime = {}
            mapOneSsdLifeTime['dev'] = disk
            mapOneSsdLifeTime['life_time'] = self.GetSsdLifeTime(disk)
            lsSsdLifeTime.append(mapOneSsdLifeTime)
        return lsSsdLifeTime

if '__main__' == __name__:
    disk = CDisk()
    print json.dumps(disk.GetDiskInfo(), indent=4)
    print json.dumps(disk.GetAllSsdLifeTime(), indent=4)
    #print json.dumps(disk.GetNvmeSsdDiskInfo('nvme0n1'), indent=4)
