#!/home/tops/bin/python2.7
# -*- coding: utf-8 -*-
'''
Notice: current file need sudo privilage for some function, such as:
GetMemorySpeed,GetProcOpenFilesLimitList
'''
import os;
import sys;
import socket;
import fcntl;
import struct;
import commands;
from common import CProcess;
from exception import CStrException,CFatalException;
from log import fatal, error, warning, info, debug;

class CLinuxEnv:
    @staticmethod
    def GetProcPathList(strBinPath, bAbsPath = False):
        lsDir = [];
        strDir = "/proc/";
        for strFile in os.listdir(strDir):
            if os.path.isfile(strFile):
                continue;

            try:
                #print(strFile);
                int(strFile);#check is int dir
                strCurrBin = open(os.path.join(strDir, strFile, 'cmdline'), 'r').readline(1024);
                strCurrBin = strCurrBin.split('\0', 1)[0];
                #print(strCurrBin);

                if not bAbsPath:
                    strCurrBin = strCurrBin.split('/')[-1];

                if strCurrBin == strBinPath:
                    lsDir.append(os.path.join(strDir, strFile));

            except Exception, e:
                continue;

        debug("find process dir:%s" %lsDir, True);
        return lsDir;

    @staticmethod
    def GetProcFdCountList(strBinPath, bAbsPath = False):
        lsFds = [];
        for strDir in CLinuxEnv.GetProcPathList(strBinPath, bAbsPath):
            try:
                lsFds.append(len(os.listdir(os.path.join(strDir, 'fd'))));
            except OSError, e:
                debug(str(e), True);

        return lsFds;

    @staticmethod
    def GetProcOpenFilesLimitList(strBinPath, bAbsPath = False):
        lsOpenFilesLimit = [];
        for strDir in CLinuxEnv.GetProcPathList(strBinPath, bAbsPath):
            try:
                strOutput = CProcess.RunCmd("sudo cat %s/limits" %strDir);
                strVal = [x.strip() for x in strOutput.split('\n') if -1 != x.find('Max open files')][0];
                lsLimitSrc = strVal[len('Max open files') :].split();
                for strLimit in lsLimitSrc[0:2]:
                    if 'unlimited' == strLimit.lower():
                        lsOpenFilesLimit.append(sys.maxint);
                    else:
                        lsOpenFilesLimit.append(int(strLimit));
            except OSError, e:
                warning(str(e), True);

        return lsOpenFilesLimit;

    @staticmethod
    def GetMemBytesMapByLinuxFile(strFile, lsKeys):
        mapMem = {};
        for strLine in open(strFile, 'r').readlines():
            strKey, strVal = strLine.split(':');
            strKey = strKey.strip();
            if (strKey in lsKeys) or (not lsKeys):
                mapMem[strKey] = CLinuxEnv.ConvertToByte(strVal.strip());

        return mapMem;

    @staticmethod
    def GetProcMemList(strBinPath, bAbsPath = False):
        lsMem = [];
        for strDir in CLinuxEnv.GetProcPathList(strBinPath, bAbsPath):
            mapMem = CLinuxEnv.GetMemBytesMapByLinuxFile(os.path.join(strDir, 'status'), ['VmRSS']);
            lsMem.append(mapMem['VmRSS']);

        return lsMem;

    @staticmethod
    def ConvertToByte(strVal):
        nVal, strType = strVal.lower().split();
        if strType in ['', 'b', 'hz']:
            return int(nVal);
        elif strType in ['kb', 'khz']:
            return int(1024 * float(nVal));
        elif strType in ['mb', 'mhz']:
            return int(1024 * 1024 * float(nVal));
        elif strType in ['gb', 'ghz']:
            return int(1024 * 1024 * 1024 * float(nVal));
        elif strType in ['tb', 'thz']:
            return int(1024 * 1024 * 1024 * 1024 * float(nVal));
        else:
            raise CStrException("invalid type(%s)" %strType);

    @staticmethod
    def GetMemInfo():
        mapMem = CLinuxEnv.GetMemBytesMapByLinuxFile('/proc/meminfo', ['MemTotal', 'MemFree']);
        mapMem['MemUsed'] = mapMem['MemTotal'] - mapMem['MemFree'];

        debug(mapMem, True);
        return mapMem;

    @staticmethod
    def GetDiskInfo(strDir):
        stDisk = os.statvfs(strDir)
        return {'free' : stDisk.f_bavail * stDisk.f_frsize, 'total' : stDisk.f_blocks * stDisk.f_frsize};

    @staticmethod
    def GetLoadInfo():
        lsLoad = open("/proc/loadavg", 'r').read().split();
        return {'lavg_1' : lsLoad[0], 'lavg_5' : lsLoad[1], 'lavg_15' : lsLoad[2]};

    @staticmethod
    def GetNetBuff():
        lsLine = open("/proc/net/dev", 'r').readlines();
        mapNet = {};
        for strLine in lsLine[2:]:
            strLine = strLine.replace(':' ,  ' ');
            lsInfo = strLine.split();
            debug("%s: recv:%s  send:%s" %(lsInfo[0], lsInfo[1], lsInfo[9]), True);
            mapNet[lsInfo[0]] = (float(lsInfo[1]), float(lsInfo[9]));

        return mapNet;

    @staticmethod
    def GetIpByIFrame(strIFrameName = 'eth0'):
        sockTest = socket.socket(socket.AF_INET, socket.SOCK_DGRAM);
        inet = fcntl.ioctl(sockTest.fileno(), 0x8915, struct.pack('256s', strIFrameName[:15]));
        return socket.inet_ntoa(inet[20:24]);

    @staticmethod
    def GetMemorySpeed(bIngnoreUnknown):
        lsMemSpeed = [];
        strOutput = CProcess.RunCmd("sudo /usr/sbin/dmidecode --type=memory");
        lsMemInfo = [x.strip() for x in strOutput.split('\n\n') if (-1 != x.find("Memory Device"))];
        for strMemInfo in lsMemInfo:
            #print(strMemInfo);
            lsSpeed = [x.strip() for x in strMemInfo.split('\n') if 0 == x.find("\tSpeed:")];
            if 1 != len(lsSpeed):
                raise CStrException("invlid memory speed len!\n%s" %lsSpeed);

            strSpeed = lsSpeed[0].split(':', 1)[1].strip();
            if bIngnoreUnknown and 'unknown' == strSpeed.lower():
                error("skip unknown memory speed", True);
                continue;

            lsMemSpeed.append(CLinuxEnv.ConvertToByte(strSpeed));

        return lsMemSpeed;

    @staticmethod
    def GetMemoryOverCommitVal():
        return int(open("/proc/sys/vm/overcommit_memory", 'r').read().strip());

