#!/home/tops/bin/python2.7
# -*- coding: utf-8 -*-
import os
import sys
import commands
import json
import uuid;
import copy
import datetime

__strCurrentDir__ = os.path.abspath(os.path.dirname(__file__))
__strModuleDir__ = os.path.dirname(__strCurrentDir__)
if __strModuleDir__ not in sys.path:
    sys.path.insert(0, __strModuleDir__)

from common.common import CProcess
from common.common import CEnv
from common.log import info, error

# 分析 0 replica 的数据， 对应的chunk 是否存在，是否有恢复的可能
# 分析 lessmax 的abnchunk, 看是否是因为replica 数大于chunkserver 数导致的
class AbnchunkAnalyze:
    def __init__(self):
        #
        self.m_mapChunkExists = {};
        pass;

    # 获取非 Normal 的cs 的数量
    def GetNotNormalCsCount(self):
        strCmd = "/apsara/deploy/puadmin lscs|grep tcp |grep -v NORMAL"
        nRet, strOut = CProcess.RunCmd4NoException(strCmd);
        if len(strOut) > 0:
            return len(strOut.split("\n")), strOut;
        else:
            return 0, ""


    # 获取 abnchunk replica 的 fileid 和 chunkid 列表，执行聚合，按文件分类，以确定影响
    def FindReplicaChunkId(self, type):

        # key 为 fileid，value 为 chunkid list
        mapChunks = {}
        strCmd = "/apsara/deploy/puadmin fs -abnchunk -t " + type
        nRet, strOut = CProcess.RunCmd4NoException(strCmd);
        if nRet == 0:
            for each in strOut.split('\n'):
                if each.find("_") != -1:
                    fileid = each.split()[0].split("_")[0];
                    chunkid = each.split()[0].split("_")[1]
                    if mapChunks.has_key(fileid):
                        mapChunks[fileid].append(chunkid)
                    else:
                        mapChunks[fileid] = [chunkid]

        return mapChunks

    # 获取文件路径
    def GetFilePath(self, fileId):
        strCmd = "/apsara/deploy/puadmin  whois " + fileId
        nRet, strOut = CProcess.RunCmd4NoException(strCmd);
        if nRet == 0:
            for each in strOut.split('\n'):
                if each.find(fileId) != -1:
                    return each.split()[-1];

        return "";


    # puadmin gfi $fileid
    #FileType:NORMAL_FILE_TYPE_FLAG
    #CompressType:NO_COMPRESSION
    #ReferenceCount:1
    #ChunkNumber:4
    #AppName:BIGFILE_APPNAME
    #PartName:BIGFILE_PARTNAME
    #CreateTime:20190219094256
    #LastModifyTime:20190219094339
    #FileLength:246294602
    def GetFileInfo(self, fileId):
        fileInfoMap = {}
        strCmd = "/apsara/deploy/puadmin  gfi " + fileId
        nRet, strOut = CProcess.RunCmd4NoException(strCmd);
        if nRet == 0:
            for each in strOut.split('\n'):
                if each.find("FileType") != -1:
                    fileInfoMap["FileType"] = each.split(":")[1]
                if each.find("ChunkNumber") != -1:
                    fileInfoMap["ChunkNumber"] = each.split(":")[1]
                if each.find("CreateTime") != -1:
                    fileInfoMap["CreateTime"] = each.split(":")[1]
                if each.find("LastModifyTime") != -1:
                    fileInfoMap["LastModifyTime"] = each.split(":")[1]
                if each.find("FileLength") != -1:
                    fileInfoMap["FileLength"] = each.split(":")[1]

        return fileInfoMap

    # 调用 gfi 命令，获取一个文件的 chunk 的分布详情，数据样例
    # ChunkId:35 Version:584 ChunkFlag:0 ReplicaNum:3 CSInfo:
    # {
    #        [Status:normal]
    #        [Location:tcp://100.81.240.140:10260]
    #        [DiskId:5]
    #
    #        [Status:normal]
    #        [Location:tcp://100.81.240.139:10260]
    #        [DiskId:8]
    #
    #        [Status:normal]
    #        [Location:tcp://100.81.240.144:10260]
    #        [DiskId:5]
    #
    #  }
    # ChunkId:36 Version:597 ChunkFlag:0 ReplicaNum:4 CSInfo:
    # {
    #        [Status:DISCONNECTED]
    #        [Location:tcp://100.81.240.143:10260]
    #        [DiskId:8]
    #
    #        [Status:normal]
    #        [Location:tcp://100.81.240.139:10260]
    #        [DiskId:2]
    #
    #        [Status:CONNECTING]
    #        [Location:tcp://100.81.240.137:10260]
    #        [DiskId:8]
    #
    #        [Status:normal]
    #        [Location:tcp://100.81.240.140:10260]
    #        [DiskId:7]
    #
    # }
    def GetChunkDetail(self, fileId):
        # 返回一个map，key 为chunkIndex，value 为详情, 详情内是一个list
        mapRet = {}
        strCmd = "/apsara/deploy/puadmin  gfi " + fileId ;
        nRet, strOut = CProcess.RunCmd4NoException(strCmd);

        #记录每个chunk 的状态，因为要反复使用，往 mapRet 中赋值时要使用深拷贝
        tmpDetail = {}
        if nRet == 0:
            for each in strOut.split("\n"):
                if each.find("ChunkId") != -1:
                    chunkIndex = each.split()[0].split(":")[1]
                if each.find("Status") != -1:
                    tmpDetail["Status"] = each.replace("[Status:", " ").replace("]", " ").strip();
                if each.find("Location") != -1:
                    tmpDetail["Location"] = each.replace("[Location:", " ").replace("]", " ").strip();
                if each.find("DiskId") != -1:
                    tmpDetail["DiskId"] = each.replace("[DiskId:", " ").replace("]", " ").strip();
                    if mapRet.has_key(chunkIndex):
                        mapRet[chunkIndex].append(copy.deepcopy(tmpDetail))
                    else:
                        mapRet[chunkIndex] = [copy.deepcopy(tmpDetail)]

        return mapRet;

    # 判断文件在磁盘上是否存在
    # ssh 192.168.109.236 'find /apsarapangu/disk7/ -name "243065084182531*" '
    def JudgeChunkExists(self, fileId, chunkIndex, chunkDetail):
        retResult = []
        if chunkDetail.has_key(chunkIndex):
            for each in chunkDetail[chunkIndex]:
                ip = each["Location"].split("//")[1].split(":")[0]
                strCmd = "ssh " + ip + ' sudo find /apsarapangu/disk' + each["DiskId"] + ' -name "' + fileId + '_' + chunkIndex + '*"'
                print strCmd
                nRet, strOut = CProcess.RunCmd4NoException(strCmd);
                if nRet == 0 and len(strOut) > 0:
                    retResult.append(ip + ":" + strOut)
                    if self.m_mapChunkExists.has_key(ip):
                        self.m_mapChunkExists[ip].append(strOut)
                        self.m_mapChunkExists[ip].sort()
                    else:
                        self.m_mapChunkExists[ip] = [strOut]

        return retResult

    # 0 replica 分析
    def NoneAnalyze(self):

        mapRet = {}
        # 获取 0 replica abnchunk 的列表
        abnChunks = self.FindReplicaChunkId("none");
        print "find " + '%d'%len(abnChunks) + " files has 0 replica"

        # 如果abnchunk过多， 怀疑cs有批量down机的情况，查一下chunkserver 的状态
        if len(abnChunks) > 1000:
            count, detail = self.GetNotNormalCsCount();
            if count > 3:
                error("find " + '%d'%count + ' abnormal chunkserver, please check first', True)
                error(detail, True)
                return {"######": "find too much abnormal chunkserver, please check first"}, {};

        if len(abnChunks) == 0:
            return {}, {};

        for fileId, chunkIndexs in abnChunks.items():
            filePath = self.GetFilePath(fileId);
            print "FileId: " + fileId + ", FilePath:" + filePath + ", AbnChunks count:" + '%s'%len(chunkIndexs)
            print "    FileDetail", self.GetFileInfo(fileId)

            chunkDetail = self.GetChunkDetail(fileId)
            mapRet[filePath] = []

            # 将 chunk 加入 列表即可
            #for chunk in chunkIndexs:
            #    mapRet[filePath].append({chunk:chunkDetail[chunk]})

            # 检查每一个chunk，文件是否存在

            for chunk in chunkIndexs:
                chunkExist = self.JudgeChunkExists(fileId, chunk, chunkDetail)
                if len(chunkExist) == 0:
                    detail = ""
                    if chunkDetail.has_key(chunk):
                        detail = str(chunkDetail[chunk])
                    mapRet[filePath].append({chunk: ["Cann't find Chunk in chunkserver " + detail]})
                else:
                    mapRet[filePath].append({chunk: chunkExist})


        return mapRet, self.m_mapChunkExists;

    # lessmax  abnchunk 分析，看是否是因为副本数大于cs 数量
    def LessmaxAnalyze(self):

        mapRet = {}

        # 获取cs 数量
        nRet, nCsCount = CProcess.RunCmd4NoException("/apsara/deploy/puadmin lscs|grep tcp|wc -l");
        if nRet != 0:
            error("Get Cs Count failed, /apsara/deploy/puadmin may not exist", True)
            return {}

        # 获取 lessmax  abnchunk 的列表
        abnChunks = self.FindReplicaChunkId("lessmax");
        print "find " + '%d'%len(abnChunks) + " files has lessmax replica"
        if len(abnChunks) == 0:
            return {};

        for fileId, chunkIndexs in abnChunks.items():
            filePath = self.GetFilePath(fileId);
            print "FileId: " + fileId + ", FilePath:" + filePath + ", AbnChunks count:" + '%s'%len(chunkIndexs)
            chunkDetail = self.GetChunkDetail(fileId)

            # 检查每一个文件，看文件副本数是否大于cs 数量
            nRet, strRet = CProcess.RunCmd4NoException("/apsara/deploy/puadmin gfi " + fileId);
            for each in strRet.split("\n"):
                if each.find("MaxCopy:") != -1:
                    maxCopy = int(each.split(":")[1])
                    if maxCopy > int(nCsCount):
                        mapRet[fileId] = "FilePath: " + filePath + ", MaxCopy:" + '%s'%maxCopy + " > CsCount:" + nCsCount

        return mapRet;

    # lessmin  abnchunk 分析, 生成两个脚本，一个用来把副本数降低，一个恢复，用来消除lessmin的abnchunk
    def LessminExecute(self):

        mapRet = {}

        # 获取cs 数量
        nRet, nCsCount = CProcess.RunCmd4NoException("/apsara/deploy/puadmin lscs|grep tcp|wc -l");
        if nRet != 0:
            error("Get Cs Count failed, /apsara/deploy/puadmin may not exist", True)
            return {}

        # 获取 lessmin abnchunk 的列表
        abnChunks = self.FindReplicaChunkId("lessmin");
        print "find " + '%d'%len(abnChunks) + " files has lessmin replica"
        if len(abnChunks) == 0:
            return;

        # 将文件列表生成出来
        lsPath = []
        for fileId, chunkIndexs in abnChunks.items():
            nRet, strRet = CProcess.RunCmd4NoException("/apsara/deploy/puadmin whois " + fileId);
            for each in strRet.split('\n'):
                if each.find(fileId) != -1:
                    lsPath.append("pangu://localcluster" +each.split()[-1])

        # 生成两个脚本，一个用来降低副本数，一个恢复
        nowTime=datetime.datetime.now().strftime('%Y-%m-%d_%H:%M:%S')
        self.GenLessMinRepairStript("lessmin_abnchunnk_to_2_2_" + nowTime + ".sh", "2", lsPath)
        info("lessmin_abnchunnk_to_2_2.sh gen success, please chmod +x,and run it first", True)
        self.GenLessMinRepairStript("lessmin_abnchunnk_to_3_3_" + nowTime + ".sh", "3", lsPath)
        info("lessmin_abnchunnk_to_3_3.sh gen success, please chmod +x, wait 1 hour run it second", True)

    def GenLessMinRepairStript(self, filename, replicanum, lsPath):
        dstFile = os.path.join(__strCurrentDir__, filename)
        dstFd = open(dstFile, "w")
        for each in lsPath:
            dstFd.write("/apsara/deploy/pu setreplica " + each + " " + replicanum + " " + replicanum + "\n");
        dstFd.close()

    # onecopy  abnchunk 分析, 生成两个脚本，一个用来把副本数降低，一个恢复，用来消除lessmin的abnchunk
    def OnecopyExecute(self):

        mapRet = {}

        # 获取cs 数量
        nRet, nCsCount = CProcess.RunCmd4NoException("/apsara/deploy/puadmin lscs|grep tcp|wc -l");
        if nRet != 0:
            error("Get Cs Count failed, /apsara/deploy/puadmin may not exist", True)
            return {}

        # 获取 lessmin abnchunk 的列表
        abnChunks = self.FindReplicaChunkId("onecopy");
        print "find " + '%d'%len(abnChunks) + " files has onecopy replica"
        if len(abnChunks) == 0:
            return;

        # 将文件列表生成出来
        lsPath = []
        for fileId, chunkIndexs in abnChunks.items():
            nRet, strRet = CProcess.RunCmd4NoException("/apsara/deploy/puadmin whois " + fileId);
            for each in strRet.split('\n'):
                if each.find(fileId) != -1:
                    lsPath.append("pangu://localcluster" +each.split()[-1])

        # 生成两个脚本，一个用来降低副本数，一个恢复
        nowTime=datetime.datetime.now().strftime('%Y-%m-%d_%H:%M:%S')
        self.GenLessMinRepairStript("onecopy_abnchunnk_to_1_1_" + nowTime + ".sh", "1", lsPath)
        info("onecopy_abnchunnk_to_1_1.sh gen success, please chmod +x,and run it first", True)
        self.GenLessMinRepairStript("onecopy_abnchunnk_to_3_3_" + nowTime + ".sh", "3", lsPath)
        info("onecopy_abnchunnk_to_3_3.sh gen success, please chmod +x, wait 1 hour run it second", True)


if '__main__' == __name__:

    if  len(sys.argv) != 2:
        print ("Usage: " + sys.argv[0] + " [none|lessmin|lessmax]")
        exit(2)

    obj = AbnchunkAnalyze();
    type =  sys.argv[1];

    if type == "none":
        ret, ret2 = obj.NoneAnalyze()
        info("0 replica abnchunk analyze result: " + json.dumps(ret, indent = 4), True)
        info("exist chunk distribute: " + json.dumps(ret2, indent = 4), True)
    elif type == "lessmax":
        maxRet = obj.LessmaxAnalyze()
        info("lessmax abnchunk analyze result: " + json.dumps(maxRet, indent = 4), True)
    elif type == "lessmin":
        obj.LessminExecute();
    elif type == "onecopy":
        obj.OnecopyExecute();
    else:
        print ("Usage: " + sys.argv[0] + " [none|onecopy|lessmin|lessmax]")


