#!/home/tops/bin/python2.7
# -*- coding: utf-8 -*-
import os
import sys
import time
import commands
import json

__strCurrentDir__ = os.path.abspath(os.path.dirname(__file__))
__strModuleDir__ = os.path.dirname(__strCurrentDir__)
if __strModuleDir__ not in sys.path:
    sys.path.insert(0, __strModuleDir__)

from common.log import info, warning, error, fatal
from  pangu_status_collect import PanguStatusCollect

class PanguStatusCheck:
    def __init__(self):
        self.m_result = {};
        self.m_collect = PanguStatusCollect();
        self.m_clusterName = self.m_collect.GetClusterName()

    def loadSaveData(self):

        filename = "pangu_status_" + self.m_clusterName + ".json"

        path = os.path.join(__strCurrentDir__, filename)
        if not os.path.exists(path):
            error("saved result file is not exist, path" + path, True)
            return -1, "";

        return 0, json.loads(open(os.path.join(path)).read())

    # 主逻辑
    def Run(self):
        info("begin check pangu status", True)

        # 读取保存的pangu 状态
        ret, savedResult = self.loadSaveData()
        if ret != 0:
            error("load saved result file failed", True)
            return 1;

        # 收集当前pangu 的状态
        if self.m_collect.GetLogId() != 0:
            return 2;

        if self.m_collect.GetAbnchunk() != 0:
            return 3;

        if self.m_collect.GetCapability() != 0:
            return 4;

        result = self.m_collect.GetData()

        # 对pangu 状态进行对比，比较标准：
        # 1、AbnChunk_Count* 不能增长, 不能出现多余的abnchunk
        # 2、PrimaryLogId NormalCsTotalCount TotalDiskSizeGB 不能减少，pangu状态不能回退
        lsResult = []
        lsCheckKey = ["PrimaryLogId", "NormalCsTotalCount", "AbnChunk_Count_none", "AbnChunk_Count_onecopy", "AbnChunk_Count_lessmin", "TotalDiskSizeGB"]
        for each in lsCheckKey:
            if each not in savedResult.keys():
                errorInfo = "the key " + each + " not exist in saveResult"
                lsResult.append(errorInfo)
                error(errorInfo, True)
                continue;
            if (each.startswith("AbnChunk_Count") and result[each] > savedResult[each]) or (not each.startswith("AbnChunk_Count") and result[each] < savedResult[each]):
                errorInfo = each + ": check failed, old value is : " + str(savedResult[each]) + ", curr value is : " + str(result[each])
                lsResult.append(errorInfo)
                error(errorInfo, True)
            else:
                successInfo = each + ": pass check, old value is : " + str(savedResult[each]) + ", curr value is : " + str(result[each])
                info(successInfo, True)

        if len(lsResult) > 0:
            error("\n+++++ pangu status check FAIL, ClusterName: " + self.m_clusterName + '\n', True)
            return 5;

        info("\n>>>>> pangu status check PASS, ClusterName: " + self.m_clusterName + '\n', True)

        return 0;

if __name__ == '__main__':

    obj = PanguStatusCheck();
    ret = obj.Run();
    sys.exit(ret)

