#!/home/tops/bin/python2.7
# -*- coding: utf-8 -*-
import os
import sys
import commands
import json
import yaml
import uuid;
import copy

__strCurrentDir__ = os.path.abspath(os.path.dirname(__file__))
__strModuleDir__ = os.path.dirname(__strCurrentDir__)
if __strModuleDir__ not in sys.path:
    sys.path.insert(0, __strModuleDir__)

from common.common import CProcess
from common.common import CEnv
from common.log import info, error

#
# service-pangu 中的模板是从公有云中继承过来，有许多修改
#
# ##### 监控模板部分修改 #####
# service-pangu 基线库里有38+ 个盘古模板
# 更新 pangu_monitor/conf/pangu_monitor_items.json 等类似文件时，因为模板数量太多，手工操作容易出错
# 写代码来解决
# 1、优先从tianji 上下载对应的模板，看里面是否都有对应文件，如果有，则使用里面的
# 2、如果 tianji 模板不存在，则从代码里取，根据cluster_type 的tag 拿到对应版本和对应应用的配置文件
# 3、对文件进行修改，修改后，写到目标位置
#
# pangu_global_monitor.json 这个有合法性问题，先删掉里面的注释
#
# 注意：
#  1、hdfs 那两个模板的分支不是 016*， 不在此处理，需要手动处理
#  2、注意代码分支的正确性，一定要切换到正确的分支
#  3、处理后 json 文件会重新dump，diff的差异可能会比较大
#
# ##### flag 部分修改 #####
# 有一些flag 需要增加，并制定指定的值
# 有一些flag 需要去掉，检查有无性
# ADS 模板比较特殊，需要渲染，名字为 pangu_chunkserver_flag.json.jj2, ADS 的模板跳过，请手工修改
#
# ##### tianjimon 检查 ####
# 检查 tianjimon 是否为 base-template, 仅检查，不修改
#
# #### 公有云和专有云模板 对比功能 ####
# 比较flag 的值是否一致，以 self.m_lsFlagFile 为比较列表，并跳过专有云特化的值
#
#
class ServicePanguConfUpdate:
    def __init__(self):
        # 要操作的监控配置文件列表
        self.m_lsDstFile = ["pangu_monitor/conf/pangu_monitor_items.json", "pangu_supervisor/conf/pangu_global_monitor.json"]

        # 要操作的flag 文件列表
        self.m_lsFlagFile = ["pangu_master/pangu_master_flag.json", "pangu_chunkserver/pangu_chunkserver_flag.json", "pangu_supervisor/pangu_supervisor_flag.json"]
        #self.m_lsFlagFile = ["pangu_chunkserver/pangu_chunkserver_flag.json"]

        # 要修改的基线库本地存放地址，即要操作的本地目录
        self.m_TempleTargetDir = "/apsarapangu/disk5/yanning/apsara_stack/service-pangu/tianji_templates"

        # 从ppe 上下载的tianji 模板的存放地址, 请手动自己下载，脚本不提供下载功能
        self.m_tianjiTmpDir = "/apsarapangu/disk5/yanning/apsara_stack/all_template_export/all_for_update"

        #代码路径
        self.m_codePath_016 = "/apsarapangu/disk5/yanning/trunk_016/pangu/tianji_config/bootstrap"
        self.m_codePath_0161 = "/apsarapangu/disk5/yanning/trunk_0161/pangu/tianji_config/bootstrap"
        self.m_codePath_0162 = "/apsarapangu/disk5/yanning/trunk_0162/pangu/tianji_config/bootstrap"

        #监控项里要修改的内容
        self.m_updateContent = {}
        monitorModify = []
        monitorModify.append({"key": "ms_memory_speed", "field": "enable", "value": 0})
        monitorModify.append({"key": "check_top_temp_file_size", "field": "enable", "value": 0})
        monitorModify.append({"key": "check_top_temp_file_count", "field": "enable", "value": 0})
        monitorModify.append({"key": "cs_ssdcache_lifetime", "field": "enable", "value": 0})

        supervisorModify = []
        supervisorModify.append({"key": "check_pangu_usage", "field": "interval", "value": 3600})
        supervisorModify.append({"key": "check_pangu_usage", "field": "warning", "value": "75"})
        supervisorModify.append({"key": "check_pangu_usage", "field": "error", "value": "80"})


        self.m_updateContent["pangu_monitor_items.json"] = monitorModify
        self.m_updateContent["pangu_global_monitor.json"] = supervisorModify

        # flag 里的需要修改的内容,
        # action 支持 add 或delete，如果是add，则修改或增加，如果是 delete，则删除
        # scope 支持all 和 hybrid，分别用来应用到全部模板或混合存储模板上

        masterFlagModify = []
        masterFlagModify.append({"key": "pangu_master_DelayTimeForFileGC", "value": "86400", "action": "delete", "scope": "all"})
        masterFlagModify.append({"key": "pangu_master_IgnoreMasterStateInNuwa", "value": "false", "action": "add", "scope": "all"})

        chunkserverFlagModify = []
        chunkserverFlagModify.append({"key": "pangu_chunkserver_DeleteTempChunkSubDirFileIntervalTime", "value": "172800", "action": "add", "scope": "all"})
        chunkserverFlagModify.append({"key": "pangu_chunkserver_SSD_DumpWriteMaxRetryTimes", "value": "1000000000", "action": "add", "scope": "hybrid"})
        chunkserverFlagModify.append({"key": "pangu_chunkserver_SSD_DumpReadMaxRetryTimes", "value": "1000000000", "action": "add", "scope": "hybrid"})
        chunkserverFlagModify.append({"key": "pangu_chunkserver_SSD_DumpSyncMaxRetryTimes", "value": "1000000000", "action": "add", "scope": "hybrid"})
        chunkserverFlagModify.append({"key": "pangu_chunkserver_FileSystemStartTimeout", "value": "86400", "action": "add", "scope": "hybrid"})

        supervisorFlagModify = []
        supervisorFlagModify.append({"key": "pangu_supervisor_MaxConcurrentChunkserverShutdownCount", "value": "1", "action": "add", "scope": "all"})
        supervisorFlagModify.append({"key": "pangu_supervisor_ChunkserverShutdownSetBlackList", "value": "false", "action": "add", "scope": "all"})
        supervisorFlagModify.append({"key": "pangu_supervisor_CheckAtLastTwoRepLoopCountForOfflineCs", "value": "false", "action": "add", "scope": "all"})
        supervisorFlagModify.append({"key": "pangu_supervisor_SupportRmaEtcActions", "value": "false", "action": "delete", "scope": "all"})

        self.m_updateContent["pangu_master_flag.json"] = masterFlagModify
        self.m_updateContent["pangu_chunkserver_flag.json"] = chunkserverFlagModify
        self.m_updateContent["pangu_supervisor_flag.json"] = supervisorFlagModify


    def  TraverseDirForUpdateMonitor(self):
        print "BEGIN check and update monitor and gmonitor config"
        for each in os.listdir(self.m_TempleTargetDir):
            # HDFS 那两个模板的分支是 dfs 分支，手工处理
            if each.find("HDFS") != -1:
                print "skip", each, "please check"
                continue

            # 获取代码路径，及tag
            codePath, tag = self.GetCodePath(each)
            for dst in self.m_lsDstFile:
                # 获取要操作的文件路径，优先级依次为 模板，ppe模板，代码
                srcFilePath = self.GetTargetFilePath(each, dst, codePath, tag)
                targetFilePath = os.path.join(self.m_TempleTargetDir, each, "user", dst)
                self.ExeuteOneMonitor(srcFilePath, targetFilePath)
        print "END check and update monitor and gmonitor config\n"


    # 根据输入的模板名称，获取代码路径，及tag
    def GetCodePath(self, path):
        codePath = self.m_codePath_016
        if path.find("0.16.2") != -1:
            codePath = self.m_codePath_0162
        elif path.find("0.16.1") != -1:
            codePath = self.m_codePath_0161

        ret = yaml.load(open(os.path.join(self.m_TempleTargetDir, path, "tag.conf")))
        tag = ret["Tags"]["CLUSTER_TYPE"]

        return codePath, tag

    def GetTargetFilePath(self, template, dstFile, codePath, tag):
        # 判断目标是否存在
        dstFileFullPath = os.path.join(self.m_TempleTargetDir, template, "user", dstFile);
        if os.path.exists(dstFileFullPath):
            # 如果目标文件存在，直接返回
            return  dstFileFullPath
        else:
            # 如果目标文件不存在，检查上一级文件夹是否存在，如果不存在，就创建
            dstFilePath = dstFileFullPath[:-len(dstFileFullPath.split("/")[-1])]
            if not os.path.exists(dstFilePath):
                os.makedirs(dstFilePath)

        # 判断 ppe 模板中对应的文件是否存在
        templateFilePath = os.path.join(self.m_tianjiTmpDir, template, "user", dstFile)
        if os.path.exists(templateFilePath):
            return templateFilePath

        # 判断代码中是否存在
        codeFilePath = os.path.join(codePath, dstFile[: -len(dstFile.split('/')[-1])], tag, dstFile.split("/")[-1])
        if os.path.exists(codeFilePath):
            return codeFilePath

        return ""


    # 修改监控选项，包含 pangu_monitor_items.json 和 pangu_global_monitor.json
    def ExeuteOneMonitor(self, srcFilePath, targetFilePath):
        # 因为 pangu_global_monitor.json 里面有一些 // 开头的注释，这个json 解析不出来，要先处理掉
        src = open(srcFilePath, "r").read()
        dst = ""
        if srcFilePath.find("pangu_global_monitor.json") != -1:
            for each in src.split("\n"):
                if each.find("//") != -1:
                    dst = dst + each.split("//")[0]
                else:
                    dst = dst + each
        else:
            dst = src

        srcJson = json.loads(dst)

        lsModify = []
        if srcFilePath.find("pangu_monitor_items.json") != -1:
            lsModify =  self.m_updateContent["pangu_monitor_items.json"]
        elif srcFilePath.find("pangu_global_monitor.json") != -1:
            lsModify = self.m_updateContent["pangu_global_monitor.json"]

        updateTimes = 0;
        for each in lsModify:
            key =  each["key"]
            if  srcJson.has_key(key):
                #判断对应的字段是否存在，如果不存在，就输出
                if srcJson[key].has_key(each["field"]):
                    if srcJson[key][each["field"]] !=  each["value"]:
                        print "update ",  key, " for ", targetFilePath
                        srcJson[key][each["field"]] =  each["value"]
                        updateTimes = updateTimes + 1;
                else:
                    print "standard is", each, "but", each["field"], "not exist in file", targetFilePath

        # 只有监控项发生变化才需要修改
        if updateTimes > 0:
            open(targetFilePath, "w").write(json.dumps(srcJson, indent = 4, sort_keys = True))
            # 不知道为什么，dumps 后，行后面多了个空格，处理掉
            s = open(targetFilePath, 'r').read();
            ls = [x.rstrip() for x in s.split('\n')];
            open(targetFilePath, 'w').write('\n'.join(ls))
            print "update", targetFilePath

    def  TraverseDirForUpdateFlag(self):
        print "BEGIN check and modify flag"
        for each in os.listdir(self.m_TempleTargetDir):
            # HDFS DCM 的分支是 dfs 分支，手工处理
            # ADS 需要手动处理
            if each.find("TMPL-PANGU-ADS") != -1 or (each.find("HDFS") != -1 and each.find("DCM") != -1):
                print "skip", each, "please manual processing"
                continue

            isHybrid = self.IsHybrid(each)

            # 依次对每一个flag 文件进行修改
            for flagFile in self.m_lsFlagFile:
                targetFilePath = os.path.join(self.m_TempleTargetDir, each, "user", flagFile)
                self.ExecuteOneFlag(targetFilePath, isHybrid)

        print "END check and modify flag\n"


    def ExecuteOneFlag(self, targetFilePath, isHybrid):
        srcJson = {}
        # 如果目标文件不存在，检查上一级文件夹是否存在，如果不存在，就创建
        if not os.path.exists(targetFilePath):
            filePath = targetFilePath[:-len(targetFilePath.split("/")[-1])]
            if not os.path.exists(filePath):
                os.makedirs(filePath)
                print "create dir", filePath
        else:
            src = open(targetFilePath, "r").read()
            srcJson = json.loads(src)

        lsModify = []
        if targetFilePath.find("pangu_master_flag.json") != -1:
            lsModify =  self.m_updateContent["pangu_master_flag.json"]
        elif targetFilePath.find("pangu_chunkserver_flag.json") != -1:
            lsModify = self.m_updateContent["pangu_chunkserver_flag.json"]
        elif targetFilePath.find("pangu_supervisor_flag.json") != -1:
            lsModify = self.m_updateContent["pangu_supervisor_flag.json"]

        #chunkserverFlagModify.append({"key": "pangu_chunkserver_FileSystemStartTimeout", "value": "86400", "action": "add", "scope": "hybrid"})
        updateTimes = 0;
        for each in lsModify:
            needDo = False
            if each["scope"]  == "all" or (each["scope"]  == "hybrid" and isHybrid):
                needDo = True;

            if needDo:
                if each["action"] == "add":
                    if srcJson.has_key(each["key"]):
                        if srcJson[each["key"]] != each["value"]:
                            # 修改flag，赋值
                            print "modify key", each["key"],"from", srcJson[each["key"]], "to", each["value"]
                            srcJson[each["key"]] = each["value"]
                            updateTimes = updateTimes + 1
                    else:
                        # flag不存在，增加
                        srcJson[each["key"]] = each["value"]
                        updateTimes = updateTimes + 1

                        print "add key ", each["key"], "=", each["value"]
                elif each["action"] == "delete":
                    if srcJson.has_key(each["key"]):
                        # 删除key
                        print "delete key", each["key"]
                        srcJson.pop(each["key"])
                        updateTimes = updateTimes + 1

        if updateTimes > 0:
            open(targetFilePath, "w").write(json.dumps(srcJson, indent = 4, sort_keys = True))
            # 不知道为什么，dumps 后，行后面多了个空格，处理掉
            s = open(targetFilePath, 'r').read();
            ls = [x.rstrip() for x in s.split('\n')];
            open(targetFilePath, 'w').write('\n'.join(ls))
            print "update", targetFilePath

    # 根据模板名字判读是否为混合存储
    def IsHybrid(self, templateName):
        if templateName.find("SATASSD") != -1 or templateName.find("HYBRID") != -1 or templateName.find("ADS") != -1:
            return True;

        return False;

    # 检查 tianjiMonTemplate 是不是 base-template
    def  TraverseDirForCheckTianjiMon(self):
        print "BEGIN check tianjiMonTemplate"
        for each in os.listdir(self.m_TempleTargetDir):

            targetFilePath = os.path.join(self.m_TempleTargetDir, each, "tianji", "tianjimon.conf")

            src = open(targetFilePath, "r").read()
            srcJson = json.loads(src)
            if srcJson["tianjiMonTemplate"] != "base-template":
                print targetFilePath, "tianjiMonTemplate is not base-template, please modify it"

        print "END check tianjiMonTemplate\n"


    # 模板对比功能，对比公有云和专有云模板差异，跳过特化的部分
    def TraverseDirForDiff(self):
        print "BEGIN diff public cloud and apsara stack "

        lsApsaraStackTemplate = os.listdir(self.m_TempleTargetDir)
        lsApsaraStackTemplate.sort()
        for each in lsApsaraStackTemplate:
            if not os.path.exists(os.path.join(self.m_tianjiTmpDir, each)):
                print each, "is not exist in public, SKIP"
                continue;

            if each.find("apsara_stack") != -1 or each.find("HDFS") != -1:
                print "Skip", each, "this template if used only for apsara stack"
                continue;

            for flagFile in self.m_lsFlagFile:
                apsaraStackFilePath = os.path.join(self.m_TempleTargetDir, each, "user", flagFile)
                publicFilePath = os.path.join(self.m_tianjiTmpDir, each, "user", flagFile)
                self.DiffOneFlagFile(apsaraStackFilePath, publicFilePath)

        print "END diff public cloud and apsara stack\n "

    # 比较两个flag 文件的不同，跳过特化的部分，双向比较
    def DiffOneFlagFile(self, apsaraStackFile, publicFile):
        lsResult = []
        apsaraStackJson = {}
        if  os.path.exists(apsaraStackFile):
            apsaraStackJson = json.loads(open(apsaraStackFile, "r").read())
        publicJson = {}
        if  os.path.exists(publicFile):
            publicJson = json.loads(open(publicFile, "r").read())

        lsSkip = []
        if apsaraStackFile.find("pangu_master_flag.json") != -1:
            lsSkip = self.m_updateContent["pangu_master_flag.json"]
        elif apsaraStackFile.find("pangu_chunkserver_flag.json") != -1:
            lsSkip = self.m_updateContent["pangu_chunkserver_flag.json"]
        elif apsaraStackFile.find("pangu_supervisor_flag.json") != -1:
            lsSkip = self.m_updateContent["pangu_supervisor_flag.json"]

        diffSet = set()
        for each in lsSkip:
            diffSet.add(each["key"])

        for k, v in apsaraStackJson.items():
            if k in diffSet:
                continue
            diffSet.add(k)
            if publicJson.has_key(k):
                if v != publicJson[k]:
                    lsResult.append(k +  " is diff, apsarastack = " + str(v) + " public = " + str(publicJson[k]))
            else:
                lsResult.append(k + " = " + str(v) + ", ONLY exist in APSARA_STACK")

        for publicK, publicV in publicJson.items():
            if publicK in diffSet:
                continue;
            diffSet.add(publicK)
            if apsaraStackJson.has_key(publicK):
                if publicV != apsaraStackJson[publicK]:
                    lsResult.append(publicK + " is diff, public = " + str(publicV) + " apsarastack = ", str(apsaraStackJson[publicK]))
            else:
                lsResult.append(str(publicK) + " = " + str(publicV) +  ", ONLY exist in PUBLIC")

        if  lsResult:
            print "\nFIND DIFF IN", apsaraStackFile
            print json.dumps(lsResult, indent = 4)

if '__main__' == __name__:

    obj = ServicePanguConfUpdate();
#    obj.TraverseDirForUpdateMonitor()
#    obj.TraverseDirForUpdateFlag()
#    obj.TraverseDirForCheckTianjiMon();
    obj.TraverseDirForDiff();



