#!/home/tops/bin/python
# -*- coding: utf-8 -*-
import sys
import os
import json
import datetime
import argparse
import traceback
import hashlib
from hashlib import sha1
import urllib, urllib2
import base64
import hmac
import time
import uuid
from optparse import OptionParser
import ConfigParser
# from Crypto.PublicKey import RSA
import yaml
import subprocess
import tempfile
import shutil

reload(sys)
sys.setdefaultencoding('utf-8')

# work_dir/
#   |_ tianji_clt.conf
#   |_ AT01/
#     |_ a.conf
#     |_ .tianji/
#         |_ meta
#         |_ origin_repo/
#             |_ a.conf
#
gMetaDir = ".tianji"
gSkipList = [gMetaDir, ".git"]
gMetaFile = "meta"
gOriginRepo = "origin_repo"
# cluster and baseversion is stored in gMetaFile
gCluster = "cluster"
gBaseVersion = "base_version"
gMd5OfPkgFiles = 'md5_of_pkg_files'

Default_Beiming_Address = "10.101.172.46"


def checkIfTianjiGitDir(root):
    if not os.path.exists(os.path.join(root, gMetaDir)):
        return False
    if not os.path.exists(os.path.join(root, gMetaDir, gMetaFile)):
        return False
    if not os.path.exists(os.path.join(root, gMetaDir, gOriginRepo)):
        return False
    return True


def getClusterName(root):
    metaFile = os.path.join(root, gMetaDir, gMetaFile)
    with open(metaFile) as f:
        meta = json.load(f)
    return meta[gCluster]


def getBaseVersion(root):
    metaFile = os.path.join(root, gMetaDir, gMetaFile)
    with open(metaFile) as f:
        meta = json.load(f)
    return meta[gBaseVersion]


def saveBaseVersion(root, baseVersion):
    metaFile = os.path.join(root, gMetaDir, gMetaFile)
    with open(metaFile) as f:
        meta = json.load(f)
    meta[gBaseVersion] = baseVersion
    with open(metaFile, 'w') as f:
        json.dump(meta, f)


def SaveFile(filePath, fileContent):
    f = open(filePath, "w")
    f.write(fileContent)
    f.close()


def ExecCmdWithoutCheck(cmd):
    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()
    return (p.returncode, stdout, stderr)


def Exec(cmd):
    ret, stdout, stderr = ExecCmdWithoutCheck(cmd)
    if ret:
        msg = "exec cmd[%s] failed, stdout[%s], stderr[%s]" % (cmd, stdout, stderr)
        ReturnError(500, msg)
    return ret, stdout, stderr


def ValidateMachineReplace(clusterConfPath, roleConfPath, sourceMachine, destMachine):
    roleEvalBinary = os.path.join(os.path.dirname(os.path.abspath(__file__)), "role_eval")
    if not os.path.exists(roleEvalBinary):
        msg = "ValidateMachineReplace failed, binary[role_eval] not exist"
        ReturnError(404, msg)
    cmd = "%s -RoleConf %s -ClusterConf %s" % (roleEvalBinary, roleConfPath, clusterConfPath)
    rc, stdout, stderr = Exec(cmd)
    serverRoles = json.loads(stdout)
    for serverRole in serverRoles:
        if serverRole.startswith("TianjiClient#") or serverRole.startswith("DockerDaemon#"):
            continue
        if sourceMachine in serverRoles[serverRole] and destMachine in serverRoles[serverRole]:
            msg = "ValidateMachineReplace failed, [%s] and [%s] exist in the same serverRole[%s]" % (
            sourceMachine, destMachine, serverRole)
            ReturnError(409, msg)


def DoMachineReplace(roleConfStr, sourceMachine, destMachine):
    roleConf = yaml.load(roleConfStr)
    if roleConf.has_key("MachineGroups"):
        for machineGroup in roleConf["MachineGroups"]:
            if type(roleConf["MachineGroups"][machineGroup]) != type("str"):
                if sourceMachine in roleConf["MachineGroups"][machineGroup]:
                    roleConf["MachineGroups"][machineGroup].remove(sourceMachine)
                    if destMachine not in roleConf["MachineGroups"][machineGroup]:
                        roleConf["MachineGroups"][machineGroup].append(destMachine)
    if roleConf.has_key("ServerRoles"):
        for serverRole in roleConf["ServerRoles"]:
            if type(roleConf["ServerRoles"][serverRole]) != type("str"):
                if sourceMachine in roleConf["ServerRoles"][serverRole]:
                    if not serverRole.startswith("TianjiClient#") and not serverRole.startswith("DockerDaemon#"):
                        roleConf["ServerRoles"][serverRole].remove(sourceMachine)
                    if destMachine not in roleConf["ServerRoles"][serverRole]:
                        roleConf["ServerRoles"][serverRole].append(destMachine)
                index = 0
                for machine in roleConf["ServerRoles"][serverRole]:
                    if machine.startswith(sourceMachine + ":"):
                        strIndex = machine.find(":")
                        roleConf["ServerRoles"][serverRole][index] = destMachine + machine[strIndex:]
                    index = index + 1

    return yaml.dump(roleConf)


class NotFound(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class VersionNotExist(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class TianjiAPI:
    def __init__(self, configPath, beiming_address="", tianji_api_address="", primary_api_address=0,
                 tianji_api_addresses=[], project_name="", access_key_id="", access_key_secret="",
                 tempConfigPath="/tmp/tianji.conf"):
        if configPath == "":
            self.configPath = tempConfigPath
            self.primaryApiAddress = primary_api_address
            self.primaryApiAddresses = []
            for addr in tianji_api_addresses:
                self.primaryApiAddresses.append(addr.rstrip("/"))
            self.userKeyID = access_key_id
            self.userKeySecret = access_key_secret
            self.projectName = project_name
            self.beimingAddress = beiming_address
            self.baseUrl = self.primaryApiAddresses[self.primaryApiAddress]
            return

        try:
            with open(configPath) as f:
                config = json.load(f)
        except Exception as err:
            ReturnError(500, "Load config(" + configPath + ") failed: " + str(err))

        beiming_address = ""
        tianji_api_address = ""
        primary_api_address = 0
        tianji_api_addresses = []
        access_key_id = ""
        access_key_secret = ""

        if "beiming_address" in config:
            beiming_address = config["beiming_address"]
        if "tianji_api_address" in config:
            tianji_api_address = config["tianji_api_address"].rstrip("/")
        if "tianji_api_addresses" in config:
            tianji_api_addresses = []
            for addr in config["tianji_api_addresses"]:
                tianji_api_addresses.append(addr.rstrip("/"))
        if tianji_api_address != "" and len(tianji_api_addresses) == 0:
            tianji_api_addresses.append(tianji_api_address.rstrip("/"))
        if "primary_api_address" in config:
            primary_api_address = config["primary_api_address"]
            if len(tianji_api_addresses) > 0:
                primary_api_address = primary_api_address % len(tianji_api_addresses)

        # to compatible with legacy
        if "access_id" in config:
            access_key_id = config["access_id"]
        if "access_key_id" in config:
            access_key_id = config["access_key_id"]
        if "access_key" in config:
            access_key_secret = config["access_key"]
        if "access_key_secret" in config:
            access_key_secret = config["access_key_secret"]

        self.configPath = configPath
        self.primaryApiAddress = primary_api_address
        self.primaryApiAddresses = tianji_api_addresses
        self.userKeyID = str(access_key_id)
        self.userKeySecret = str(access_key_secret)
        if "project_name" in config:
            self.projectName = str(config["project_name"])
        else:
            self.projectName = ""
        self.beimingAddress = str(beiming_address)
        self.baseUrl = self.primaryApiAddresses[self.primaryApiAddress]

    def GetRegions(self):
        apiUrl = "/api/regions"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def CheckoutCluster(self, projectName, clusterName, version="", extend=False, plain=False):
        if version == "":
            apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "?plain=" + str(plain).lower()
            if extend == True:
                apiUrl = apiUrl + "&extend=true"
        else:
            apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "?version=" + version + "&plain=" + str(
                plain).lower()
            if extend == True:
                apiUrl = apiUrl + "&extend=true"

        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue['ClusterConf'], retValue['Version']

    def SubmitCluster(self, projectName, clusterName, version, commitDelta, commitMessage):
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "?version=" + version
        params = {}
        params["X-Tianji-Comment"] = commitMessage.strip()
        ret = self.http_post(self.userKeyID, self.userKeySecret, apiUrl, commitDelta, params)
        return ret

    def CreateProject(self, projectName):
        apiUrl = "/api/projects/" + projectName
        self.http_put(self.userKeyID, self.userKeySecret, apiUrl, "")

    def ListProjects(self):
        apiUrl = "/api/allprojects"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def DeleteProject(self, projectName):
        apiUrl = "/api/projects/" + projectName
        self.http_delete(self.userKeyID, self.userKeySecret, apiUrl)

    def CreateService(self, serviceName, isSysService="false"):
        apiUrl = "/api/services/" + serviceName + "?issys=" + isSysService
        self.http_put(self.userKeyID, self.userKeySecret, apiUrl, "")

    def ListServices(self):
        apiUrl = "/api/allservices"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def ServiceExists(self, serviceName):
        services = self.ListServices()
        if serviceName in services:
            return True
        else:
            return False

    def DeleteService(self, serviceName):
        apiUrl = "/api/services/" + serviceName
        self.http_delete(self.userKeyID, self.userKeySecret, apiUrl)

    def CreateCluster(self, projectName, clusterName, createJson):
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName
        ret = self.http_put(self.userKeyID, self.userKeySecret, apiUrl, createJson)
        return ret

    def ListResource(self, project, cluster):
        #  apiUrl = "/api/projects/" + project + "/clusters/" + cluster + "/serverroles/" + serverrole + "/resources"
        apiUrl = "/api/projects/" + project + "/clusters/" + cluster + "/resources"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)

    def DeleteCluster(self, projectName, clusterName):
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName
        if self.GetClusterState(projectName, clusterName) != "pre-deleted":
            self.http_put(self.userKeyID, self.userKeySecret, apiUrl, {"Operation": "Pre-Delete"})
        self.http_delete(self.userKeyID, self.userKeySecret, apiUrl)

    # Elgant Poweron Cluster: operationJson = {"Operation": "Start"}
    # Elgant Poweroff Cluster: operationJson = {"Operation": "Stop"}
    def OperateCluster(self, projectName, clusterName, operationJson):
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName
        ret = self.http_put(self.userKeyID, self.userKeySecret, apiUrl, operationJson)
        return ret

    def ListClusters(self, projectName):
        apiUrl = "/api/projects/" + projectName + "/clusters"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        clusterList = []
        for item in retValue:
            clusterList.append(item["ClusterName"])
        return clusterList

    def ClusterExists(self, projectName, clusterName):
        clusters = self.ListClusters(projectName)
        if clusterName in clusters:
            return True
        return False

    def GetClusterForServiceTemplate(self, serviceName, templateName):
        apiUrl = "/api/services/" + serviceName + "/templates/" + templateName + "/serviceinstances"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def GetClusterVersion(self, clusterName):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/serviceinstances?detail=false"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue['Version']

    def GetClusterState(self, projectName, clusterName):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue['ClusterState'].lower()

    def GetClusterLog(self, clusterName, logNum):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/deploylogs?count=" + str(logNum)
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def GetClusterLogContent(self, clusterName, version):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + self.projectName + "/clusters/" + clusterName + "/deploylogs/" + version
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def GetServiceInstanceLog(self, clusterName, serviceName, logNum):
        apiUrl = "/api/services/" + serviceName + "/serviceinstances/" + clusterName + "/deploylogs?count=" + str(
            logNum)
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def GetServiceInstanceLogContent(self, clusterName, serviceName, version):
        apiUrl = "/api/services/" + serviceName + "/serviceinstances/" + clusterName + "/deploylogs/" + version
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def GetClusterVersionFromLog(self, clusterName):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/deploylogs?count=1"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        if not isinstance(retValue, list) or len(retValue) == 0:
            ReturnError(404, "Can not get cluster version")
        return retValue[0]['Version']

    def CreateServiceInstance(self, clusterName, serviceName):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/serviceinstances/" + serviceName
        self.http_put(self.userKeyID, self.userKeySecret, apiUrl, "")

    def DeleteServiceInstance(self, clusterName, serviceName):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/serviceinstances/" + serviceName
        ret = self.http_delete(self.userKeyID, self.userKeySecret, apiUrl)
        return ret

    def SubmitService(self, clusterName, serviceName, version, commitDelta, commitMessage):
        apiUrl = "/api/services/" + serviceName + "/serviceinstances/" + clusterName + "?version=" + version
        params = {}
        params["X-Tianji-Comment"] = commitMessage.strip()
        ret = self.http_post(self.userKeyID, self.userKeySecret, apiUrl, commitDelta, params)
        return ret

    def ListServiceInstance(self, clusterName, detail=False):
        projectName = self.GetClusterProject(clusterName)
        if detail == True:
            apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/serviceinstances"
        else:
            apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/serviceinstances?detail=false"

        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def ListInstanceInService(self, serviceName):
        apiUrl = "/api/services/" + serviceName + "/serviceinstances"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def ListInstanceServerrolesInCluster(self, clusterName, serviceName):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/serviceinstances/" + serviceName + "/serverroles?detail=false"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def ListInstanceServerroleMachinesInCluster(self, clusterName, serviceName, serverroleName):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/serviceinstances/" + serviceName + "/serverroles/" + serverroleName + "/machines?detail=false"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def ListInstanceServerrolesInService(self, clusterName, serviceName):
        apiUrl = "/api/services/" + serviceName + "/serviceinstances/" + clusterName + "/serverroles?detail=false"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def ListInstanceServerroleMachinesInService(self, clusterName, serviceName, serverroleName):
        apiUrl = "/api/services/" + serviceName + "/serviceinstances/" + clusterName + "/serverroles/" + serverroleName + "/machines?detail=false"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def GetServiceInstanceServerroles(self, clusterName, serviceName):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/serviceinstances/" + serviceName + "/serverroles"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def GetServiceInstanceTags(self):
        apiUrl = "/api/serviceinstancetags"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def GetServiceInstanceByTag(self, tag):
        apiUrl = "/api/serviceinstancetags/" + tag
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def GetServiceInstanceTagInCluster(self, clusterName, serviceName):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/serviceinstances/" + serviceName + "/tags"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def AddServiceInstanceTagInCluster(self, clusterName, serviceName, tag):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/serviceinstances/" + serviceName + "/tags/" + tag
        self.http_put(self.userKeyID, self.userKeySecret, apiUrl, "")

    def DeleteServiceInstanceTagInCluster(self, clusterName, serviceName, tag):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/serviceinstances/" + serviceName + "/tags/" + tag
        self.http_delete(self.userKeyID, self.userKeySecret, apiUrl)

    def GetServiceInstanceTagInService(self, clusterName, serviceName):
        apiUrl = "/api/services/" + serviceName + "/serviceinstances/" + clusterName + "/tags"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def AddServiceInstanceTagInService(self, clusterName, serviceName, tag):
        apiUrl = "/api/services/" + serviceName + "/serviceinstances/" + clusterName + "/tags/" + tag
        self.http_put(self.userKeyID, self.userKeySecret, apiUrl, "")

    def DeleteServiceInstanceTagInService(self, clusterName, serviceName, tag):
        apiUrl = "/api/services/" + serviceName + "/serviceinstances/" + serviceName + "/tags/" + tag
        self.http_delete(self.userKeyID, self.userKeySecret, apiUrl)

    def CreateServiceTemplate(self, serviceName, templateName, contentJson, commitMessage, desc="", group=""):
        templates = self.ListServiceTemplate(serviceName)
        exist = False
        for template in templates:
            if template["TmplateName"] == templateName:
                exist = True
        if exist:
            raise Exception, "CreateServiceTemplate failed, service[%s]'s template[%s] has exist" % (
            serviceName, templateName)

        apiUrl = "/api/services/" + serviceName + "/templates/" + templateName
        params = {}
        params["X-Tianji-Comment"] = commitMessage.strip()
        params["X-Tianji-Description"] = desc
        params["X-Tianji-Group"] = group
        ret = self.http_post(self.userKeyID, self.userKeySecret, apiUrl, contentJson, params)
        return ret

    def ListServiceTemplate(self, serviceName):
        apiUrl = "/api/services/" + serviceName + "/templates"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def CheckServiceTemplateExist(self, serviceName, templateName, apiName):
        templates = self.ListServiceTemplate(serviceName)
        exist = False
        for template in templates:
            if template["TmplateName"] == templateName:
                exist = True
        if not exist:
            raise Exception, "%s failed, service[%s]'s template[%s] not exist" % (apiName, serviceName, templateName)

    def UpdateServiceTemplate(self, serviceName, templateName, contentJson, commitMessage, desc="", group=""):
        self.CheckServiceTemplateExist(serviceName, templateName, "UpdateServiceTemplate")

        apiUrl = "/api/services/" + serviceName + "/templates/" + templateName
        params = {}
        params["X-Tianji-Comment"] = commitMessage.strip()
        params["X-Tianji-Description"] = desc
        params["X-Tianji-Group"] = group
        ret = self.http_post(self.userKeyID, self.userKeySecret, apiUrl, contentJson, params)
        return ret

    def DeleteServiceTemplate(self, serviceName, templateName):
        self.CheckServiceTemplateExist(serviceName, templateName, "DeleteServiceTemplate")
        apiUrl = "/api/services/" + serviceName + "/templates/" + templateName
        self.http_delete(self.userKeyID, self.userKeySecret, apiUrl)

    def GetServiceTemplateLog(self, serviceName, templateName, logNum):
        apiUrl = "/api/services/" + serviceName + "/templates/" + templateName + "/logs?count=" + str(logNum)
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def GetServiceTemplate(self, serviceName, templateName, version="", plain=False):
        self.CheckServiceTemplateExist(serviceName, templateName, "GetServiceTemplate")
        apiUrl = "/api/services/" + serviceName + "/templates/" + templateName + "?version=" + version + "&plain=" + str(
            plain).lower()
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return yaml.load(retValue["Content"])

    def GetServiceInstanceTemplateReferenceInCluster(self, projectName, clusterName, serviceName):
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/serviceinstances/" + serviceName + "/template"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def SetClusterTemplateReference(self, projectName, clusterName, serviceName, templateName, version):
        self.CheckServiceTemplateExist(serviceName, templateName, "SetServiceTemplateReference")
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/serviceinstances/" + serviceName + "/template?name=" + templateName + "&version=" + version
        content = {}
        params = {}
        ret = self.http_post(self.userKeyID, self.userKeySecret, apiUrl, content, params)
        return ret

    def DeleteClusterTemplateReference(self, projectName, clusterName, serviceName):
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/serviceinstances/" + serviceName + "/template"
        ret = self.http_delete(self.userKeyID, self.userKeySecret, apiUrl)
        return ret

    def GetServiceInstanceTemplateReferenceInService(self, clusterName, serviceName):
        apiUrl = "/api/services/" + serviceName + "/serviceinstances/" + clusterName + "/template"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def SetServiceTemplateReference(self, clusterName, serviceName, templateName, version):
        self.CheckServiceTemplateExist(serviceName, templateName, "SetServiceTemplateReference")
        apiUrl = "/api/services/" + serviceName + "/serviceinstances/" + clusterName + "/template?name=" + templateName + "&version=" + version
        content = {}
        params = {}
        ret = self.http_post(self.userKeyID, self.userKeySecret, apiUrl, content, params)
        return ret

    def DeleteServiceTemplateReference(self, clusterName, serviceName):
        apiUrl = "/api/services/" + serviceName + "/serviceinstances/" + clusterName + "/template"
        ret = self.http_delete(self.userKeyID, self.userKeySecret, apiUrl)
        return ret

    def CreateServiceSchema(self, serviceName, schemaName, contentJson, commitMessage, desc=""):
        schemas = self.ListServiceSchema(serviceName)
        exist = False
        for schema in schemas:
            if schema["SchemaName"] == schemaName:
                exist = True
        if exist:
            raise Exception, "CreateServiceSchema failed, service[%s]'s schema[%s] has exist" % (
            serviceName, schemaName)

        apiUrl = "/api/services/" + serviceName + "/schemas/" + schemaName
        params = {}
        params["X-Tianji-Comment"] = commitMessage.strip()
        params["X-Tianji-Description"] = desc
        ret = self.http_post(self.userKeyID, self.userKeySecret, apiUrl, contentJson, params)
        return ret

    def ListServiceSchema(self, serviceName):
        apiUrl = "/api/services/" + serviceName + "/schemas"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def CheckServiceSchemaExist(self, serviceName, schemaName, apiName):
        schemas = self.ListServiceSchema(serviceName)
        exist = False
        for schema in schemas:
            if schema["SchemaName"] == schemaName:
                exist = True
        if not exist:
            raise Exception, "%s failed, service[%s]'s schema[%s] not exist" % (apiName, serviceName, schemaName)

    def UpdateServiceSchema(self, serviceName, schemaName, contentJson, commitMessage, desc=""):
        self.CheckServiceSchemaExist(serviceName, schemaName, "UpdateServiceSchema")

        apiUrl = "/api/services/" + serviceName + "/schemas/" + schemaName
        params = {}
        params["X-Tianji-Comment"] = commitMessage.strip()
        params["X-Tianji-Description"] = desc
        ret = self.http_post(self.userKeyID, self.userKeySecret, apiUrl, contentJson, params)
        return ret

    def DeleteServiceSchema(self, serviceName, schemaName):
        self.CheckServiceSchemaExist(serviceName, schemaName, "DeleteServiceSchema")
        apiUrl = "/api/services/" + serviceName + "/schemas/" + schemaName
        self.http_delete(self.userKeyID, self.userKeySecret, apiUrl)

    def GetServiceSchemaLog(self, serviceName, schemaName, logNum):
        apiUrl = "/api/services/" + serviceName + "/schemas/" + schemaName + "/logs?count=" + str(logNum)
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def GetServiceSchema(self, serviceName, schemaName, version=""):
        self.CheckServiceSchemaExist(serviceName, schemaName, "GetServiceSchema")
        apiUrl = "/api/services/" + serviceName + "/schemas/" + schemaName + "?version=" + version
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue["Content"]

    def CreateServiceValidator(self, serviceName, validatorName, contentJson, commitMessage, desc=""):
        validators = self.ListServiceValidator(serviceName)
        exist = False
        for validator in validators:
            if validator["ValidatorName"] == validatorName:
                exist = True
        if exist:
            raise Exception, "CreateServiceValidator failed, service[%s]'s validator[%s] has exist" % (
            serviceName, validatorName)

        apiUrl = "/api/services/" + serviceName + "/validators/" + validatorName
        params = {}
        params["X-Tianji-Comment"] = commitMessage.strip()
        params["X-Tianji-Description"] = desc
        ret = self.http_post(self.userKeyID, self.userKeySecret, apiUrl, contentJson, params)
        return ret

    def ListServiceValidator(self, serviceName):
        apiUrl = "/api/services/" + serviceName + "/validators"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def CheckServiceValidatorExist(self, serviceName, validatorName, apiName):
        validators = self.ListServiceValidator(serviceName)
        exist = False
        for validator in validators:
            if validator["ValidatorName"] == validatorName:
                exist = True
        if not exist:
            raise Exception, "%s failed, service[%s]'s validator[%s] not exist" % (apiName, serviceName, validatorName)

    def UpdateServiceValidator(self, serviceName, validatorName, contentJson, commitMessage, desc=""):
        self.CheckServiceValidatorExist(serviceName, validatorName, "UpdateServiceValidator")

        apiUrl = "/api/services/" + serviceName + "/validators/" + validatorName
        params = {}
        params["X-Tianji-Comment"] = commitMessage.strip()
        params["X-Tianji-Description"] = desc
        ret = self.http_post(self.userKeyID, self.userKeySecret, apiUrl, contentJson, params)
        return ret

    def DeleteServiceValidator(self, serviceName, validatorName):
        self.CheckServiceValidatorExist(serviceName, validatorName, "DeleteServiceValidator")
        apiUrl = "/api/services/" + serviceName + "/validators/" + validatorName
        self.http_delete(self.userKeyID, self.userKeySecret, apiUrl)

    def GetServiceValidatorLog(self, serviceName, validatorName, logNum):
        apiUrl = "/api/services/" + serviceName + "/validators/" + validatorName + "/logs?count=" + str(logNum)
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def GetServiceValidator(self, serviceName, validatorName, version=""):
        self.CheckServiceValidatorExist(serviceName, validatorName, "GetServiceValidator")
        apiUrl = "/api/services/" + serviceName + "/validators/" + validatorName + "?version=" + version
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue["Content"]

    def CreateServiceTjmonTemplate(self, serviceName, templateName, contentJson, commitMessage):
        templates = self.ListServiceTjmonTemplate(serviceName)
        exist = False
        for template in templates:
            if template["TmplateName"] == templateName:
                exist = True
        if exist:
            raise Exception, "CreateServiceTjmonTemplate failed, service[%s]'s template[%s] has exist" % (
            serviceName, templateName)

        apiUrl = "/api/services/" + serviceName + "/tjmontemplates/" + templateName
        params = {}
        params["X-Tianji-Comment"] = commitMessage.strip()
        params["X-Tianji-Description"] = commitMessage.strip()
        ret = self.http_post(self.userKeyID, self.userKeySecret, apiUrl, contentJson, params)
        return ret

    def ListServiceTjmonTemplate(self, serviceName):
        apiUrl = "/api/services/" + serviceName + "/tjmontemplates"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return retValue

    def CheckServiceTjmonTemplateExist(self, serviceName, templateName, apiName):
        templates = self.ListServiceTjmonTemplate(serviceName)
        exist = False
        for template in templates:
            if template["TmplateName"] == templateName:
                exist = True
        if not exist:
            raise Exception, "%s failed, service[%s]'s Tjmontemplate[%s] not exist" % (
            apiName, serviceName, templateName)

    def UpdateServiceTjmonTemplate(self, serviceName, templateName, contentJson, commitMessage):
        self.CheckServiceTjmonTemplateExist(serviceName, templateName, "UpdateServiceTjmonTemplate")

        apiUrl = "/api/services/" + serviceName + "/tjmontemplates/" + templateName
        params = {}
        params["X-Tianji-Comment"] = commitMessage.strip()
        params["X-Tianji-Description"] = commitMessage.strip()
        ret = self.http_post(self.userKeyID, self.userKeySecret, apiUrl, contentJson, params)
        return ret

    def DeleteServiceTjmonTemplate(self, serviceName, templateName):
        apiUrl = "/api/services/" + serviceName + "/tjmontemplates/" + templateName
        self.http_delete(self.userKeyID, self.userKeySecret, apiUrl)

    def GetServiceTjmonTemplate(self, serviceName, templateName, version=""):
        apiUrl = "/api/services/" + serviceName + "/tjmontemplates/" + templateName + "?version=" + version
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        retValue = yaml.load(ret)
        return yaml.load(retValue["Content"])

    def GetServiceInstanceListByTjmonTmplName(self, serviceName, templateName, mode=""):
        if mode == "legacy":
            apiUrl = "/api/services/" + serviceName + "/tjmontemplates/" + templateName + "/serviceinstances?legacy=true"
        else:
            apiUrl = "/api/services/" + serviceName + "/tjmontemplates/" + templateName + "/serviceinstances"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def GetRemoteDiff(self, clusterName, version1, version2):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/diff?v1=" + version1 + "&v2=" + version2
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return ret

    def GetClusterProject(self, clusterName):
        apiUrl = "/api/clusters/" + clusterName + "/project"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return ret

    def ListAllRolling(self, clusterName):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/rollings"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def ListRolling(self, clusterName, version):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/rollings?commit_id=" + version
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def GetRolling(self, clusterName, version, subVersion=""):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/rollings/" + version + "?subversion=" + subVersion
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def OperateRolling(self, clusterName, version, operationJson, subVersion=""):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/rollings/" + version + "?subversion=" + subVersion
        ret = self.http_put(self.userKeyID, self.userKeySecret, apiUrl, operationJson)
        return yaml.load(ret)

    def ListMachineByProjectAndSR(self, project, serverrole):
        apiUrl = "/api/machines?project=" + project + "&serverrole=" + serverrole
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def ListMachineByIP(self, ip):
        apiUrl = "/api/machines?ip=" + ip
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def ListMachineByCluster(self, clusterName):
        apiUrl = "/api/machines?cluster=" + clusterName
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return json.loads(ret)

    def GetMachineInfo(self, hostname, attr, type=""):
        apiUrl = "/api/machines/" + hostname + "?attr=" + attr
        if type == "local":
            apiUrl = apiUrl + "&type=local"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def BatchGetMachineInfo(self, hostnames, attr, type=""):
        apiUrl = "/api/machines/batchgetinfo?attr=" + attr
        if type == "local":
            apiUrl = apiUrl + "&type=local"
        params = {}
        ret = self.http_post(self.userKeyID, self.userKeySecret, apiUrl, hostnames, params)
        return yaml.load(ret)

    def ListMachineSRByCluster(self, clusterName, serverRole=None):
        if serverRole != None:
            apiUrl = "/api/machinesserverroles?cluster=" + clusterName + "&serverrole=" + serverRole
        else:
            apiUrl = "/api/machinesserverroles?cluster=" + clusterName
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def ListMachineSRByHost(self, hostName):
        apiUrl = "/api/machinesserverroles?hostname=" + hostName
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def GetMachineSR(self, hostName, serverRole):
        apiUrl = "/api/machines/" + hostName + "/serverroles/" + serverRole
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def ListMachineSRMonitor(self, hostName, serverRole):
        apiUrl = "/api/machines/" + hostName + "/serverroles/" + serverRole + "/monitors"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def GetClusterSRMonitor(self, projectName, clusterName, serverRoleName, monitorName):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/serverroles/" + serverRoleName + "/monitors/" + monitorName
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return json.loads(ret)

    def ListBuildJob(self, clusterName, version):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/builds?cluster=" + clusterName + "&commit_id=" + version
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def GetBuildInfo(self, clusterName, version):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/builds/" + version
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def ListAdmin(self):
        apiUrl = "/api/admin"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def AddAdmin(self, adminuser):
        apiUrl = "/api/admin?opid=" + adminuser
        self.http_put(self.userKeyID, self.userKeySecret, apiUrl, "")

    def RemoveAdmin(self, adminuser):
        apiUrl = "/api/admin?opid=" + adminuser
        self.http_delete(self.userKeyID, self.userKeySecret, apiUrl)

    def ListProjectUser(self, projectName):
        apiUrl = "/api/projects/" + projectName + "/users"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def ListServiceUser(self, serviceName):
        apiUrl = "/api/services/" + serviceName + "/users"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def AddProjectUser(self, userName, projectName, isOwner):
        if isOwner:
            apiUrl = "/api/projects/" + projectName + "/users?opid=" + userName + "&beowner=true"
        else:
            apiUrl = "/api/projects/" + projectName + "/users?opid=" + userName
        self.http_put(self.userKeyID, self.userKeySecret, apiUrl, "")

    def AddServiceUser(self, userName, serviceName, isOwner):
        if isOwner:
            apiUrl = "/api/services/" + serviceName + "/users?opid=" + userName + "&beowner=true"
        else:
            apiUrl = "/api/services/" + serviceName + "/users?opid=" + userName
        self.http_put(self.userKeyID, self.userKeySecret, apiUrl, "")

    def RemoveProjectUser(self, userName, projectName):
        apiUrl = "/api/projects/" + projectName + "/users?opid=" + userName
        self.http_delete(self.userKeyID, self.userKeySecret, apiUrl)

    def RemoveServiceUser(self, userName, serviceName):
        apiUrl = "/api/services/" + serviceName + "/users?opid=" + userName
        self.http_delete(self.userKeyID, self.userKeySecret, apiUrl)

    def TurnOnAPIServer(self):
        apiUrl = "/api/passport"
        content = {}
        params = {}
        self.http_post(self.userKeyID, self.userKeySecret, apiUrl, content, params)

    def GetMode(self):
        apiUrl = "/api/mode"
        return self.http_get(self.userKeyID, self.userKeySecret, apiUrl)

    def EnableMode(self, ummJson, mode=""):
        apiUrl = "/api/mode?mode=" + mode
        self.http_put(self.userKeyID, self.userKeySecret, apiUrl, ummJson)

    def FetchFileList(self, projectName, clusterName, hostName, serviceName, serverroleName, appName, path, pattern):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/serviceinstances/" + serviceName + "/serverroles/" + serverroleName + "/machines/" + hostName + "/apps/" + appName + "/filelist?path=" + path + "&pattern=" + pattern
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def FetchAllDir(self, projectName, clusterName, hostName, serviceName, serverroleName, appName):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/serviceinstances/" + serviceName + "/serverroles/" + serverroleName + "/machines/" + hostName + "/apps/" + appName + "/filelist?type=alldir"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def FetchLog(self, projectName, clusterName, hostName, serviceName, serverroleName, appName, path, pattern, num,
                 bytes, delimiter):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/serviceinstances/" + serviceName + "/serverroles/" + serverroleName + "/machines/" + hostName + "/apps/" + appName + "/debuglog?path=" + path + "&pattern=" + pattern + "&num=" + num + "&bytes=" + bytes + "&delimiter=" + delimiter
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return ret

    def DockerInspect(self, projectName, clusterName, serviceName, serverroleName, appName, machine):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/services/" + serviceName + "/serverroles/" + serverroleName + "/apps/" + appName + "/dockerinspect?machine=" + machine
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return ret

    def DockerRestart(self, projectName, clusterName, serviceName, serverroleName, appName, machine):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/services/" + serviceName + "/serverroles/" + serverroleName + "/apps/" + appName + "/dockerrestart?machine=" + machine
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return ret

    def DockerLogs(self, projectName, clusterName, serviceName, serverroleName, appName, machine, count):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/serviceinstances/" + serviceName + "/serverroles/" + serverroleName + "/machines/" + machine + "/apps/" + appName + "/dockerlogs?count=" + str(
            count)
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return ret

    def ListClusterMachine(self, projectName, clusterName):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/machines"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def GetMachineAllSRInfo(self, projectName, clusterName, machine, monitorFlag=False, monitorList=""):
        projectName = self.GetClusterProject(clusterName)
        if monitorFlag == True:
            withMonitor = "?monitor=true"
        else:
            withMonitor = "?monitor=false"

        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/machines/" + machine + withMonitor + "&monitors=" + monitorList
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def AllocateClusterMachine(self, projectName, clusterName, machineRequire):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/machines"
        retVersion = self.http_put(self.userKeyID, self.userKeySecret, apiUrl, machineRequire)
        retJson = {}
        retJson["version"] = retVersion
        return retJson

    def ExtendClusterMachine(self, clusterName, operation, allService, extendJson, commitMessage=""):
        projectName = self.GetClusterProject(clusterName)
        params = {}
        params["X-Tianji-Comment"] = commitMessage.strip()
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/machines?operation=" + operation + "&allservice=" + allService
        retVersion = self.http_put(self.userKeyID, self.userKeySecret, apiUrl, extendJson, params)
        return retVersion

    def RemoveClusterMachine(self, clusterName, operation, removeJson, commitMessage=""):
        projectName = self.GetClusterProject(clusterName)
        params = {}
        params["X-Tianji-Comment"] = commitMessage.strip()
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/machines?operation=" + operation
        retVersion = self.http_put(self.userKeyID, self.userKeySecret, apiUrl, removeJson, params)
        return retVersion

    def ListClusterMachineGroup(self, clusterName):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/machinegroups"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def GetClusterMachineGroup(self, clusterName, machineGroupName):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/machinegroups/" + machineGroupName
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return ret

    def OperateClusterMachineGroup(self, clusterName, operation, machinegroup, tomachinegroup, operateJson,
                                   commitMessage=""):
        projectName = self.GetClusterProject(clusterName)
        params = {}
        params["X-Tianji-Comment"] = commitMessage.strip()
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/machinegroups?operation=" + operation + "&machinegroup=" + machinegroup + "&tomachinegroup=" + tomachinegroup
        retVersion = self.http_put(self.userKeyID, self.userKeySecret, apiUrl, operateJson, params)
        return retVersion

    def GetNextAPIServerAddress(self):
        self.primaryApiAddress = (self.primaryApiAddress + 1) % len(self.primaryApiAddresses)
        self.baseUrl = self.primaryApiAddresses[self.primaryApiAddress]
        return

    def StorePrimaryApiAddress(self):
        configContent = {}
        configContent["primary_api_address"] = self.primaryApiAddress
        configContent["tianji_api_addresses"] = self.primaryApiAddresses
        configContent["access_key_id"] = self.userKeyID
        configContent["access_key_secret"] = self.userKeySecret
        configContent["project_name"] = self.projectName
        configContent["beiming_address"] = self.beimingAddress

        with open(self.configPath, 'w') as f:
            f.write(json.dumps(configContent, sort_keys=True, indent=2))
        return

    def ListMachineBucket(self, projectName):
        apiUrl = "/api/projects/" + projectName + "/machinebuckets"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def GetMachineBucket(self, projectName, bucketName, detail="true"):
        apiUrl = "/api/projects/" + projectName + "/machinebuckets/" + bucketName + "/machines?detail=" + detail
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def CreateMachineBucket(self, projectName, bucketName):
        apiUrl = "/api/projects/" + projectName + "/machinebuckets/" + bucketName
        self.http_put(self.userKeyID, self.userKeySecret, apiUrl, "")
        ret = {}
        return ret

    def DeleteMachineBucket(self, projectName, bucketName):
        apiUrl = "/api/projects/" + projectName + "/machinebuckets/" + bucketName
        self.http_delete(self.userKeyID, self.userKeySecret, apiUrl)
        ret = {}
        return ret

    def UpdateMachineBucket(self, projectName, bucketName, machinesInfo):
        apiUrl = "/api/projects/" + projectName + "/machinebuckets/" + bucketName
        params = {}
        params["X-Tianji-Comment"] = "update machine bucket " + bucketName + " with machines: " + str(machinesInfo)
        ret = self.http_post(self.userKeyID, self.userKeySecret, apiUrl, machinesInfo, params)
        return ret

    def MigrateMachineBucket(self, projectName, bucketName, toProjectName, toBucketName, migrateJson):
        apiUrl = "/api/projects/" + projectName + "/machinebuckets/" + bucketName + "?project=" + toProjectName + "&bucket=" + toBucketName
        params = {}
        params["X-Tianji-Comment"] = "migrate machines " + str(
            migrateJson) + " from project " + projectName + " bucket " + bucketName + " to project " + toProjectName + " bucket " + toBucketName
        ret = self.http_post(self.userKeyID, self.userKeySecret, apiUrl, migrateJson, params)
        return ret

    def ListClusterQuota(self, projectName, clusterName):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/quota"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def UpdateClusterQuota(self, projectName, clusterName, quotasInfo):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/quota"
        params = {}
        params["X-Tianji-Comment"] = "update cluster quota: " + str(quotasInfo)
        ret = self.http_post(self.userKeyID, self.userKeySecret, apiUrl, quotasInfo, params)
        return ret

    def ListUsergroups(self, projectName):
        apiUrl = "/api/projects/" + projectName + "/usergroups"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def CreateUsergroup(self, projectName, usergroupName, createJson):
        apiUrl = "/api/projects/" + projectName + "/usergroups/" + usergroupName
        self.http_put(self.userKeyID, self.userKeySecret, apiUrl, createJson)
        ret = {}
        return ret

    def DeleteUsergroup(self, projectName, usergroupName):
        apiUrl = "/api/projects/" + projectName + "/usergroups/" + usergroupName
        self.http_delete(self.userKeyID, self.userKeySecret, apiUrl)
        ret = {}
        return ret

    def ShowUsergroup(self, projectName, usergroupName):
        apiUrl = "/api/projects/" + projectName + "/usergroups/" + usergroupName
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def ModifyUsergroup(self, projectName, usergroupName, updateJson):
        apiUrl = "/api/projects/" + projectName + "/usergroups/" + usergroupName
        params = {}
        params["X-Tianji-Comment"] = "update usergroup: " + usergroupName
        ret = self.http_post(self.userKeyID, self.userKeySecret, apiUrl, updateJson, params)
        return ret

    def ListClusterUsergroups(self, projectName, clusterName):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/usergroups"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def CreateClusterUsergroup(self, projectName, clusterName, usergroupName):
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/usergroups/" + usergroupName
        self.http_put(self.userKeyID, self.userKeySecret, apiUrl, "")
        ret = {}
        return ret

    def DeleteClusterUsergroup(self, projectName, clusterName, usergroupName):
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/usergroups/" + usergroupName
        self.http_delete(self.userKeyID, self.userKeySecret, apiUrl)
        ret = {}
        return ret

    def ListMachines(self, projectName, clusterName):
        projectName = self.GetClusterProject(clusterName)
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/machinesinfo"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def ListMachineMonitor(self, machine):
        apiUrl = "/api/machines/" + machine + "/monitors"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def OperateSRMachinesByProject(self, projectName, clusterName, serviceName, serverroleName, machine, action, status,
                                   dimpact="", simpact="", appimpact=""):
        if machine == "Machines":
            apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/serviceinstances/" + serviceName + "/serverroles/" + serverroleName + "/machines?action=" + action + "&status=" + status + "&dimpact=" + dimpact + "&simpact=" + simpact + "&appimpact=" + appimpact
        else:
            apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/serviceinstances/" + serviceName + "/serverroles/" + serverroleName + "/machines/" + machine + "?action=" + action + "&status=" + status + "&dimpact=" + dimpact + "&simpact=" + simpact + "&appimpact=" + appimpact
        ret = self.http_put(self.userKeyID, self.userKeySecret, apiUrl, "")
        return ret

    def DeleteSRMachineActionByProject(self, projectName, clusterName, serviceName, serverroleName, machine, action,
                                       status=""):
        if machine == "Machines":
            apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/serviceinstances/" + serviceName + "/serverroles/" + serverroleName + "/machines?action=" + action + "&status=" + status
        else:
            apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/serviceinstances/" + serviceName + "/serverroles/" + serverroleName + "/machines/" + machine + "?action=" + action + "&status=" + status
        ret = self.http_delete(self.userKeyID, self.userKeySecret, apiUrl)
        return ret

    def OperateSRMachinesByService(self, clusterName, serviceName, serverroleName, machine, action, status="",
                                   dimpact="", simpact="", appimpact=""):
        if machine == "Machines":
            apiUrl = "/api/services/" + serviceName + "/serviceinstances/" + clusterName + "/serverroles/" + serverroleName + "/machines?action=" + action + "&status=" + status + "&dimpact=" + dimpact + "&simpact=" + simpact + "&appimpact=" + appimpact
        else:
            apiUrl = "/api/services/" + serviceName + "/serviceinstances/" + clusterName + "/serverroles/" + serverroleName + "/machines/" + machine + "?action=" + action + "&status=" + status + "&dimpact=" + dimpact + "&simpact=" + simpact + "&appimpact=" + appimpact
        ret = self.http_put(self.userKeyID, self.userKeySecret, apiUrl, "")
        return ret

    def DeleteSRMachineActionByService(self, clusterName, serviceName, serverroleName, machine, action, status=""):
        if machine == "Machines":
            apiUrl = "/api/services/" + serviceName + "/serviceinstances/" + clusterName + "/serverroles/" + serverroleName + "/machines?action=" + action + "&status=" + status
        else:
            apiUrl = "/api/services/" + serviceName + "/serviceinstances/" + clusterName + "/serverroles/" + serverroleName + "/machines/" + machine + "?action=" + action + "&status=" + status
        ret = self.http_delete(self.userKeyID, self.userKeySecret, apiUrl)
        return ret

    def DeleteMachineAction(self, projectName, clusterName, machine, action, status=""):
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/machines/" + machine + "?action=" + action + "&status=" + status
        ret = self.http_delete(self.userKeyID, self.userKeySecret, apiUrl)
        return ret

    def SetMachineAction(self, projectName, clusterName, machine, action, description, status="", dimpact="",
                         simpact="", appimpact=""):
        apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/machines/" + machine + "?action=" + action + "&status=" + status + "&description=" + description + "&dimpact=" + dimpact + "&simpact=" + simpact + "&appimpact=" + appimpact
        ret = self.http_put(self.userKeyID, self.userKeySecret, apiUrl, "")
        return ret

    def GetMachineSRInfoByProject(self, projectName, clusterName, serviceName, serverroleName, machine, monitorFlag,
                                  monitorList=""):
        projectName = self.GetClusterProject(clusterName)
        if monitorFlag == True:
            withMonitor = "?monitor=true"
        else:
            withMonitor = "?monitor=false"

        if machine == "Machines":
            apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/serviceinstances/" + serviceName + "/serverroles/" + serverroleName + "/machines" + withMonitor + "&monitors=" + monitorList
        else:
            apiUrl = "/api/projects/" + projectName + "/clusters/" + clusterName + "/serviceinstances/" + serviceName + "/serverroles/" + serverroleName + "/machines/" + machine + withMonitor + "&monitors=" + monitorList
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def GetMachineSRInfoByService(self, clusterName, serviceName, serverroleName, machine, monitorFlag):
        withMonitor = ""
        if monitorFlag == True:
            withMonitor = "?monitor=true"

        if machine == "Machines":
            apiUrl = "/api/services/" + serviceName + "/serviceinstances/" + clusterName + "/serverroles/" + serverroleName + "/machines" + withMonitor
        else:
            apiUrl = "/api/services/" + serviceName + "/serviceinstances/" + clusterName + "/serverroles/" + serverroleName + "/machines/" + machine + withMonitor
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    # hardware SDK
    def ListIdc(self):
        apiUrl = "/api/idcs"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def ListRoom(self, idcName):
        apiUrl = "/api/idcs/" + idcName + "/rooms"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def ListRack(self, idcName, roomName):
        apiUrl = "/api/idcs/" + idcName + "/rooms/" + roomName + "/racks"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def ListRackunit(self, idcName, roomName, rackName):
        apiUrl = "/api/idcs/" + idcName + "/rooms/" + roomName + "/racks/" + rackName + "/rackunits"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def CreateIdc(self, idcName, attr):
        apiUrl = "/api/idcs/" + idcName
        ret = self.http_put(self.userKeyID, self.userKeySecret, apiUrl, attr)
        return ret

    def CreateRoom(self, idcName, roomName, attr):
        apiUrl = "/api/idcs/" + idcName + "/rooms/" + roomName
        ret = self.http_put(self.userKeyID, self.userKeySecret, apiUrl, attr)
        return ret

    def CreateRack(self, idcName, roomName, rackName, attr):
        apiUrl = "/api/idcs/" + idcName + "/rooms/" + roomName + "/racks/" + rackName
        ret = self.http_put(self.userKeyID, self.userKeySecret, apiUrl, attr)
        return ret

    def CreateRackunit(self, idcName, roomName, rackName, rackunitName, attr):
        apiUrl = "/api/idcs/" + idcName + "/rooms/" + roomName + "/racks/" + rackName + "/rackunits/" + rackunitName
        ret = self.http_put(self.userKeyID, self.userKeySecret, apiUrl, attr)
        return ret

    def SetIdcAttr(self, idcName, attr):
        apiUrl = "/api/idcs/" + idcName
        params = {}
        ret = self.http_post(self.userKeyID, self.userKeySecret, apiUrl, attr, params)
        return ret

    def SetRoomAttr(self, idcName, roomName, attr):
        apiUrl = "/api/idcs/" + idcName + "/rooms/" + roomName
        params = {}
        ret = self.http_post(self.userKeyID, self.userKeySecret, apiUrl, attr, params)
        return ret

    def SetRackAttr(self, idcName, roomName, rackName, attr):
        apiUrl = "/api/idcs/" + idcName + "/rooms/" + roomName + "/racks/" + rackName
        params = {}
        ret = self.http_post(self.userKeyID, self.userKeySecret, apiUrl, attr, params)
        return ret

    def SetRackunitAttr(self, idcName, roomName, rackName, rackunitName, attr):
        apiUrl = "/api/idcs/" + idcName + "/rooms/" + roomName + "/racks/" + rackName + "/rackunits/" + rackunitName
        params = {}
        ret = self.http_post(self.userKeyID, self.userKeySecret, apiUrl, attr, params)
        return ret

    def GetIdcAttr(self, idcName):
        apiUrl = "/api/idcs/" + idcName
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def GetRoomAttr(self, idcName, roomName):
        apiUrl = "/api/idcs/" + idcName + "/rooms/" + roomName
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def GetRackAttr(self, idcName, roomName, rackName):
        apiUrl = "/api/idcs/" + idcName + "/rooms/" + roomName + "/racks/" + rackName
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def GetRackunitAttr(self, idcName, roomName, rackName, rackunitName):
        apiUrl = "/api/idcs/" + idcName + "/rooms/" + roomName + "/racks/" + rackName + "/rackunits/" + rackunitName
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def DeleteIdc(self, idcName):
        apiUrl = "/api/idcs/" + idcName
        ret = self.http_delete(self.userKeyID, self.userKeySecret, apiUrl)
        return ret

    def DeleteRoom(self, idcName, roomName):
        apiUrl = "/api/idcs/" + idcName + "/rooms/" + roomName
        ret = self.http_delete(self.userKeyID, self.userKeySecret, apiUrl)
        return ret

    def DeleteRack(self, idcName, roomName, rackName):
        apiUrl = "/api/idcs/" + idcName + "/rooms/" + roomName + "/racks/" + rackName
        ret = self.http_delete(self.userKeyID, self.userKeySecret, apiUrl)
        return ret

    def DeleteRackunit(self, idcName, roomName, rackName, rackunitName):
        apiUrl = "/api/idcs/" + idcName + "/rooms/" + roomName + "/racks/" + rackName + "/rackunits/" + rackunitName
        ret = self.http_delete(self.userKeyID, self.userKeySecret, apiUrl)
        return ret

    def ListIdcMachine(self, idcName):
        apiUrl = "/api/idcs/" + idcName + "/machines"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def ListRoomMachine(self, idcName, roomName):
        apiUrl = "/api/idcs/" + idcName + "/rooms/" + roomName + "/machines"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def ListRackMachine(self, idcName, roomName, rackName):
        apiUrl = "/api/idcs/" + idcName + "/rooms/" + roomName + "/racks/" + rackName + "/machines"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def ListRackunitMachine(self, idcName, roomName, rackName, rackunitName):
        apiUrl = "/api/idcs/" + idcName + "/rooms/" + roomName + "/racks/" + rackName + "/rackunits/" + rackunitName + "/machines"
        ret = self.http_get(self.userKeyID, self.userKeySecret, apiUrl)
        return yaml.load(ret)

    def BatchGetClusterConf(self, postJson, plain=False):
        apiUrl = "/internal/batchclusterconf?plain=" + str(plain).lower()
        params = {}
        ret = self.http_post(self.userKeyID, self.userKeySecret, apiUrl, postJson, params)
        return ret

    def http_get_raw(self, url):
        try:
            request = urllib2.Request(url)
            response = urllib2.urlopen(request, timeout=200)
            if response.code != 200:
                ReturnError(response.code, response.read())
            else:
                return response.read()
        except Exception as err:
            ReturnError(500, str(err))

    def http_get(self, userKeyID, userKeySecret, apiUrl):
        updatePrimaryApiAddress = 0
        errMessage = ""
        for i in range(0, len(self.primaryApiAddresses)):
            try:
                response = make_request(userKeyID, userKeySecret, self.baseUrl, apiUrl, "GET", "")
                if updatePrimaryApiAddress == 1:
                    self.StorePrimaryApiAddress()

                if response.code != 200:
                    ReturnError(response.code, response.read())
                else:
                    responseContent = response.read()
                    if "<!DOCTYPE html>" in responseContent:
                        raise Exception, "tianji_api_address(%s) not valid" % self.baseUrl
                    return responseContent
            except NotFound as err:
                ReturnError(404, str(err))
            except urllib2.HTTPError, err:
                ReturnError(500, str(err))
            except ReturnErrorException, err:
                ReturnError(500, str(err))
                return
            except Exception as err:
                self.GetNextAPIServerAddress()
                updatePrimaryApiAddress = 1
                errMessage = "Tianji ApiServer not serving, " + str(err)
                continue
        ReturnError(500, errMessage)

    def http_put(self, userKeyID, userKeySecret, apiUrl, values, params={}):
        updatePrimaryApiAddress = 0
        errMessage = ""
        for i in range(0, len(self.primaryApiAddresses)):
            try:
                jdata = json.dumps(values)
                response = make_request(userKeyID, userKeySecret, self.baseUrl, apiUrl, "PUT", jdata, params)
                if updatePrimaryApiAddress == 1:
                    self.StorePrimaryApiAddress()

                if response.code != 200:
                    ReturnError(response.code, response.read())
                else:
                    responseContent = response.read()
                    if "<!DOCTYPE html>" in responseContent:
                        raise Exception, "tianji_api_address(%s) not valid" % self.baseUrl
                    return responseContent
            except urllib2.HTTPError, err:
                ReturnError(500, str(err))
            except ReturnErrorException, err:
                ReturnError(500, str(err))
                return
            except Exception as err:
                self.GetNextAPIServerAddress()
                updatePrimaryApiAddress = 1
                errMessage = "Tianji ApiServer not serving, " + str(err)
                continue
        ReturnError(500, errMessage)

    def http_post(self, userKeyID, userKeySecret, apiUrl, values, params):
        updatePrimaryApiAddress = 0
        errMessage = ""
        for i in range(0, len(self.primaryApiAddresses)):
            try:
                jdata = json.dumps(values)
                response = make_request(userKeyID, userKeySecret, self.baseUrl, apiUrl, "POST", jdata, params)
                if updatePrimaryApiAddress == 1:
                    self.StorePrimaryApiAddress()

                if response.code != 200:
                    ReturnError(response.code, response.read())
                else:
                    responseContent = response.read()
                    if "<!DOCTYPE html>" in responseContent:
                        raise Exception, "tianji_api_address(%s) not valid" % self.baseUrl
                    return responseContent
            except urllib2.HTTPError, err:
                ReturnError(500, str(err))
            except VersionNotExist as err:
                ReturnError(500, str(err))
            except ReturnErrorException, err:
                ReturnError(500, str(err))
            except Exception as err:
                self.GetNextAPIServerAddress()
                updatePrimaryApiAddress = 1
                errMessage = "Tianji ApiServer not serving, " + str(err)
                continue
        ReturnError(500, errMessage)

    def http_delete(self, userKeyID, userKeySecret, apiUrl):
        updatePrimaryApiAddress = 0
        errMessage = ""
        for i in range(0, len(self.primaryApiAddresses)):
            try:
                response = make_request(userKeyID, userKeySecret, self.baseUrl, apiUrl, "DELETE", "")
                if updatePrimaryApiAddress == 1:
                    self.StorePrimaryApiAddress()

                if response.code != 200:
                    ReturnError(response.code, response.read())
                else:
                    responseContent = response.read()
                    if "<!DOCTYPE html>" in responseContent:
                        raise Exception, "tianji_api_address(%s) not valid" % self.baseUrl
                    return responseContent
            except urllib2.HTTPError, err:
                ReturnError(500, str(err))
            except ReturnErrorException, err:
                ReturnError(500, str(err))
                return
            except Exception as err:
                self.GetNextAPIServerAddress()
                updatePrimaryApiAddress = 1
                errMessage = "Tianji ApiServer not serving, " + str(err)
                continue
        ReturnError(500, errMessage)


def ConvertToDeltaJson(fileMap):
    resultDeltaJson = {}
    items = fileMap.keys()
    items.sort(reverse=True)
    for filePath in items:
        ProcessFile(filePath, fileMap[filePath], resultDeltaJson)
    return resultDeltaJson


def ConvertToFileMap(fullJson):
    resultFileMap = {}
    ProcessJsonMap("", fullJson, resultFileMap)
    return resultFileMap


def SaveToDisk(rootDir, fileMap, overwrite=False):
    for elem in fileMap:
        path = os.path.join(rootDir, elem)
        if elem.endswith("/"):
            if not os.path.exists(path):
                cmd = "mkdir -p %s" % path
                ExecCmd(cmd)
        else:
            if os.path.exists(path) and not overwrite:
                raise Exception, "file[%s] has exist" % path
            if not os.path.exists(os.path.dirname(path)):
                cmd = "mkdir -p %s" % os.path.dirname(path)
                ExecCmd(cmd)
            f = open(path, "w")
            f.write(fileMap[elem]["content"])
            f.close()


def LoadFromDisk(rootDir, skipList):
    resultFileMap = {}
    ScanDir(rootDir, "", skipList, resultFileMap)
    return resultFileMap


def DiffFileMap(sourceFileMap, destFileMap):
    diffFileMap = {}
    for sourceFile in sourceFileMap:
        if sourceFile not in destFileMap:
            diffFileMap[sourceFile] = {}
            diffFileMap[sourceFile]["type"] = "delete"
            diffFileMap[sourceFile]["content"] = sourceFileMap[sourceFile]["content"]
        else:
            if sourceFileMap[sourceFile]["content"] != destFileMap[sourceFile]["content"]:
                diffFileMap[sourceFile] = {}
                diffFileMap[sourceFile]["type"] = "update"
                diffFileMap[sourceFile]["content"] = destFileMap[sourceFile]["content"]
    for destFile in destFileMap:
        if destFile not in sourceFileMap:
            diffFileMap[destFile] = {}
            diffFileMap[destFile]["type"] = "add"
            diffFileMap[destFile]["content"] = destFileMap[destFile]["content"]
    return diffFileMap


def CheckMosaicTag(fileMap):
    for f in fileMap:
        if fileMap[f]["type"] != "delete" and fileMap[f]["content"] != None:
            if fileMap[f]["content"].find("<**** Secret Data ****>") != -1:
                return False
    return True


# internal function impl

def get_current_date():
    date = datetime.datetime.strftime(datetime.datetime.utcnow(), "%a, %d %b %Y %H:%M:%S GMT")
    return date


def get_authorization(access_key_id, access_key_secret, method, content_md5, content_type, date, api_resource, params):
    sorted_user_params = sorted(params.items(), key=lambda it: it[0])
    canon_string = '\n'
    for k, v in sorted_user_params:
        canon_string += k.lower() + ':' + v + '\n'
    if "?" in api_resource:
        api_resource_array = api_resource.split("?")
        resource_item = api_resource_array[1].split("&")
        resource_item.sort()
        api_resource_sorted = api_resource_array[0] + '?' + '&'.join(resource_item)
    else:
        api_resource_sorted = api_resource

    stringToSign = method + "\n" + content_md5 + "\n" + content_type + "\n" + date + str(
        canon_string) + api_resource_sorted
    hmacsha1 = hmac.new(access_key_secret, stringToSign, hashlib.sha1)
    signature = base64.b64encode(hmacsha1.digest())
    auth = "Tianji " + access_key_id + ":" + signature
    # print "########[INFO] stringToSign: ", repr(stringToSign), "\n"
    return auth


def make_headers(access_key_id, access_key_secret, method, body, api_resource, params):
    date = get_current_date()
    http_headers = dict()
    if 'POST' == method or 'PUT' == method:
        m = hashlib.md5()
        m.update(body)
        md5_body = base64.encodestring(m.digest()).strip()
        http_headers['Content-MD5'] = md5_body
        http_headers['Content-Type'] = "application/json"
        http_headers['Authorization'] = get_authorization(access_key_id, access_key_secret, method, md5_body,
                                                          "application/json", date, api_resource, params)
    else:
        http_headers['Authorization'] = get_authorization(access_key_id, access_key_secret, method, '', '', date,
                                                          api_resource, params)
    http_headers['Date'] = date

    for key, value in params.items():
        http_headers[str(key)] = str(value)

    return http_headers


def make_request(access_key_id, access_key_secret, api_server_address, api_resource, method, body, params={},
                 quiet=False):
    urlparams = {}
    api_resource_array = api_resource.split("?")
    api_path_array = api_resource_array[0].split("/")
    new_api_path_array = []
    for api_path in api_path_array:
        new_api_path_array.append(urllib.quote(api_path))

    url = api_server_address + "/".join(new_api_path_array)

    if len(api_resource_array) > 1:
        resource_item = api_resource_array[1].split("&")
        for item in resource_item:
            if "=" in item:
                item_array = item.split("=")
                urlparams[item_array[0]] = item_array[1]
        url = url + "?" + urllib.urlencode(urlparams)

    try:
        if 'GET' == method:
            request = urllib2.Request(url)
        if 'POST' == method:
            request = urllib2.Request(url, body)
        if 'PUT' == method:
            request = urllib2.Request(url, body)
            request.get_method = lambda: 'PUT'
        if 'DELETE' == method:
            request = urllib2.Request(url)
            request.get_method = lambda: 'DELETE'

        http_headers = make_headers(access_key_id, access_key_secret, method, body, api_resource, params)

        for key, value in http_headers.items():
            request.add_header(key, value)
        response = urllib2.urlopen(request, timeout=200)
        return response
    except urllib2.HTTPError, e:
        errMsg = e.read().strip()
        if errMsg.startswith("{"):
            errJson = json.loads(errMsg)
            if errJson["Code"] == "VersionNotExist":
                raise VersionNotExist(errMsg)
            if errJson["Message"].find("not found") != -1 and errJson["Message"].find("61") != -1:
                raise NotFound(errMsg)

        ReturnError(e.code, errMsg)
    except Exception as err:
        raise Exception, "http request failed, url[%s], method[%s], err[%s] " % (url, method, err)


class ReturnErrorException(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return self.value


def ReturnError(code, message):
    ret = {}
    ret["code"] = code
    ret["message"] = message
    ret["data"] = {}
    # print json.dumps(ret)
    raise ReturnErrorException, json.dumps(ret)
    # exit(1)


def ProcessFile(filePath, fileInfo, resultDeltaJson):
    index = filePath.find("/")
    # file
    if index == -1:
        rootPart = filePath
        if fileInfo["type"] == "delete":
            resultDeltaJson[rootPart] = None
        else:
            resultDeltaJson[rootPart] = fileInfo["content"]
    # end dir
    elif index == len(filePath) - 1:
        rootPart = filePath[:index]
        if fileInfo["type"] == "delete":
            resultDeltaJson[rootPart] = None
        else:
            if rootPart not in resultDeltaJson:
                resultDeltaJson[rootPart] = {}
    # inner dir
    else:
        rootPart = filePath[:index]
        if rootPart not in resultDeltaJson:
            resultDeltaJson[rootPart] = {}
        filePath = filePath[index + 1:]
        ProcessFile(filePath, fileInfo, resultDeltaJson[rootPart])


def ProcessJsonMap(prefix, jsonMap, resultFileMap):
    for elem in jsonMap:
        if type(jsonMap[elem]) == type("") or type(jsonMap[elem]) == type(u""):
            filePath = os.path.join(prefix, elem)
            if filePath not in resultFileMap:
                resultFileMap[filePath] = {}
            resultFileMap[filePath]["type"] = "add"
            resultFileMap[filePath]["content"] = jsonMap[elem]
        elif type(jsonMap[elem]) == type({}):
            filePath = os.path.join(prefix, elem)
            if not filePath.endswith("/"):
                filePath = filePath + "/"
            if filePath not in resultFileMap:
                resultFileMap[filePath] = {}
            resultFileMap[filePath]["type"] = "add"
            resultFileMap[filePath]["content"] = None
            ProcessJsonMap(filePath, jsonMap[elem], resultFileMap)
        else:
            raise Exception, "unsupport map elem type[%s], key[%s], it must be string or dict" % (
            type(jsonMap[elem]), elem)


def ExecCmd(cmd):
    # print "exec cmd[%s]" % cmd
    if os.system(cmd):
        raise Exception, "exec cmd[%s] failed" % cmd


def ScanDir(rootDir, relativeDir, skipList, resultFileMap):
    dirFullPath = os.path.join(rootDir, relativeDir)
    elemList = os.listdir(dirFullPath)
    for elem in elemList:
        if elem in skipList and os.path.samefile(rootDir, dirFullPath):
            continue
        fullPath = os.path.join(dirFullPath, elem)
        relativeFullPath = os.path.join(relativeDir, elem)
        if os.path.isdir(fullPath):
            if not relativeFullPath.endswith("/"):
                relativeFullPath = relativeFullPath + "/"
            if relativeFullPath not in resultFileMap:
                resultFileMap[relativeFullPath] = {}
            resultFileMap[relativeFullPath]["type"] = "add"
            resultFileMap[relativeFullPath]["content"] = None
            ScanDir(rootDir, relativeFullPath, skipList, resultFileMap)
        else:
            if relativeFullPath not in resultFileMap:
                resultFileMap[relativeFullPath] = {}
            resultFileMap[relativeFullPath]["type"] = "add"
            input = open(fullPath, 'r')
            content = input.read()
            input.close()
            resultFileMap[relativeFullPath]["content"] = content
    return resultFileMap


class TianjiSDK(TianjiAPI):
    def __init__(self):
        self.Init = False
        pass

    def initTjSDK(self, tianji_api_addresses, project_name, access_key_id, access_key_secret, tianji_conf_path):
        self.tjSDK = TianjiAPI("", Default_Beiming_Address, "", 0, tianji_api_addresses, project_name, access_key_id,
                               access_key_secret, tianji_conf_path)
        self.Init = True
        self.configPath = tianji_conf_path
        self.primaryApiAddress = 0
        self.primaryApiAddresses = tianji_api_addresses
        self.userKeyID = str(access_key_id)
        self.userKeySecret = str(access_key_secret)
        self.projectName = str(project_name)
        self.beimingAddress = Default_Beiming_Address
        self.baseUrl = self.primaryApiAddresses[self.primaryApiAddress]

        return self.tjSDK

    def CheckSubmitStatus(self, project_name, cluster_name):
        try:
            returnValue = {}
            if not self.Init:
                return yaml.load('{"code: 405", "data": {}, "message": "Tianji SDK is not initialized"}')
            returnValue["code"] = 200
            returnValue["data"] = {}
            returnValue["data"]["machines"] = self.tjSDK.ListClusterMachine(project_name, cluster_name)
            confJson, version = self.tjSDK.CheckoutCluster(project_name, cluster_name)
            if "services" not in confJson or "tianji" not in confJson["services"]:
                ReturnError(409, "tianji-client is not deployed")

            returnValue["data"]["services"] = {}
            for service in confJson["services"]:
                returnValue["data"]["services"][service] = ""

            returnValue["data"]["version"] = version
            return returnValue
        except ReturnErrorException, e:
            return yaml.load(str(e))
        except Exception, e:
            return yaml.load('{"code: 500", "data": {}, "message": "' + str(e) + '"}')

    def CheckDeployStatus(self, project_name, cluster_name, expect_version=""):
        try:
            returnValue = {}
            if not self.Init:
                return yaml.load('{"code: 405", "data": {}, "message": "Tianji SDK is not initialized"}')
            returnValue["code"] = 200

            ret = self.tjSDK.ListMachineSRByCluster(cluster_name)
            for sr in ret:
                retSR = self.tjSDK.GetMachineSR(sr["hostname"], sr["serverrole"])
                if "actual_version" in retSR and retSR["actual_version"] == expect_version and retSR[
                    "state"].lower() == "good":
                    sr["status"] = "Good"
                else:
                    sr["status"] = "Upgrading"

            returnValue["data"] = ret
            return returnValue
        except ReturnErrorException, e:
            return yaml.load(str(e))
        except Exception, e:
            return yaml.load('{"code: 500", "data": {}, "message": "' + str(e) + '"}')

    def DeployCluster(self, project_name, cluster_name, cluster_quota, user_group, machine_spec, deployFileJson={}):
        try:
            returnValue = {}
            if not self.Init:
                return yaml.load('{"code: 405", "data": {}, "message": "Tianji SDK is not initialized"}')

            # create cluster
            create_json = {}
            create_json["Operation"] = "Create"
            create_json["Quotas"] = cluster_quota
            self.tjSDK.CreateCluster(project_name, cluster_name, create_json)

            # create user group
            self.tjSDK.CreateUsergroup(project_name, cluster_name + "_usergroup", user_group)

            # set cluster user group
            self.tjSDK.CreateClusterUsergroup(project_name, cluster_name, cluster_name + "_usergroup")

            # deploy tianji-client, using default template of tianji
            tianjiJson = self.tjSDK.GetServiceTemplate("tianji", "default", "")
            deployJson = {}
            if len(deployFileJson) > 0:
                deployJson = deployFileJson
            deployJson["cluster.conf"] = "Group: DefaultGroup\nMachines: []\nName: %s\nRegion: DefaultRegion" % (
            cluster_name)
            if "sys.conf" not in deployJson:
                deployJson["sys.conf"] = "os: {'*': 'os-default'}\npart: {'*': 'ALdefault'}"
            if "services" not in deployJson:
                deployJson["services"] = {}
            deployJson["services"]["tianji"] = tianjiJson
            self.tjSDK.CreateServiceInstance(cluster_name, "tianji")
            version = self.tjSDK.GetClusterVersion(cluster_name)
            version = self.tjSDK.SubmitCluster(project_name, cluster_name, version, deployJson,
                                               "Deploy default tianji-client")

            # allocate machines
            try:
                ret_version = self.tjSDK.AllocateClusterMachine(project_name, cluster_name, machine_spec)
            except ReturnErrorException, e:
                returnValue["code"] = 404
                returnValue["data"] = {}
                returnValue["message"] = str(e)
                return returnValue

            returnValue["code"] = 200
            returnValue["data"] = {}
            returnValue["data"]["version"] = ret_version

            return returnValue

        except ReturnErrorException, e:
            return yaml.load(str(e))
        except Exception, e:
            return yaml.load('{"code: 500", "data": {}, "message": "' + str(e) + '"}')

    def AllocateMachine(self, project_name, cluster_name, machine_spec):
        try:
            returnValue = {}
            if not self.Init:
                return yaml.load('{"code: 405", "data": {}, "message": "Tianji SDK is not initialized"}')

            ret = self.tjSDK.AllocateClusterMachine(project_name, cluster_name, machine_spec)

            returnValue["code"] = 200
            returnValue["data"] = {}
            returnValue["data"]["version"] = ret

            return returnValue

        except ReturnErrorException, e:
            return yaml.load(str(e))
        except Exception, e:
            return yaml.load('{"code: 500", "data": {}, "message": "' + str(e) + '"}')

    def UpdateMachine(self, project_name, cluster_name, machine_json):
        try:
            returnValue = {}
            if not self.Init:
                return yaml.load('{"code: 405", "data": {}, "message": "Tianji SDK is not initialized"}')

            confJson, version = self.tjSDK.CheckoutCluster(project_name, cluster_name)
            if "cluster.conf" not in confJson:
                ReturnError(409, "cluster is not deployed")

            cluster_conf = yaml.load(confJson["cluster.conf"])
            machine_list = cluster_conf["Machines"]
            for machine in machine_json["Add"]:
                machine_list.append(machine)
            for machine in machine_json["Delete"]:
                machine_list.remove(machine)

            confJson["cluster.conf"] = yaml.dump(cluster_conf)
            ret = self.tjSDK.SubmitCluster(project_name, cluster_name, version, confJson,
                                           "UpdateMachine: " + str(machine_json))

            returnValue["code"] = 200
            returnValue["data"] = {}
            returnValue["data"]["version"] = ret

            return returnValue

        except ReturnErrorException, e:
            return yaml.load(str(e))
        except Exception, e:
            return yaml.load('{"code: 500", "data": {}, "message": "' + str(e) + '"}')

    def DeleteClusterAndUsergroup(self, project_name, cluster_name):
        try:
            returnValue = {}
            if not self.Init:
                return yaml.load('{"code: 405", "data": {}, "message": "Tianji SDK is not initialized"}')

            # delete cluster
            cluster_list = self.tjSDK.ListClusters(project_name)
            if cluster_name in cluster_list:
                self.tjSDK.DeleteCluster(project_name, cluster_name)

            # delete user group
            usergroup_list = self.tjSDK.ListUsergroups(project_name)
            if cluster_name + "_usergroup" in usergroup_list:
                self.tjSDK.DeleteUsergroup(project_name, cluster_name + "_usergroup")

            returnValue["code"] = 200
            returnValue["data"] = {}
            return returnValue

        except ReturnErrorException, e:
            return yaml.load(str(e))
        except Exception, e:
            return yaml.load('{"code: 500", "data": {}, "message": "' + str(e) + '"}')

    def GetQuota(self, project_name, cluster_name):
        try:
            returnValue = {}
            if not self.Init:
                return yaml.load('{"code: 405", "data": {}, "message": "Tianji SDK is not initialized"}')

            ret = self.tjSDK.ListClusterQuota(project_name, cluster_name)
            returnValue["code"] = 200
            returnValue["data"] = ret
            return returnValue

        except ReturnErrorException, e:
            return yaml.load(str(e))
        except Exception, e:
            return yaml.load('{"code: 500", "data": {}, "message": "' + str(e) + '"}')

    def UpdateQuota(self, project_name, cluster_name, quota_json):
        try:
            returnValue = {}
            if not self.Init:
                return yaml.load('{"code: 405", "data": {}, "message": "Tianji SDK is not initialized"}')

            ret = self.tjSDK.UpdateClusterQuota(project_name, cluster_name, quota_json)
            returnValue["code"] = 200
            returnValue["data"] = ret
            return returnValue

        except ReturnErrorException, e:
            return yaml.load(str(e))
        except Exception, e:
            return yaml.load('{"code: 500", "data": {}, "message": "' + str(e) + '"}')

    def GetUsergroup(self, project_name, usergroup_name):
        try:
            returnValue = {}
            if not self.Init:
                return yaml.load('{"code: 405", "data": {}, "message": "Tianji SDK is not initialized"}')

            ret = self.tjSDK.ShowUsergroup(project_name, usergroup_name)
            returnValue["code"] = 200
            returnValue["data"] = ret
            return returnValue

        except ReturnErrorException, e:
            return yaml.load(str(e))
        except Exception, e:
            return yaml.load('{"code: 500", "data": {}, "message": "' + str(e) + '"}')

    def UpdateUsergroup(self, project_name, usergroup_name, usergroup_json):
        try:
            returnValue = {}
            if not self.Init:
                return yaml.load('{"code: 405", "data": {}, "message": "Tianji SDK is not initialized"}')

            ret = self.tjSDK.ModifyUsergroup(project_name, usergroup_name, usergroup_json)
            returnValue["code"] = 200
            returnValue["data"] = ret
            return returnValue

        except ReturnErrorException, e:
            return yaml.load(str(e))
        except Exception, e:
            return yaml.load('{"code: 500", "data": {}, "message": "' + str(e) + '"}')

    def ListClusterMachines(self, project_name, cluster_name):
        try:
            returnValue = {}
            if not self.Init:
                return yaml.load('{"code: 405", "data": {}, "message": "Tianji SDK is not initialized"}')

            ret = self.tjSDK.ListMachines(project_name, cluster_name)
            bucketList = self.tjSDK.ListMachineBucket(project_name)
            vmInfo = {}
            if "tianchi_dynamic_virtual_bucket" in bucketList:
                vmInfo = self.tjSDK.GetMachineBucket(project_name, "tianchi_dynamic_virtual_bucket")

            for vm in ret:
                if vm in vmInfo and "Status" in vmInfo[vm]["Attributes"]:
                    ret[vm]["Attributes"]["Status"] = vmInfo[vm]["Attributes"]["Status"]

            returnValue["code"] = 200
            returnValue["data"] = ret
            return returnValue

        except ReturnErrorException, e:
            return yaml.load(str(e))
        except Exception, e:
            return yaml.load('{"code: 500", "data": {}, "message": "' + str(e) + '"}')

    def ListMachineBuckets(self, project_name):
        try:
            returnValue = {}
            if not self.Init:
                return yaml.load('{"code: 405", "data": {}, "message": "Tianji SDK is not initialized"}')

            ret = self.tjSDK.ListMachineBucket(project_name)
            returnValue["code"] = 200
            returnValue["data"] = ret
            return returnValue

        except ReturnErrorException, e:
            return yaml.load(str(e))
        except Exception, e:
            return yaml.load('{"code: 500", "data": {}, "message": "' + str(e) + '"}')

    def ListBucketMachines(self, project_name, machinebucket_name):
        try:
            returnValue = {}
            if not self.Init:
                return yaml.load('{"code: 405", "data": {}, "message": "Tianji SDK is not initialized"}')

            ret = self.tjSDK.GetMachineBucket(project_name, machinebucket_name)
            returnValue["code"] = 200
            returnValue["data"] = ret
            return returnValue

        except ReturnErrorException, e:
            return yaml.load(str(e))
        except Exception, e:
            return yaml.load('{"code: 500", "data": {}, "message": "' + str(e) + '"}')

    def ListBucketMachineNumber(self, project_name):
        try:
            returnValue = {}
            if not self.Init:
                return yaml.load('{"code: 405", "data": {}, "message": "Tianji SDK is not initialized"}')

            project_bucketmachine_num = {}
            bucket_list = self.tjSDK.ListMachineBucket(project_name)
            for machine_bucket in bucket_list:
                project_bucketmachine_num[machine_bucket] = {}
                project_bucketmachine_num[machine_bucket]["Used"] = 0
                project_bucketmachine_num[machine_bucket]["All"] = 0
                machine_list = self.tjSDK.GetMachineBucket(project_name, machine_bucket)
                for machine in machine_list:
                    if "Attributes" in machine_list[machine] and "TianjiCluster" in machine_list[machine][
                        "Attributes"] and machine_list[machine]["Attributes"]["TianjiCluster"] != "":
                        project_bucketmachine_num[machine_bucket]["Used"] = project_bucketmachine_num[machine_bucket][
                                                                                "Used"] + 1
                    project_bucketmachine_num[machine_bucket]["All"] = project_bucketmachine_num[machine_bucket][
                                                                           "All"] + 1

            returnValue["code"] = 200
            returnValue["data"] = project_bucketmachine_num
            return returnValue

        except ReturnErrorException, e:
            return yaml.load(str(e))
        except Exception, e:
            return yaml.load('{"code: 500", "data": {}, "message": "' + str(e) + '"}')

    def DeployServerRole(self, cluster_name, service_name, serverrole_ip_map):
        try:
            returnValue = {}
            if not self.Init:
                return yaml.load('{"code: 405", "data": {}, "message": "Tianji SDK is not initialized"}')

            project_name = self.GetClusterProject(cluster_name)

            cluster_conf, version = self.CheckoutCluster(project_name, cluster_name)
            serviceInstances = self.ListServiceInstance(cluster_name, False)["ServiceInstances"]

            # deploy user service if no exist
            if service_name not in serviceInstances:
                self.CreateServiceInstance(cluster_name, service_name)

            # generate availability.conf
            availability_conf = {}
            availability_conf["availability"] = {}

            # generate version.conf
            version_conf = {}
            version_conf["Versions"] = {}

            # search hostname for ip from tjmaster
            machine_list = []
            serverrole_host_map = {}
            for sri in serverrole_ip_map:
                serverrole_host_map[sri] = []
                for machineIP in serverrole_ip_map[sri]["machines"]:
                    machineHost = self.ListMachineByIP(machineIP)
                    if len(machineHost) == 0:
                        ReturnError(404, "Can not find hostname for ip: " + machineIP)
                    serverrole_host_map[sri].append(machineHost[0])
                    if machineHost[0] not in machine_list:
                        machine_list.append(machineHost[0])
                availability_conf["availability"][sri] = {}
                availability_conf["availability"][sri]["mode"] = "100%"
                version_conf["Versions"][sri] = {}
                version_conf["Versions"][sri]["Applications"] = {}
                version_conf["Versions"][sri]["Applications"]["*"] = serverrole_ip_map[sri]["version"]

            # generate role.conf
            roleConf = {}
            roleConf["ServerRoles"] = {}
            roleConf["ServerRoles"] = serverrole_host_map

            # generate service conf
            cluster_conf["services"][service_name] = {}
            cluster_conf["services"][service_name]["role.conf"] = yaml.dump(roleConf)
            cluster_conf["services"][service_name]["version.conf"] = yaml.dump(version_conf)
            cluster_conf["services"][service_name]["tianji"] = {}
            cluster_conf["services"][service_name]["tianji"]["availability.conf"] = yaml.dump(availability_conf)

            commitVersion = self.SubmitCluster(project_name, cluster_name, version, cluster_conf,
                                               "Submit serverrole and ip config")

            returnValue["code"] = 200
            returnValue["data"] = commitVersion
            return returnValue

        except ReturnErrorException, e:
            return yaml.load(str(e))
        except Exception, e:
            return yaml.load('{"code: 500", "data": {}, "message": "' + str(e) + '"}')


class TianjiGit:
    def __init__(self):
        pass

    def initTjApi(self, configPath):
        if not os.path.exists(configPath):
            raise Exception, "tianji_clt config file[%s] not exist" % configPath

        tjApi = TianjiAPI(configPath)
        return tjApi

    def _checkout(self, project, cluster, target_dir, overwrite=False, extend=False, plain=False):
        fullJson, baseVersion = self.tjApi.CheckoutCluster(project, cluster, "", extend, plain)
        filemap = ConvertToFileMap(fullJson)
        if not os.path.exists(target_dir):
            os.makedirs(target_dir)
        SaveToDisk(target_dir, filemap, overwrite)
        metaDir = os.path.join(target_dir, gMetaDir)
        if overwrite:
            if not os.path.exists(metaDir):
                os.makedirs(metaDir)
        else:
            if os.path.exists(metaDir):
                raise Exception, "%s already exists" % metaDir
            os.makedirs(metaDir)
        metaFile = os.path.join(metaDir, gMetaFile)
        metaMap = {gCluster: cluster, gBaseVersion: baseVersion}
        with open(metaFile, 'w') as f:
            json.dump(metaMap, f)
        # file map -> target_dir/.tianji/origin_repo/
        originRepoPath = os.path.join(metaDir, gOriginRepo)
        if os.path.exists(originRepoPath):
            cmd = "rm -rf %s" % originRepoPath
            ExecCmd(cmd)
        if not os.path.exists(originRepoPath):
            os.makedirs(originRepoPath)
        SaveToDisk(originRepoPath, filemap, overwrite)

    def _getLocalDiff(self, clusterDir):
        filemap1 = LoadFromDisk(clusterDir, gSkipList)
        metaDir = os.path.join(clusterDir, gMetaDir, gOriginRepo)
        filemap2 = LoadFromDisk(metaDir, gSkipList)
        return DiffFileMap(filemap2, filemap1)

    def Checkout(self, args):
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)

        if args.project:
            project = args.project
        else:
            project = self.tjApi.projectName
        cluster = args.cluster
        target_dir = args.target_dir
        if target_dir == None:
            target_dir = args.cluster
        if os.path.exists(target_dir):
            raise Exception, "%s already exists." % target_dir
        if args.full:
            self._checkout(project, cluster, target_dir, False, True, args.plain)
        else:
            self._checkout(project, cluster, target_dir, False, False, args.plain)
        print "checkout into %s ... ok." % target_dir

    def Log(self, args):
        if not checkIfTianjiGitDir(args.dir):
            raise Exception, "not a tianji_clt dir"
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        cluster = getClusterName(args.dir)
        logs = self.tjApi.GetClusterLog(cluster, args.num)
        print json.dumps(logs, indent=2)

    def Show(self, args):
        if not checkIfTianjiGitDir(args.dir):
            raise Exception, "not a tianji_clt dir"
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        cluster = getClusterName(args.dir)
        logs = self.tjApi.GetClusterLogContent(cluster, args.version)
        print json.dumps(logs, indent=2)

    def Diff(self, args):
        if not checkIfTianjiGitDir(args.dir):
            raise Exception, "not a tianji_clt dir"
        if args.versions:
            configPath = args.config
            self.tjApi = self.initTjApi(configPath)
            cluster = getClusterName(args.dir)
            versions = args.versions
            diff = self.tjApi.GetRemoteDiff(cluster, args.versions[0], versions[1])
            print diff
        else:
            diff = self._getLocalDiff(args.dir)
            print json.dumps(diff, sort_keys=True, indent=2)

    def Update(self, args):
        if not checkIfTianjiGitDir(args.dir):
            raise Exception, "not a tianji_clt dir"
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)

        if args.project:
            project = args.project
        else:
            project = self.tjApi.projectName

        diff = self._getLocalDiff(args.dir)
        if diff:
            raise Exception, "local diff exists, can't update"
        cluster = getClusterName(args.dir)
        # checkout to temp dir first
        tempDir = args.dir + "_" + str(time.time())
        self._checkout(project, cluster, tempDir, True, plain=args.plain)
        # first clear local git, some dir or file may be deleted
        # checkout won't delete them
        cmd = "rm -rf %s" % args.dir
        ExecCmd(cmd)

        cmd = "mv %s %s" % (tempDir, args.dir)
        ExecCmd(cmd)
        print "update to latest version succeed"

    def Push(self, args):
        if not checkIfTianjiGitDir(args.dir):
            raise Exception, "not a tianji_clt dir"
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        if args.project:
            project = args.project
        else:
            project = self.tjApi.projectName

        filemap1 = LoadFromDisk(args.dir, gSkipList)
        metaDir = os.path.join(args.dir, gMetaDir, gOriginRepo)
        filemap2 = LoadFromDisk(metaDir, gSkipList)
        diffFileMap = DiffFileMap(filemap2, filemap1)
        if not CheckMosaicTag(diffFileMap):
            print "There is a mosaic tag in diff list, you can use diff to check it out"
            print "Please use --plain flag in 'checkout' or 'update' to get plain config"
            return ""

        deltaJson = ConvertToDeltaJson(diffFileMap)
        cluster = getClusterName(args.dir)
        version = getBaseVersion(args.dir)
        createdServiceIntances = []
        if len(diffFileMap) == 0:
            if not args.force_rolling:
                print "There is no local diff to commit for cluster[%s], skip it" % cluster
                print "But if you really want to trigger one more rolling, please add --force_rolling"
                return ""
            else:
                # print "There is no local diff to commit for cluster[%s], but user use --force_rolling" % cluster
                # print "So push all local file to trigger rolling"
                deltaJson = ConvertToDeltaJson(filemap1)
        try:
            serviceInstancesToDelete = []
            serviceInstancesToAdd = []
            for filePath in diffFileMap:
                elems = filePath.split("/")
                if filePath.startswith("services/os/") and not args.updateos:
                    print "Update 'services/os' may cause os clone, please confirm by adding '--updateos' option"
                    return ""
                if filePath.startswith("services/") and filePath.endswith("/") and len(elems) == 3:
                    serviceInstanceName = elems[1]
                    if diffFileMap[filePath]["type"] == "delete":
                        serviceInstancesToDelete.append(serviceInstanceName)
                    else:
                        serviceInstancesToAdd.append(serviceInstanceName)
            if len(serviceInstancesToDelete) > 0:
                # don't support delete service instance
                msg = "Can not delete service instance dir[%s]\n" % json.dumps(serviceInstancesToDelete)
                msg = msg + "Please use tianji_clt instance -d instance -c cluster to delete"
                raise Exception, msg

            for addedServiceInstance in serviceInstancesToAdd:
                serviceInstances = self.tjApi.ListServiceInstance(cluster, False)["ServiceInstances"]
                if addedServiceInstance not in serviceInstances:
                    self.tjApi.CreateServiceInstance(cluster, addedServiceInstance)
                    createdServiceIntances.append(addedServiceInstance)

            commitVersion = self.tjApi.SubmitCluster(project, cluster, version, deltaJson, args.comment)

        except VersionNotExist as err:
            latestFullJson, latestBaseVersion = self.tjApi.CheckoutCluster(project, cluster)
            latestFilemap = ConvertToFileMap(latestFullJson)
            latestDiffFileMap = DiffFileMap(filemap1, latestFilemap)
            if len(latestDiffFileMap) == 0:
                commitVersion = latestBaseVersion
                print "Base version updated and no diff with local, So Update base version to [%s]" % latestBaseVersion
            else:
                traceback.print_exc()
                raise err

        except Exception, e:
            for serviceInstance in createdServiceIntances:
                print "Delete Service Instance[%s]" % serviceInstance
                self.tjApi.DeleteServiceInstance(cluster, serviceInstance)
            traceback.print_exc()
            raise e

        # file map -> ./.tianji/origin_repo/
        metaDir = os.path.join(args.dir, gMetaDir)
        originRepoPath = os.path.join(metaDir, gOriginRepo)
        if os.path.exists(originRepoPath):
            cmd = "rm -rf %s" % originRepoPath
            ExecCmd(cmd)
        if not os.path.exists(originRepoPath):
            os.makedirs(originRepoPath)
        SaveToDisk(originRepoPath, filemap1, overwrite=False)
        # update local base version
        saveBaseVersion(args.dir, commitVersion)
        # print "push to server succeed, commitVersion[%s]" % commitVersion
        print commitVersion
        return commitVersion

    def Branch(self, args):
        returnValue = {}
        returnValue["code"] = 200

        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        if args.project:
            project = args.project
        else:
            project = self.tjApi.projectName
        if args.all:
            clusters = self.tjApi.ListClusters(project)
            returnValue["data"] = clusters
        elif args.delete:
            cluster = args.delete
            self.tjApi.DeleteCluster(project, cluster)
            returnValue["data"] = "delete cluster %s succeed" % cluster
        elif args.create:
            cluster = args.create
            quotaSet = {}
            createJson = {}
            createJson["Operation"] = "Create"
            if args.autogen:
                createJson["autoGenEncKey"] = True

            if args.file:
                with open(args.file) as f:
                    quotaSet = json.load(f)
                print quotaSet
                createJson["Quotas"] = quotaSet
            self.tjApi.CreateCluster(project, cluster, createJson)
            returnValue["data"] = "create cluster %s succeed" % cluster
        elif args.allocate:
            if not args.file:
                print "error: -f/--file must be given for machine allocate requirements"
                exit(1)
            with open(args.file) as f:
                machineRequire = json.load(f)

            ret = self.tjApi.AllocateClusterMachine(args.project, args.allocate, machineRequire)
        elif args.listug:
            ret = self.tjApi.ListClusterUsergroups(project, args.listug)
            returnValue["data"] = ret
        elif args.addug:
            if not args.ug:
                print "error: --ug must be given for usergroup"
                exit(1)
            ret = self.tjApi.CreateClusterUsergroup(project, args.addug, args.ug)
            returnValue["data"] = ret
        elif args.deleteug:
            if not args.ug:
                print "error: --ug must be given for usergroup"
                exit(1)
            ret = self.tjApi.DeleteClusterUsergroup(project, args.deleteug, args.ug)
            returnValue["data"] = ret
        elif args.reference:
            if not args.service_name:
                print "error: --service_name or -s must be given"
                exit(1)
            if not args.template_name:
                print "error: --template_name or -t must be given"
                exit(1)
            if not args.version:
                print "error: --version or -v must be given"
                exit(1)
            if args.remove:
                ret = self.tjApi.DeleteClusterTemplateReference(project, args.reference, args.service_name)
            else:
                ret = self.tjApi.SetClusterTemplateReference(project, args.reference, args.service_name,
                                                             args.template_name, args.version)
            returnValue["data"] = ret

        print json.dumps(returnValue, sort_keys=True, indent=2)

    def Instance(self, args):
        returnValue = {}
        returnValue["code"] = 200

        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        cluster = args.cluster
        if args.project:
            project = args.project
        else:
            project = self.tjApi.projectName
        if args.all:
            serviceInstances = self.tjApi.ListServiceInstance(cluster, False)["ServiceInstances"]
            returnValue["data"] = serviceInstances
        if args.add:
            ret = self.tjApi.CreateServiceInstance(cluster, args.add)
            returnValue["data"] = ret
        elif args.delete:
            ret = self.tjApi.DeleteServiceInstance(cluster, args.delete)
            returnValue["data"] = ret

        print json.dumps(returnValue, sort_keys=True, indent=2)

    def Project(self, args):
        returnValue = {}
        returnValue["code"] = 200

        if args.all:
            configPath = args.config
            self.tjApi = self.initTjApi(configPath)
            projects = self.tjApi.ListProjects()
            returnValue["data"] = projects
        elif args.create:
            configPath = args.config
            self.tjApi = self.initTjApi(configPath)
            project = args.create
            self.tjApi.CreateProject(project)
            returnValue["data"] = "create project %s succeed" % project
        elif args.delete:
            configPath = args.config
            self.tjApi = self.initTjApi(configPath)
            project = args.delete
            self.tjApi.DeleteProject(project)
            returnValue["data"] = "delete project %s succeed" % project
        print json.dumps(returnValue, sort_keys=True, indent=2)

    def Service(self, args):
        if args.all:
            configPath = args.config
            self.tjApi = self.initTjApi(configPath)
            services = self.tjApi.ListServices()
            print json.dumps(services, sort_keys=True, indent=2)
        elif args.create:
            configPath = args.config
            self.tjApi = self.initTjApi(configPath)
            service = args.create
            if args.sys:
                self.tjApi.CreateService(service, "true")
            else:
                self.tjApi.CreateService(service)
            print "create service %s succeed" % service

    def ServiceTemplate(self, args):
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        desc = ""
        if args.desc:
            desc = args.desc
        group = ""
        if args.group:
            group = args.group
        if args.all:
            if not args.service_name:
                print "error: --service_name must be given"
                exit(1)
            templates = self.tjApi.ListServiceTemplate(args.service_name)
            print json.dumps(templates, sort_keys=True, indent=2)
        elif args.delete:
            if not args.service_name:
                print "error: --service_name or -s must be given"
                exit(1)
            if not args.template_name:
                print "error: --template_name or -t must be given"
                exit(1)
            answer = raw_input("Do you really want to delete this service[%s]'s template[%s]?[yes/no]: " % (
            args.service_name, args.template_name))
            if answer != "yes":
                exit(0)
            self.tjApi.DeleteServiceTemplate(args.service_name, args.template_name)
            print "delete service template succeed"
        elif args.create:
            if not args.service_name:
                print "error: --service_name or -s must be given"
                exit(1)
            if not args.template_name:
                print "error: --template_name or -t must be given"
                exit(1)
            if not args.dir:
                print "error: --dir must be given"
                exit(1)
            if not args.comment:
                print "error: --comment or -m must be given"
                exit(1)
            filemap = LoadFromDisk(args.dir, gSkipList)
            contentJson = ConvertToDeltaJson(filemap)
            self.tjApi.CreateServiceTemplate(args.service_name, args.template_name, contentJson, args.comment, desc,
                                             group)
            print "create service template succeed"
        elif args.update:
            if not args.service_name:
                print "error: --service_name or -s must be given"
                exit(1)
            if not args.template_name:
                print "error: --template_name or -t must be given"
                exit(1)
            if not args.dir:
                print "error: --dir must be given"
                exit(1)
            if not args.comment:
                print "error: --comment or -m must be given"
                exit(1)
            filemap = LoadFromDisk(args.dir, gSkipList)
            contentJson = ConvertToDeltaJson(filemap)
            self.tjApi.UpdateServiceTemplate(args.service_name, args.template_name, contentJson, args.comment, desc,
                                             group)
            print "Update service template succeed"
        elif args.get:
            if not args.service_name:
                print "error: --service_name or -s must be given"
                exit(1)
            if not args.template_name:
                print "error: --template_name or -t must be given"
                exit(1)
            if not args.dir:
                print "error: --dir must be given"
                exit(1)
            if os.path.exists(args.dir):
                raise Exception, "%s already exists." % args.dir
            version = ""
            if args.version:
                version = args.version

            contentJson = self.tjApi.GetServiceTemplate(args.service_name, args.template_name, version, args.plain)
            filemap = ConvertToFileMap(contentJson)
            if not os.path.exists(args.dir):
                os.makedirs(args.dir)
            overwrite = True
            SaveToDisk(args.dir, filemap, overwrite)
            print "Get service template succeed, save into dir[%s]" % args.dir
        elif args.log:
            if not args.service_name:
                print "error: --service_name or -s must be given"
                exit(1)
            if not args.template_name:
                print "error: --template_name or -t must be given"
                exit(1)
            num = ""
            if args.num:
                num = args.num

            logs = self.tjApi.GetServiceTemplateLog(args.service_name, args.template_name, num)
            print json.dumps(logs, indent=2)
        elif args.reference:
            if not args.service_name:
                print "error: --service_name or -s must be given"
                exit(1)
            if not args.template_name:
                print "error: --template_name or -t must be given"
                exit(1)
            if not args.cluster_name:
                print "error: --cluster_name or -c must be given"
                exit(1)
            if not args.version:
                print "error: --version or -v must be given"
                exit(1)
            if args.remove:
                ret = self.tjApi.DeleteServiceTemplateReference(args.cluster_name, args.service_name)
                print "Delete service template reference succeed, new cluster version: " + ret
            else:
                ret = self.tjApi.SetServiceTemplateReference(args.cluster_name, args.service_name, args.template_name,
                                                             args.version)
                print "Set service template reference succeed, new cluster version: " + ret
        elif args.show_reference:
            if not args.service_name:
                print "error: --service_name or -s must be given"
                exit(1)
            if not args.template_name:
                print "error: --template_name or -t must be given"
                exit(1)
            ret = self.tjApi.GetClusterForServiceTemplate(args.service_name, args.template_name)
            print json.dumps(ret, indent=2)

    def ServiceValidator(self, args):
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        desc = ""
        if args.desc:
            desc = args.desc
        if args.all:
            if not args.service_name:
                print "error: --service_name must be given"
                exit(1)
            validators = self.tjApi.ListServiceValidator(args.service_name)
            print json.dumps(validators, sort_keys=True, indent=2)
        elif args.delete:
            if not args.service_name:
                print "error: --service_name or -s must be given"
                exit(1)
            if not args.validator_name:
                print "error: --validator_name or -sh must be given"
                exit(1)
            answer = raw_input("Do you really want to delete this service[%s]'s validator[%s]?[yes/no]: " % (
            args.service_name, args.validator_name))
            if answer != "yes":
                exit(0)
            self.tjApi.DeleteServiceValidator(args.service_name, args.validator_name)
            print "delete service validator succeed"
        elif args.create:
            if not args.service_name:
                print "error: --service_name or -s must be given"
                exit(1)
            if not args.validator_name:
                print "error: --validator_name or -sh must be given"
                exit(1)
            if not args.file:
                print "error: -f must be given"
                exit(1)
            if not args.comment:
                print "error: --comment or -m must be given"
                exit(1)
            with open(args.file) as f:
                fileContent = f.read()
            self.tjApi.CreateServiceValidator(args.service_name, args.validator_name, fileContent, args.comment, desc)
            print "create service validator succeed"
        elif args.update:
            if not args.service_name:
                print "error: --service_name or -s must be given"
                exit(1)
            if not args.validator_name:
                print "error: --validator_name or -sh must be given"
                exit(1)
            if not args.file:
                print "error: -f must be given"
                exit(1)
            if not args.comment:
                print "error: --comment or -m must be given"
                exit(1)
            with open(args.file) as f:
                fileContent = f.read()
            self.tjApi.UpdateServiceValidator(args.service_name, args.validator_name, fileContent, args.comment, desc)
            print "Update service validator succeed"
        elif args.get:
            if not args.service_name:
                print "error: --service_name or -s must be given"
                exit(1)
            if not args.validator_name:
                print "error: --validator_name or -sh must be given"
                exit(1)
            if not args.output:
                print "error: -o must be given"
                exit(1)
            version = ""
            if args.version:
                version = args.version

            content = self.tjApi.GetServiceValidator(args.service_name, args.validator_name, version)
            f = open(args.output, "w")
            f.write(content)
            f.close()
            print "Get service validator succeed, save into file[%s]" % args.output
        elif args.log:
            if not args.service_name:
                print "error: --service_name or -s must be given"
                exit(1)
            if not args.validator_name:
                print "error: --validator_name or -t must be given"
                exit(1)
            num = ""
            if args.num:
                num = args.num

            logs = self.tjApi.GetServiceValidatorLog(args.service_name, args.validator_name, num)
            print json.dumps(logs, indent=2)

    def ServiceSchema(self, args):
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        desc = ""
        if args.desc:
            desc = args.desc
        if args.all:
            if not args.service_name:
                print "error: --service_name must be given"
                exit(1)
            schemas = self.tjApi.ListServiceSchema(args.service_name)
            print json.dumps(schemas, sort_keys=True, indent=2)
        elif args.delete:
            if not args.service_name:
                print "error: --service_name or -s must be given"
                exit(1)
            if not args.schema_name:
                print "error: --schema_name or -sh must be given"
                exit(1)
            answer = raw_input("Do you really want to delete this service[%s]'s schema[%s]?[yes/no]: " % (
            args.service_name, args.schema_name))
            if answer != "yes":
                exit(0)
            self.tjApi.DeleteServiceSchema(args.service_name, args.schema_name)
            print "delete service schema succeed"
        elif args.create:
            if not args.service_name:
                print "error: --service_name or -s must be given"
                exit(1)
            if not args.schema_name:
                print "error: --schema_name or -sh must be given"
                exit(1)
            if not args.file:
                print "error: -f must be given"
                exit(1)
            if not args.comment:
                print "error: --comment or -m must be given"
                exit(1)
            with open(args.file) as f:
                fileContent = yaml.load(f)
            self.tjApi.CreateServiceSchema(args.service_name, args.schema_name, fileContent, args.comment, desc)
            print "create service schema succeed"
        elif args.update:
            if not args.service_name:
                print "error: --service_name or -s must be given"
                exit(1)
            if not args.schema_name:
                print "error: --schema_name or -sh must be given"
                exit(1)
            if not args.file:
                print "error: -f must be given"
                exit(1)
            if not args.comment:
                print "error: --comment or -m must be given"
                exit(1)
            with open(args.file) as f:
                fileContent = yaml.load(f)
            self.tjApi.UpdateServiceSchema(args.service_name, args.schema_name, fileContent, args.comment, desc)
            print "Update service schema succeed"
        elif args.get:
            if not args.service_name:
                print "error: --service_name or -s must be given"
                exit(1)
            if not args.schema_name:
                print "error: --schema_name or -sh must be given"
                exit(1)
            if not args.output:
                print "error: -o must be given"
                exit(1)
            version = ""
            if args.version:
                version = args.version

            content = self.tjApi.GetServiceSchema(args.service_name, args.schema_name, version)
            f = open(args.output, "w")
            f.write(content)
            f.close()
            print "Get service schema succeed, save into file[%s]" % args.output
        elif args.log:
            if not args.service_name:
                print "error: --service_name or -s must be given"
                exit(1)
            if not args.schema_name:
                print "error: --schema_name or -t must be given"
                exit(1)
            num = ""
            if args.num:
                num = args.num

            logs = self.tjApi.GetServiceSchemaLog(args.service_name, args.schema_name, num)
            print json.dumps(logs, indent=2)

    def Check(self, args):
        returnValue = {}
        returnValue["code"] = 200
        returnValue["data"] = {}
        if args.all:
            configPath = args.config
            self.tjApi = self.initTjApi(configPath)
            cluster = args.cluster
            if args.all.lower() == "head":
                version = self.tjApi.GetClusterVersion(cluster)
            else:
                version = args.all
            ret = self.tjApi.ListRolling(cluster, version)
            print "Job: %-68s State: %s" % (cluster + "-" + version, ret[0]["state"])
            allSR = self.tjApi.ListMachineSRByCluster(cluster)
            for sr in allSR:
                retSR = self.tjApi.GetMachineSR(sr["hostname"], sr["serverrole"])
                print "ServerRole: %-30s HostName: %-20s ActualVersion: %-45s State: %s" % (
                sr["serverrole"], sr["hostname"], retSR["actual_version"], retSR["state"])
                # print json.dumps(ret, sort_keys=True, indent=2)
        elif args.rolling:
            configPath = args.config
            self.tjApi = self.initTjApi(configPath)
            cluster = args.cluster
            if args.rolling.lower() == "head":
                version = self.tjApi.GetClusterVersion(cluster)
            else:
                version = args.rolling
            returnValue["data"] = self.tjApi.ListRolling(cluster, version)
        elif args.allsr:
            configPath = args.config
            self.tjApi = self.initTjApi(configPath)
            cluster = args.cluster
            returnValue["data"] = self.tjApi.ListMachineSRByCluster(cluster)
        elif args.serverrole:
            configPath = args.config
            self.tjApi = self.initTjApi(configPath)
            returnValue["data"] = self.tjApi.GetMachineSR(args.serverrole[0], args.serverrole[1])
        print json.dumps(returnValue, sort_keys=True, indent=2)

    # get machine attributes
    def MachineInfo(self, args):
        returnValue = {}
        returnValue["code"] = 200
        returnValue["data"] = {}
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        if not args.attr:
            attr = "cluster,ip,state,actionname,actionstatus"
        else:
            attr = args.attr

        hosts = []
        if args.machines:
            hosts = args.machines.split(",")
        else:
            f = open(args.file, 'r')
            hosts = f.readlines()
            f.close()

        if args.local:
            type = "local"
            if args.batch:
                newhosts = []
                for hostname in hosts:
                    newhosts.append(hostname.strip())
                machinesInfo = self.tjApi.BatchGetMachineInfo(newhosts, attr, "local")
                for machineInfo in machinesInfo:
                    if "Hostname" in machineInfo["Attributes"]:
                        hostname = machineInfo["Attributes"]["Hostname"]
                    else:
                        hostname = machineInfo["Attributes"]["hostname"]
                    returnValue["data"][hostname] = machineInfo["Attributes"]
            else:
                for hostname in hosts:
                    hostname = hostname.strip()
                    machineInfo = self.tjApi.GetMachineInfo(hostname, attr, "local")
                    returnValue["data"][hostname] = machineInfo["Attributes"]

        else:
            if args.batch:
                newhosts = []
                for hostname in hosts:
                    newhosts.append(hostname.strip())
                machinesInfo = self.tjApi.BatchGetMachineInfo(newhosts, attr, "")
                for machineInfo in machinesInfo:
                    if "Hostname" in machineInfo:
                        hostname = machineInfo["Hostname"]
                    else:
                        hostname = machineInfo["hostname"]
                    returnValue["data"][hostname] = {}
                    for k in machineInfo.keys():
                        if k == "hostname":
                            continue
                        returnValue["data"][hostname][k] = machineInfo[k]

            else:
                for hostname in hosts:
                    hostname = hostname.strip()
                    returnValue["data"][hostname] = {}
                    machineInfo = self.tjApi.GetMachineInfo(hostname, attr, "")
                    for k in machineInfo.keys():
                        if k == "hostname":
                            continue
                        returnValue["data"][hostname][k] = machineInfo[k]

        print json.dumps(returnValue, sort_keys=True, indent=2)

    # list all rolling jobs
    def Rollings(self, args):
        returnValue = {}
        returnValue["code"] = 200
        returnValue["data"] = []
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        cluster = args.cluster
        if args.list:
            returnValue["data"] = self.GetAllRollings(cluster)

        if args.show:
            if args.show.lower() == "head":
                args.show = self.tjApi.GetClusterVersion(cluster)
            returnValue["data"] = self.tjApi.ListRolling(cluster, args.show)
            # returnValue["data"] = self.tjApi.GetRolling(cluster, args.show, args.subversion)
        elif args.terminate:
            if args.terminate.lower() == "head":
                args.terminate = self.tjApi.GetClusterVersion(cluster)
            operationJson = {}
            operationJson["operation"] = "terminate"
            returnValue["data"] = self.tjApi.OperateRolling(cluster, args.terminate, operationJson, args.subversion)
        elif args.pause:
            if args.pause.lower() == "head":
                args.pause = self.tjApi.GetClusterVersion(cluster)
            operationJson = {}
            operationJson["operation"] = "pause"
            returnValue["data"] = self.tjApi.OperateRolling(cluster, args.pause, operationJson, args.subversion)
        elif args.resume:
            if args.resume.lower() == "head":
                args.resume = self.tjApi.GetClusterVersion(cluster)
            operationJson = {}
            operationJson["operation"] = "resume"
            returnValue["data"] = self.tjApi.OperateRolling(cluster, args.resume, operationJson, args.subversion)
        elif args.retry:
            if args.retry.lower() == "head":
                args.retry = self.tjApi.GetClusterVersion(cluster)
            operationJson = {}
            operationJson["operation"] = "retry"
            returnValue["data"] = self.tjApi.OperateRolling(cluster, args.retry, operationJson, args.subversion)

        print json.dumps(returnValue, sort_keys=True, indent=2)

    def GetAllRollings(self, cluster):
        rollings = []
        ret = self.tjApi.ListAllRolling(cluster)
        rollingOrder = {}
        for rolling in ret:
            if "start_time" not in rolling:
                continue
            rollingOrder[rolling["start_time"]] = {}
            rollingOrder[rolling["start_time"]]["id"] = rolling["commit_id"]
            rollingOrder[rolling["start_time"]]["state"] = rolling["state"]

        keys = rollingOrder.keys()
        keys.sort()
        currentFlag = 1
        currentRolling = {}

        for k in keys:
            rollingInfo = {}
            rollingInfo["id"] = str(rollingOrder[k]["id"])
            rollingInfo["state"] = str(rollingOrder[k]["state"])
            rollingInfo["starttime"] = str(k)
            if str(k) == "" and rollingInfo["state"].lower() == "running":
                rollingInfo["current"] = 1
                currentRolling = rollingInfo
                continue
            if currentFlag == 1 and (
                    rollingInfo["state"].lower() == "running" or rollingInfo["state"].lower() == "waiting"):
                rollingInfo["current"] = 1
                currentFlag = 0
            else:
                rollingInfo["current"] = 0
            rollings.append(rollingInfo)

        if len(currentRolling) > 0:
            rollings.append(currentRolling)
        return rollings

    # check latest rolling job
    def RollingStatus(self, args):
        returnValue = {}
        returnValue["code"] = 200
        returnValue["data"] = {}
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        cluster = args.cluster
        if not args.version or args.version.lower() == "head":
            version = self.tjApi.GetClusterVersion(cluster)
        else:
            version = args.version

        ret = self.tjApi.ListRolling(cluster, version)
        rollings = self.GetAllRollings(cluster)
        rollingExist = 0
        for roll in rollings:
            if roll["id"] == version and roll["current"] == 1:
                # version is current rolling id
                rollingExist = 1
                break
            elif roll["id"] == version and roll["current"] == 0:
                if version == rollings[len(rollings) - 1]["id"]:
                    # version is the last rolling id
                    rollingExist = 1
                    break
                else:
                    # version is an old rolling id
                    returnValue["data"]["JobState"] = str(ret[0]["state"]).lower()
                    returnValue["data"]["Services"] = {}
                    returnValue["data"][
                        "Message"] = "New rolling has been submitted, can not check service status, use listrollings to check"
                    print json.dumps(returnValue, sort_keys=True, indent=2)
                    return
            elif roll["id"] != version and roll["current"] == 1:
                # an old rolling is still blocking
                returnValue["data"]["JobState"] = "Blocking"
                returnValue["data"]["Services"] = {}
                returnValue["data"]["Message"] = "Wait for rolling job: " + roll["id"]
                print json.dumps(returnValue, sort_keys=True, indent=2)
                return

        if len(ret) == 0 or rollingExist == 0:
            returnValue["data"]["JobState"] = "None"
            returnValue["data"]["Services"] = {}
            returnValue["data"]["Message"] = "No such job"
            print json.dumps(returnValue, sort_keys=True, indent=2)
            return

        rollingStr = ret[0]["status"]
        rollingInfo = yaml.load(rollingStr)
        serviceInfo = {}

        #      "status" : "{\"TaskStatus\":[{\"SRI\":{\"tianji.TianjiClient#\":{\"ApproveRate\":\"1\",\"FailureRate\":\"0\",\"ScaleUnitStatus\":\"{}\",\"Status\":\"succeeded\",\"SuccessRate\":\"1\"}},\"Status\":\"succeeded\",\"RollingInfo\":{\"StartTime\":\"\\\"2015-12-10T13:03:11.109531767+08:00\\\"\",\"EndTime\":\"\\\"2015-12-10T13:03:23.735674837+08:00\\\"\",\"State\":\"succeeded\"},\"RollbackInfo\":{\"StartTime\":\"\",\"EndTime\":\"\",\"State\":\"\"}}],\"Version\":\"0.4\"}"

        if "RUStatusMap" not in rollingInfo or not rollingInfo["RUStatusMap"]:
            if "state" in ret[0]:
                returnValue["data"]["JobState"] = ret[0]["state"]
            else:
                returnValue["data"]["JobState"] = "Waiting"
            returnValue["data"]["Services"] = {}
            if "error_message" in ret[0]:
                returnValue["data"]["Message"] = ret[0]["error_message"]
            else:
                returnValue["data"]["Message"] = "Wait for DS to roll"
            print json.dumps(returnValue, sort_keys=True, indent=2)
            return

        if rollingInfo["RUStatusMap"] and len(rollingInfo["RUStatusMap"]) > 0:
            for serverRoleName in rollingInfo["RUStatusMap"]:
                serverRoleName = str(serverRoleName)
                serviceName = serverRoleName.split(".")[0]

                if serviceName not in serviceInfo:
                    serviceInfo[serviceName] = {}
                srHosts = self.tjApi.ListMachineSRByCluster(cluster, serverRoleName)
                if serverRoleName not in serviceInfo[serviceName]:
                    serviceInfo[serviceName][serverRoleName] = []
                for srHost in srHosts:
                    serviceInfo[serviceName][serverRoleName].append(str(srHost["hostname"]))

        returnValue["data"]["Services"] = self.GetServiceRollingStatus(serviceInfo, version)

        allGood = 1
        for serviceName in returnValue["data"]["Services"]:
            if returnValue["data"]["Services"][serviceName]["State"] != "Good":
                allGood = 0
                break

        jobMessage = ret[0]["error_message"]
        # parse job state
        jobState = str(ret[0]["state"]).lower()
        retJobState = "Running"
        if jobState == "waiting":
            buildStatus = self.tjApi.GetBuildInfo(cluster, version)
            if buildStatus["build_status"].lower() == "done":
                retJobState = "Ready"
                jobMessage = "Wait for MM to set serverrole task (lingyun.ly)"
            else:
                retJobState = "Building"
                jobMessage = buildStatus["msg"]
        elif jobState == "failed":
            retJobState = "Failed"
            buildStatus = self.tjApi.GetBuildInfo(cluster, version)
            if buildStatus["build_status"].lower() != "done":
                jobMessage = buildStatus["msg"]
        elif jobState == "terminated":
            retJobState = "Terminated"
        elif jobState == "paused":
            retJobState = "Paused"
        elif jobState == "succeeded" and allGood == 1:
            retJobState = "Succeeded"
        returnValue["data"]["JobState"] = retJobState
        returnValue["data"]["Message"] = jobMessage

        if args.json:
            print json.dumps(returnValue)
        else:
            print "\nJob: %s-%s, State: %s" % (cluster, version, returnValue["data"]["JobState"])
            if returnValue["data"]["Message"] != "":
                print "Message: %s" % (returnValue["data"]["Message"])

            for serviceName in returnValue["data"]["Services"]:
                print "\n    Service: %s, State: %s" % (
                serviceName, returnValue["data"]["Services"][serviceName]["State"])
                if returnValue["data"]["Services"][serviceName]["Message"] != "":
                    print "    Message: %s" % (returnValue["data"]["Services"][serviceName]["Message"])

                for serverRoleName in returnValue["data"]["Services"][serviceName]["Serverroles"]:
                    for hostName in returnValue["data"]["Services"][serviceName]["Serverroles"][serverRoleName]:
                        print "        %-40s %-20s %-15s %s" % (serverRoleName, hostName,
                                                                returnValue["data"]["Services"][serviceName][
                                                                    "Serverroles"][serverRoleName][hostName]["State"],
                                                                returnValue["data"]["Services"][serviceName][
                                                                    "Serverroles"][serverRoleName][hostName]["Message"])

    # get all service rolling status
    def GetServiceRollingStatus(self, serviceInfo, version):
        returnValue = {}
        for serviceName in serviceInfo:
            returnValue[serviceName] = {}
            returnValue[serviceName]["Serverroles"] = {}
            returnValue[serviceName]["Message"] = ""
            serviceState = ""
            waitingSRIList = []
            postCheckFailed = 0
            goodFlag = 1

            for serverRoleName in serviceInfo[serviceName]:
                returnValue[serviceName]["Serverroles"][serverRoleName] = {}
                for hostName in serviceInfo[serviceName][serverRoleName]:
                    sriMessage = ""
                    srInfo = self.tjApi.GetMachineSR(hostName, serverRoleName)
                    srState = "Good"
                    if "kept_versions" not in srInfo or version not in srInfo["kept_versions"]:
                        srState = "Upgrading"
                    elif "available_versions" not in srInfo or version not in srInfo["available_versions"]:
                        srState = "Downloading"
                        srMonitorInfo = self.tjApi.ListMachineSRMonitor(hostName, serverRoleName)
                        for srM in srMonitorInfo:
                            if srM["monitor"].lower() == "downloaderfile":
                                sriMessage = srM["description"] + " (tongchun.xiaotc)"
                    elif "expected_version" not in srInfo or version != srInfo["expected_version"]:
                        srState = "Ready"
                    elif "actual_version" not in srInfo or version != srInfo["actual_version"]:
                        srState = "Starting"
                        if "error_message" in srInfo:
                            sriMessage = srInfo["error_message"]
                            if "sysenv state not done" in sriMessage.lower():
                                sriMessage = "Depending service not ready"
                            elif "config state not done" in sriMessage.lower():
                                sriMessage = "Config rendering..."
                            elif "no legal app" in sriMessage.lower():
                                sriMessage = ""
                    elif "state" not in srInfo or srInfo["state"].lower() == "shutdown":
                        srState = "Starting"
                    elif srInfo["state"].lower() == "error":
                        srState = "Error"
                        srMonitorInfo = self.tjApi.ListMachineSRMonitor(hostName, serverRoleName)
                        for srM in srMonitorInfo:
                            if srM["level"].lower() == "error" or srM["level"].lower() == "fatal":
                                sriMessage = sriMessage + "Monitor " + srM["monitor"] + " " + srM["level"]
                                if srM["description"] != "":
                                    sriMessage = sriMessage + ", " + srM["description"] + "; "
                    elif srInfo["state"].lower() == "probation":
                        srState = "Probation"
                        sriMessage = "Need to wait"
                        srMonitorInfo = self.tjApi.ListMachineSRMonitor(hostName, serverRoleName)
                        for srM in srMonitorInfo:
                            if srM["monitor"].lower() == "postcheck_monitor":
                                sriMessage = srM["description"]
                                if srM["level"].lower() == "fatal":
                                    srState = "PostCheckFailed"
                                else:
                                    srState = "PostChecking"

                    if srState == "PostCheckFailed":
                        postCheckFailed = 1
                    elif srState != "Good":
                        goodFlag = 0
                        waitingSRIList.append("\"" + serverRoleName + "\"")

                    returnValue[serviceName]["Serverroles"][serverRoleName][hostName] = {}
                    returnValue[serviceName]["Serverroles"][serverRoleName][hostName]["State"] = srState
                    returnValue[serviceName]["Serverroles"][serverRoleName][hostName]["Message"] = sriMessage

            if postCheckFailed == 1:
                serviceState = "PostCheckFailed"
            elif goodFlag == 1:
                serviceState = "Good"
            else:
                serviceState = "Starting"
            returnValue[serviceName]["State"] = serviceState
            if len(waitingSRIList) > 0:
                returnValue[serviceName]["Message"] = "{\"no good SRI\":[" + ",".join(waitingSRIList) + "]}"
        return returnValue

    def MachineStatus(self, args):
        returnValue = {}
        returnValue["code"] = 200
        returnValue["data"] = {}
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        cluster = args.cluster

        returnValue["data"]["Machines"] = self.GetMachineStatus(cluster)
        if args.json:
            print json.dumps(returnValue)
        else:
            print json.dumps(returnValue, sort_keys=True, indent=2)

    def GetMachineStatus(self, clusterName):
        machines = self.tjApi.ListMachineByCluster(clusterName)
        machineStatus = {}
        if len(machines) == 0:
            return machineStatus

        attr = "state"
        machinesInfo = self.tjApi.BatchGetMachineInfo(machines, attr)
        for info in machinesInfo:
            if info.has_key("hostname") and info.has_key("state"):
                machineStatus[info["hostname"]] = {}
                machineStatus[info["hostname"]]["State"] = info["state"]
        return machineStatus

    # check all services
    def ServiceStatus(self, args):
        returnValue = {}
        returnValue["code"] = 200
        returnValue["data"] = {}
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        cluster = args.cluster

        errorServices = []
        if args.services:
            serviceInstances = self.tjApi.ListServiceInstance(cluster, False)["ServiceInstances"]
            serviceList = args.services.split(",")
            for serviceName in serviceList:
                if serviceName not in serviceInstances:
                    errorServices.append(serviceName)
        if len(errorServices) > 0:
            returnValue["code"] = 400
            returnValue["data"] = {}
            returnValue["message"] = "Service can not found: " + ",".join(errorServices)
            print json.dumps(returnValue, sort_keys=True, indent=2)
            return

        serviceInfo = {}
        allSR = self.tjApi.ListMachineSRByCluster(cluster)
        for sr in allSR:
            serverRoleName = str(sr["serverrole"])
            hostName = str(sr["hostname"])
            serviceName = serverRoleName.split(".")[0]

            if args.services:
                serviceList = args.services.split(",")
                if serviceName not in serviceList:
                    continue

            if serviceName not in serviceInfo:
                serviceInfo[serviceName] = {}
            if serverRoleName not in serviceInfo[serviceName]:
                serviceInfo[serviceName][serverRoleName] = []
            serviceInfo[serviceName][serverRoleName].append(hostName)

        returnValue["data"]["Services"] = self.GetServiceStatus(serviceInfo)

        # add machine state in app status
        machineStatus = self.GetMachineStatus(cluster)
        for serviceName in returnValue["data"]["Services"]:
            for serverRoleName in returnValue["data"]["Services"][serviceName]["Serverroles"]:
                for hostName in returnValue["data"]["Services"][serviceName]["Serverroles"][serverRoleName]:
                    if machineStatus.has_key(hostName) and machineStatus[hostName].has_key("State"):
                        returnValue["data"]["Services"][serviceName]["Serverroles"][serverRoleName][hostName][
                            "MachineState"] = machineStatus[hostName]["State"]
                    else:
                        returnValue["data"]["Services"][serviceName]["Serverroles"][serverRoleName][hostName][
                            "MachineState"] = "UNKNOW"

        if args.json:
            print json.dumps(returnValue)
        else:
            print json.dumps(returnValue, sort_keys=True, indent=2)

    # get all service current status
    def GetServiceStatus(self, serviceInfo):
        returnValue = {}
        for serviceName in serviceInfo:
            returnValue[serviceName] = {}
            returnValue[serviceName]["Serverroles"] = {}
            returnValue[serviceName]["Message"] = ""
            serviceState = ""
            waitingSRIList = []
            internalError = 0
            postCheckFailed = 0
            goodFlag = 1

            for serverRoleName in serviceInfo[serviceName]:
                returnValue[serviceName]["Serverroles"][serverRoleName] = {}
                for hostName in serviceInfo[serviceName][serverRoleName]:
                    sriMessage = ""
                    srInfo = self.tjApi.GetMachineSR(hostName, serverRoleName)
                    srState = "Good"
                    containerId = ""
                    containerName = ""
                    containerHost = ""
                    containerState = "Unknown"
                    postCheckDetails = ""
                    expectVersion = ""
                    actualVersion = ""
                    pullImageFailMessage = ""
                    if "expected_version" in srInfo:
                        expectVersion = srInfo["expected_version"]
                    if "actual_version" in srInfo:
                        actualVersion = srInfo["actual_version"]

                    if "available_versions" not in srInfo or (
                            expectVersion != "" and expectVersion not in srInfo["available_versions"]):
                        srState = "Upgrading"
                        sriMessage = "Downloading"
                        srMonitorInfo = self.tjApi.ListMachineSRMonitor(hostName, serverRoleName)
                        for srM in srMonitorInfo:
                            if srM["monitor"].lower() == "downloaderfile":
                                msg = srM["description"]
                                if msg.find("need not to") != -1:
                                    pullImageFailMessage = ""
                                elif msg.find("HTTP") == -1 and msg.find("not found") != -1:
                                    errBeginIdx = msg.find("Error:")
                                    pullImageFailMessage = msg[errBeginIdx:-3]
                                else:
                                    pullImageFailMessage = msg
                    elif expectVersion == "" or actualVersion == "" or expectVersion != actualVersion:
                        srState = "Upgrading"
                        if "error_message" in srInfo:
                            sriMessage = srInfo["error_message"]
                            if "sysenv state not done" in sriMessage.lower():
                                sriMessage = "Depending service not ready"
                            elif "config state not done" in sriMessage.lower():
                                sriMessage = "Config rendering..."
                            elif "no legal app" in sriMessage.lower():
                                sriMessage = ""
                    elif "state" not in srInfo or srInfo["state"].lower() == "shutdown":
                        srState = "Upgrading"
                    else:
                        # check monitor first, then check serverrole state
                        srMonitorInfo = self.tjApi.ListMachineSRMonitor(hostName, serverRoleName)
                        for srM in srMonitorInfo:
                            if srM["monitor"].lower() == "postcheck_monitor":
                                # check postcheck monitor first
                                sriMessage = srM["description"]
                                if sriMessage.find("Container") != -1:
                                    if sriMessage.find("dead") != -1:
                                        containerState = "Dead"
                                    if sriMessage.find("alive") != -1:
                                        containerState = "Alive"
                                    sriMessageSplits = sriMessage.split("|")
                                    containerId = sriMessageSplits[0].split(":")[1]
                                    containerName = sriMessageSplits[1].split(":")[1]
                                    containerHost = sriMessageSplits[2].split(":")[1]
                                    # if len(sriMessageSplits) == 2:
                                    postCheckDetails = sriMessageSplits[-1]
                                    sriMessage = ""
                                if srM["level"].lower() == "fatal":
                                    srState = "PostCheckFailed"
                                    break
                                elif srM["level"].lower() == "error":
                                    srState = "PostChecking"
                                    break

                        # if serverrole is good, do not check other monitors, may have been expired
                        # if serverrole is probation, will return good as init value
                        if srInfo["state"].lower() == "good":
                            srState = "Good"
                        elif srState != "PostCheckFailed" and srState != "PostChecking":
                            for srM in srMonitorInfo:
                                if srM["monitor"].lower() == "postcheck_monitor":
                                    continue
                                if srM["monitor"].find("prob_to_error") != -1:
                                    continue

                                    # no monitor error and expectVersion == actualVerion, will return Good as default value

                    if srState == "PostCheckFailed":
                        postCheckFailed = 1
                        goodFlag = 0
                    elif srState != "Good":
                        goodFlag = 0
                        waitingSRIList.append("\"" + serverRoleName + "\"")

                    returnValue[serviceName]["Serverroles"][serverRoleName][hostName] = {}
                    returnValue[serviceName]["Serverroles"][serverRoleName][hostName]["State"] = srState
                    returnValue[serviceName]["Serverroles"][serverRoleName][hostName]["ContainerState"] = containerState
                    returnValue[serviceName]["Serverroles"][serverRoleName][hostName]["ContainerId"] = containerId
                    returnValue[serviceName]["Serverroles"][serverRoleName][hostName]["ContainerName"] = containerName
                    returnValue[serviceName]["Serverroles"][serverRoleName][hostName]["ContainerHost"] = containerHost
                    returnValue[serviceName]["Serverroles"][serverRoleName][hostName][
                        "PostCheckDetails"] = postCheckDetails
                    returnValue[serviceName]["Serverroles"][serverRoleName][hostName]["ExpectVersion"] = expectVersion
                    returnValue[serviceName]["Serverroles"][serverRoleName][hostName]["ActualVersion"] = actualVersion
                    returnValue[serviceName]["Serverroles"][serverRoleName][hostName]["Message"] = sriMessage
                    if pullImageFailMessage != "":
                        returnValue[serviceName]["Serverroles"][serverRoleName][hostName][
                            "PullImageMessage"] = pullImageFailMessage

            if postCheckFailed == 1:
                serviceState = "PostCheckFailed"
            elif goodFlag == 1:
                serviceState = "Good"
            else:
                serviceState = "Upgrading"
            returnValue[serviceName]["State"] = serviceState
            if len(waitingSRIList) > 0:
                returnValue[serviceName]["Message"] = "{\"no good SRI\":[" + ",".join(waitingSRIList) + "]}"
        return returnValue

    def ClearLocal(self, args):
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        answer = raw_input("Do you really want to clear seed on localhost?[yes/no]: ")
        if answer != "yes":
            exit(0)

        if args.bootconfig:
            with open(args.bootconfig) as f:
                config = yaml.load(f)
            installDir = config["InstallDir"]
        else:
            installDir = "/home/admin/install_seed_cluster"

        try:
            cmd = 'sudo rpm -e tianji-client'
            ExecCmd(cmd)
        except Exception, e:
            exstr = traceback.format_exc()
            #logger.info('Execut cmd[%s] err[%s]' % (cmd, exstr))

        try:
            cmd = 'ps aux |grep tianji_cluster_daemon |grep -v grep |awk "{print \$2}" | xargs -i kill -9 {}'
            ExecCmd(cmd)

            appBinNameList = ['mock_iclone', 'node', 'p2p_master', 'machine_monitor', 'healing_service',
                              'cstore_manager', 'availability_service', 'clone_service']
            appBinNameList.extend(
                ['tj_master_main', 'machine_manager', 'deploy_service', 'image_builder', 'tianji_proxy', 'api_server',
                 'docker'])
            for appBinName in appBinNameList:
                try:
                    cmd = 'ps aux |grep %s |grep -v grep | awk "{print \$2}" |xargs -i sudo kill -9 {}' % (appBinName)
                    #logger.info("killing app cmd: %s" % (cmd))
                    ExecCmd(cmd)
                except Exception, e:
                    #logger.info("exception occured when killing apps, cmd %s, e: %s" % (cmd, e))
                    exstr = traceback.format_exc()
                    #logger.info('StopBootstrapTJCluster occur errer when exec cmd[%s] err[%s]' % (cmd, exstr))

            cmd = 'sudo rm -rf /cloud/ /apsara/cloud %s' % (installDir)
            ExecCmd(cmd)
        except Exception, e:
            exstr = traceback.format_exc()
            #logger.info('Execut cmd[%s] err[%s]' % (cmd, exstr))

    def CreateDockerCluster(self, args):
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        if args.project:
            project = args.project
        else:
            project = self.tjApi.projectName
        cluster = args.cluster

        if self.tjApi.ClusterExists(project, cluster):
            self.tjApi.DeleteCluster(self.tjApi.projectName, cluster)
        createJson = {}
        createJson["Operation"] = "Create"
        version = self.tjApi.CreateCluster(self.tjApi.projectName, cluster, createJson)
        if not self.tjApi.ServiceExists("tianji"):
            self.tjApi.CreateService("tianji")
        self.tjApi.CreateServiceInstance(cluster, "tianji")

        role_conf_str = "Environments:\n  DockerDaemon#:\n    Client: %s\nMachineGroups:\n  Client: %s\nServerRoles:\n  DockerDaemon#: Client\n  TianjiClient#: Client" % (
        args.env, args.machines)
        version_conf_str = "Versions:\n  DockerDaemon#:\n    Applications: {'*': %s}\n  TianjiClient#:\n    Applications: {'*': %s}" % (
        args.build, args.build)
        availability_conf_str = "availability:\n  DockerDaemon#: {mode: 100%}\n  TianjiClient#: {mode: 100%}"

        deltaJson = {}
        deltaJson["cluster.conf"] = "Group: DefaultGroup\nMachines: %s\nName: %s\nRegion: DefaultRegion" % (
        args.machines, cluster)
        deltaJson["sys.conf"] = "os: {\"*\": os-default}\npart: {\"*\": ALdefault}"
        deltaJson["services"] = {}
        deltaJson["services"]["tianji"] = {}
        deltaJson["services"]["tianji"]["role.conf"] = role_conf_str
        deltaJson["services"]["tianji"]["version.conf"] = version_conf_str
        deltaJson["services"]["tianji"]["tianji"] = {}
        deltaJson["services"]["tianji"]["tianji"]["availability.conf"] = availability_conf_str

        commitVersion = self.tjApi.SubmitCluster(project, cluster, version, deltaJson,
                                                 "Create docker cluster: %s" % cluster)

        returnValue = {}
        returnValue["code"] = 200
        returnValue["data"] = {}
        returnValue["data"]["Version"] = commitVersion
        print json.dumps(returnValue)

    def IsAppHasDiff(self, app, newMd5Conf, currentMd5Conf):
        if not currentMd5Conf.has_key(app):
            return True
        else:
            for filename in newMd5Conf[app].keys():
                if filename not in currentMd5Conf[app]:
                    return True
                if newMd5Conf[app][filename] != currentMd5Conf[app][filename]:
                    return True
        return False

    def CreateDockerClusterFromTemplate(self, args):
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        if args.project:
            project = args.project
        else:
            project = self.tjApi.projectName
        cluster = args.cluster

        if self.tjApi.ClusterExists(project, cluster):
            self.tjApi.DeleteCluster(self.tjApi.projectName, cluster)
        createJson = {}
        createJson["Operation"] = "Create"
        version = self.tjApi.CreateCluster(self.tjApi.projectName, cluster, createJson)
        if not self.tjApi.ServiceExists("tianji"):
            self.tjApi.CreateService("tianji")
        self.tjApi.CreateServiceInstance(cluster, "tianji")

        contentJson = self.tjApi.GetServiceTemplate("tianji", "default_for_docker", "")
        role_conf_str = contentJson["role.conf"]
        role_conf = yaml.load(role_conf_str)
        role_conf["Environments"]["DockerDaemon#"]["Client"] = yaml.load(args.env)
        for machine in args.machines.split(","):
            role_conf["MachineGroups"]["Client"].append(machine.strip())
        version_conf_str = contentJson["version.conf"]
        availability_conf_str = contentJson["tianji"]["availability.conf"]

        deltaJson = {}
        deltaJson["cluster.conf"] = "Group: DefaultGroup\nMachines: [%s]\nName: %s\nRegion: DefaultRegion" % (
        args.machines, cluster)
        deltaJson["sys.conf"] = "os: {\"*\": os-default}\npart: {\"*\": ALdefault}"
        deltaJson["services"] = {}
        deltaJson["services"]["tianji"] = {}
        deltaJson["services"]["tianji"]["role.conf"] = yaml.dump(role_conf)
        deltaJson["services"]["tianji"]["version.conf"] = version_conf_str
        deltaJson["services"]["tianji"]["tianji"] = {}
        deltaJson["services"]["tianji"]["tianji"]["availability.conf"] = availability_conf_str

        commitVersion = self.tjApi.SubmitCluster(project, cluster, version, deltaJson,
                                                 "Create docker cluster: %s" % cluster)

        returnValue = {}
        returnValue["code"] = 200
        returnValue["data"] = {}
        returnValue["data"]["Version"] = commitVersion
        print json.dumps(returnValue)

    def UpgradeServerRoles(self, args, upgradeServerRole=False):
        self.DeployServices(args, True)

    def DeployServices(self, args, upgradeServerRole=False):
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        if self.tjApi.beimingAddress == "":
            returnValue = {}
            returnValue["code"] = 500
            returnValue["data"] = {}
            returnValue["message"] = "beiming address is not provided"
            print json.dumps(returnValue)
            return

        if args.project:
            project = args.project
        else:
            project = self.tjApi.projectName
        cluster = args.cluster

        version = self.tjApi.GetClusterVersion(cluster)
        deployedServiceInstances = self.tjApi.ListServiceInstance(cluster, False)["ServiceInstances"]
        deployedServices = ",".join(deployedServiceInstances)

        serviceList = args.services.split(",")
        curDir = os.path.dirname(os.path.abspath(__file__))
        tianjiDir = tempfile.mkdtemp()
        tmpOutputDir = tempfile.mkdtemp()

        try:
            cmd = "%s/docker_compose_2_tianji.py -i %s -d %s --beiming_host %s -o %s --ops_config_dir %s" % (
            curDir, args.dir, args.services, self.tjApi.beimingAddress, tmpOutputDir, tianjiDir)
            cmd = cmd + " > /dev/null"
            ExecCmd(cmd)

            if not args.force:
                for serviceInstanceName in serviceList:
                    versionConfPath = os.path.join("services", serviceInstanceName, "version.conf")
                    md5ConfPath = os.path.join("services", serviceInstanceName, "tianji", gMd5OfPkgFiles)

                    # new conf
                    newMd5Conf = yaml.load(open(os.path.join(tianjiDir, md5ConfPath)))
                    newVersionConf = yaml.load(open(os.path.join(tianjiDir, versionConfPath)))

                    # current conf
                    fullJson, baseVersion = self.tjApi.CheckoutCluster(project, cluster)
                    filemap = ConvertToFileMap(fullJson)
                    if (filemap.has_key(md5ConfPath) and filemap[md5ConfPath].has_key("content")) and (
                        filemap.has_key(versionConfPath) and filemap[versionConfPath].has_key("content")):
                        currentMd5Conf = yaml.load(filemap[md5ConfPath]["content"])
                        currentVersionConf = yaml.load(filemap[versionConfPath]["content"])

                        for sr in newVersionConf["Versions"].keys():
                            if sr not in currentVersionConf["Versions"]:
                                continue

                            currentBuildId = currentVersionConf["Versions"][sr]['Applications']["*"]

                            dockerAppName = sr.strip("#")
                            if self.IsAppHasDiff(dockerAppName, newMd5Conf, currentMd5Conf):
                                continue

                            dockerPostCheckAppName = sr.strip("#") + "_post_check_app"
                            if self.IsAppHasDiff(dockerPostCheckAppName, newMd5Conf, currentMd5Conf):
                                continue

                            newVersionConf["Versions"][sr]["Applications"]["*"] = currentBuildId

                        newVersionConfStr = yaml.dump(newVersionConf)
                        f = open(os.path.join(tianjiDir, versionConfPath), "w")
                        f.write(newVersionConfStr)
                        f.close()

            fileMap = LoadFromDisk(tianjiDir, gSkipList)
            serviceInstancesToAdd = []
            fileMapPush = {}
            for filePath in fileMap:
                if not filePath.startswith("services/") and not filePath.startswith("plan.conf"):
                    continue
                if filePath.startswith("services/"):
                    elems = filePath.split("/")
                    serviceInstanceName = elems[1]
                    if serviceInstanceName not in serviceList:
                        continue
                    else:
                        serviceInstancesToAdd.append(serviceInstanceName)
                        fileMapPush[filePath] = fileMap[filePath]
                else:
                    fileMapPush[filePath] = fileMap[filePath]

            deltaJson = ConvertToDeltaJson(fileMapPush)

            createdServiceIntances = []
            for addedServiceInstance in serviceInstancesToAdd:
                serviceInstances = self.tjApi.ListServiceInstance(cluster, False)["ServiceInstances"]
                if addedServiceInstance not in serviceInstances:
                    if not self.tjApi.ServiceExists(addedServiceInstance):
                        self.tjApi.CreateService(addedServiceInstance)
                    self.tjApi.CreateServiceInstance(cluster, addedServiceInstance)
                    createdServiceIntances.append(addedServiceInstance)

            commitVersion = self.tjApi.SubmitCluster(project, cluster, version, deltaJson,
                                                     "Add services: [%s]" % args.services)

            returnValue = {}
            returnValue["code"] = 200
            returnValue["data"] = {}
            returnValue["data"]["Version"] = commitVersion
            print json.dumps(returnValue)
        finally:
            shutil.rmtree(tianjiDir)
            shutil.rmtree(tmpOutputDir)

    def ReplaceMachine(self, args):
        clusterName = args.cluster
        configPath = args.config
        sourceMachine = args.source
        destMachine = args.dest

        self.tjApi = self.initTjApi(configPath)
        projectName = self.tjApi.GetClusterProject(clusterName)
        clusterOpsConf, version = self.tjApi.CheckoutCluster(projectName, clusterName)

        clusterConf = yaml.load(clusterOpsConf["cluster.conf"])
        if sourceMachine not in clusterConf["Machines"]:
            msg = "ReplaceMachine failed, source machine[%s] not exist in this cluster[%s]" % (
            sourceMachine, clusterName)
            ReturnError(404, msg)
        # clusterConf["Machines"].remove(sourceMachine)
        if destMachine not in clusterConf["Machines"]:
            clusterConf["Machines"].append(destMachine)

        newClusterOpsConf = {}
        newClusterOpsConf["cluster.conf"] = yaml.dump(clusterConf)

        tmpWorkDir = tempfile.mkdtemp()
        newClusterOpsConf["services"] = {}
        try:
            # clusterConfPath = os.path.join(tmpWorkDir, "cluster.conf")
            # SaveFile(clusterConfPath, clusterOpsConf["cluster.conf"])
            if clusterOpsConf.has_key("services"):
                for service in clusterOpsConf["services"]:
                    if clusterOpsConf["services"][service].has_key("role.conf"):
                        # serviceDir = os.path.join(tmpWorkDir, service)
                        # os.makedirs(serviceDir)
                        # roleConfPath =  os.path.join(serviceDir, "role.conf")
                        # SaveFile(roleConfPath, clusterOpsConf["services"][service]["role.conf"])
                        # ValidateMachineReplace(clusterConfPath, roleConfPath, sourceMachine, destMachine)
                        newRoleConfStr = DoMachineReplace(clusterOpsConf["services"][service]["role.conf"],
                                                          sourceMachine, destMachine)
                        newClusterOpsConf["services"][service] = {}
                        newClusterOpsConf["services"][service]["role.conf"] = newRoleConfStr
            commitMsg = "use machine[%s] to replace machine[%s] of cluster[%s]" % (
            destMachine, sourceMachine, clusterName)
            commitVersion = self.tjApi.SubmitCluster(projectName, clusterName, version, newClusterOpsConf, commitMsg)

        finally:
            shutil.rmtree(tmpWorkDir)

        returnValue = {}
        returnValue["code"] = 200
        returnValue["data"] = {}
        returnValue["data"]["Version"] = commitVersion
        print json.dumps(returnValue)

    def Admin(self, args):
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)

        returnValue = {}
        returnValue["code"] = 200
        returnValue["data"] = {}
        if args.list:
            ret = self.tjApi.ListAdmin()
            returnValue["data"] = ret
        if args.add:
            userList = args.add.split(",")
            for userName in userList:
                self.tjApi.AddAdmin(userName)
        if args.delete:
            userList = args.delete.split(",")
            for userName in userList:
                self.tjApi.RemoveAdmin(userName)

        print json.dumps(returnValue)

    def ListUser(self, args):
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        if args.project:
            ret = self.tjApi.ListProjectUser(args.project)
        elif args.service:
            ret = self.tjApi.ListServiceUser(args.service)

        returnValue = {}
        returnValue["code"] = 200
        returnValue["data"] = {}
        returnValue["data"]["OwnerList"] = ret["OwnerList"]
        returnValue["data"]["UserList"] = ret["UserList"]
        print json.dumps(returnValue)

    def AddUser(self, args):
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        userList = args.users.split(",")
        if args.project:
            for userName in userList:
                self.tjApi.AddProjectUser(userName, args.project, args.owner)
        elif args.service:
            for userName in userList:
                self.tjApi.AddServiceUser(userName, args.service, args.owner)

        returnValue = {}
        returnValue["code"] = 200
        returnValue["data"] = {}
        print json.dumps(returnValue)

    def RemoveUser(self, args):
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        userList = args.users.split(",")
        if args.project:
            for userName in userList:
                self.tjApi.RemoveProjectUser(userName, args.project)
        elif args.service:
            for userName in userList:
                self.tjApi.RemoveServiceUser(userName, args.service)

        returnValue = {}
        returnValue["code"] = 200
        returnValue["data"] = {}
        print json.dumps(returnValue)

    def TurnOn(self, args):
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        self.tjApi.TurnOnAPIServer()

        returnValue = {}
        returnValue["code"] = 200
        returnValue["data"] = {}
        print json.dumps(returnValue)

    def AuthMode(self, args):
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        ummJson = {}
        if args.get:
            ret = self.tjApi.GetMode()

        elif args.enable:
            if args.addr:
                ummJson['UAddress'] = args.addr
            else:
                ummJson['UAddress'] = ""
            if args.id:
                ummJson['UHello'] = args.id
            else:
                ummJson['UHello'] = ""
            if args.secret:
                ummJson['UWorld'] = args.secret
            else:
                ummJson['UWorld'] = ""

            ret = self.tjApi.EnableMode(ummJson)

        elif args.testmode:
            ummJson = {}
            ret = self.tjApi.EnableMode(ummJson, "test")

        returnValue = {}
        returnValue["code"] = 200
        returnValue["data"] = ret
        print json.dumps(returnValue)

    def BootStrap(self, args):
        cmd = "./run_bootstrap.sh %s" % args.file
        ExecCmd(cmd)

    def Fetch(self, args):
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        if args.project:
            project = args.project
        else:
            project = self.tjApi.projectName
        pattern = ".*"
        if args.pattern:
            pattern = args.pattern
        num = ""
        if args.num:
            num = args.num
        bytes = ""
        if args.bytes:
            bytes = args.bytes
        delimiter = ""
        if args.delimiter:
            delimiter = args.delimiter
        returnValue = {}
        returnValue["code"] = 200
        returnValue["data"] = {}
        if args.type == "filename":
            returnValue["message"] = self.tjApi.FetchFileList(project, args.cluster, args.machine, args.service,
                                                              args.serverrole, args.app, args.path, pattern)
        elif args.type == "alldir":
            returnValue["message"] = self.tjApi.FetchAllDir(project, args.cluster, args.machine, args.service,
                                                            args.serverrole, args.app)
        elif args.type == "log":
            returnValue["message"] = self.tjApi.FetchLog(project, args.cluster, args.machine, args.service,
                                                         args.serverrole, args.app, args.path, pattern, num, bytes,
                                                         delimiter)
        else:
            returnValue["code"] = 400
            returnValue["message"] = "only support fetch [filename|log]"

        print json.dumps(returnValue)

    def Docker(self, args):
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        if args.project:
            project = args.project
        else:
            project = self.tjApi.projectName
        if args.cluster:
            cluster = args.cluster
        else:
            cluster = "tianji"

        returnValue = {}
        returnValue["code"] = 200
        returnValue["data"] = {}
        if args.operation == "inspect":
            returnValue["message"] = self.tjApi.DockerInspect(project, cluster, args.service, args.serverrole, args.app,
                                                              args.machine)
        elif args.operation == "logs":
            returnValue["message"] = self.tjApi.DockerLogs(project, cluster, args.service, args.serverrole, args.app,
                                                           args.machine, args.num)
        else:
            returnValue["code"] = 400
            returnValue["message"] = "only support docker [inspect|logs]"

        print json.dumps(returnValue)

    def ClusterMonitor(self, args):
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        ret = self.tjApi.GetClusterSRMonitor(args.project, args.cluster, args.serverrole, args.monitor)
        returnValue = {}
        returnValue["code"] = 200
        returnValue["data"] = ret
        print json.dumps(returnValue, sort_keys=True, indent=2)

    def Machine(self, args):
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        returnValue = {}
        returnValue["code"] = 200
        returnValue["data"] = {}
        if args.all:
            ret = self.tjApi.ListClusterMachine(args.project, args.cluster)

            returnValue["data"] = ret
            print json.dumps(returnValue, sort_keys=True, indent=2)

    def Bucket(self, args):
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        returnValue = {}
        returnValue["code"] = 200
        returnValue["data"] = {}
        if args.all:
            ret = self.tjApi.ListMachineBucket(args.project)
        elif args.show:
            detail = "true"
            if args.detail:
                detail = args.detail
            ret = self.tjApi.GetMachineBucket(args.project, args.show, detail)
        elif args.create:
            ret = self.tjApi.CreateMachineBucket(args.project, args.create)
        elif args.delete:
            ret = self.tjApi.DeleteMachineBucket(args.project, args.delete)
        elif args.update:
            if not args.file:
                print "error: --file must be given for machines info"
                exit(1)
            with open(args.file) as f:
                machinesInfo = json.load(f)
            ret = self.tjApi.UpdateMachineBucket(args.project, args.update, machinesInfo)
        elif args.migrate:
            toProject = args.project
            toBucket = args.migrate
            if args.toproject:
                toProject = args.toproject
            if args.tobucket:
                toBucket = args.tobucket
            if not args.machines:
                print "error: --machines must be given for migrated machines"
                exit(1)
            machines = args.machines.split(",")
            migrateJson = {}
            migrateJson["Migrate"] = machines
            ret = self.tjApi.MigrateMachineBucket(args.project, args.migrate, toProject, toBucket, migrateJson)

        returnValue["data"] = ret
        print json.dumps(returnValue, sort_keys=True, indent=2)

    def Quota(self, args):
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        returnValue = {}
        returnValue["code"] = 200
        returnValue["data"] = {}
        if args.show:
            ret = self.tjApi.ListClusterQuota(args.project, args.show)
        elif args.update:
            if not args.file:
                print "error: --file must be given for quota info"
                exit(1)
            with open(args.file) as f:
                quotasInfo = json.load(f)
            ret = self.tjApi.UpdateClusterQuota(args.project, args.update, quotasInfo)

        returnValue["data"] = ret
        print json.dumps(returnValue, sort_keys=True, indent=2)

    def Usergroup(self, args):
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        returnValue = {}
        returnValue["code"] = 200
        returnValue["data"] = {}
        if args.all:
            ret = self.tjApi.ListUsergroups(args.project)
        elif args.create:
            if not args.owners and not args.users:
                print "error: at least one of --owners/--users should be given"
                exit(1)
            createJson = {}
            if args.owners:
                createJson["Owners"] = args.owners.split(",")
            if args.users:
                createJson["Users"] = args.users.split(",")
            ret = self.tjApi.CreateUsergroup(args.project, args.create, createJson)
        elif args.delete:
            ret = self.tjApi.DeleteUsergroup(args.project, args.delete)
        elif args.show:
            ret = self.tjApi.ShowUsergroup(args.project, args.show)
        elif args.update:
            if not args.owners and not args.users and not args.deleteusers:
                print "error: at least one of --owners/--users/--deleteusers should be given"
                exit(1)
            updateJson = {}
            if args.owners:
                updateJson["AddOwners"] = args.owners.split(",")
            if args.users:
                updateJson["AddUsers"] = args.users.split(",")
            if args.deleteusers:
                updateJson["Delete"] = args.deleteusers.split(",")
            ret = self.tjApi.ModifyUsergroup(args.project, args.update, updateJson)

        returnValue["data"] = ret
        print json.dumps(returnValue, sort_keys=True, indent=2)

    def DoAction(self, args):
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        if args.project:
            project = args.project
        else:
            project = self.tjApi.projectName
        cluster = args.cluster
        machine = args.machine
        action = args.action
        status = args.status

        returnValue = {}
        returnValue["code"] = 200
        returnValue["data"] = {}
        if args.byproject:
            service = args.service
            serverrole = args.serverrole
            if args.delete:
                returnValue["data"] = self.tjApi.DeleteSRMachineActionByProject(project, cluster, service, serverrole,
                                                                                machine, action, status)
            else:
                returnValue["data"] = self.tjApi.OperateSRMachinesByProject(project, cluster, service, serverrole,
                                                                            machine, action, status, args.dimpact,
                                                                            args.simpact, args.appimpact)
        elif args.byservice:
            service = args.service
            serverrole = args.serverrole
            if args.delete:
                returnValue["data"] = self.tjApi.DeleteSRMachineActionByService(cluster, service, serverrole, machine,
                                                                                action, status)
            else:
                returnValue["data"] = self.tjApi.OperateSRMachinesByService(cluster, service, serverrole, machine,
                                                                            action, status, args.dimpact, args.simpact,
                                                                            args.appimpact)
        elif args.bymachine:
            if args.delete:
                returnValue["data"] = self.tjApi.DeleteMachineAction(project, cluster, machine, action, status)
            else:
                if args.description:
                    description = args.description
                else:
                    description = ""
                returnValue["data"] = self.tjApi.SetMachineAction(project, cluster, machine, action, status,
                                                                  description, args.dimpact, args.simpact,
                                                                  args.appimpact)

        print json.dumps(returnValue, sort_keys=True, indent=2)

    def MachineSRInfo(self, args):
        configPath = args.config
        self.tjApi = self.initTjApi(configPath)
        if args.project:
            project = args.project
        else:
            project = self.tjApi.projectName
        cluster = args.cluster
        service = args.service
        serverrole = args.serverrole
        machine = args.machine

        returnValue = {}
        returnValue["code"] = 200
        returnValue["data"] = {}
        if args.byproject:
            returnValue["data"] = self.tjApi.GetMachineSRInfoByProject(project, cluster, service, serverrole, machine,
                                                                       args.monitor)
        else:
            returnValue["data"] = self.tjApi.GetMachineSRInfoByService(cluster, service, serverrole, machine,
                                                                       args.monitor)

        print json.dumps(returnValue, sort_keys=True, indent=2)


def ParseCheckout(subparsers):
    parser = subparsers.add_parser('checkout', help='checkout ops config by cluster to local dir')
    parser.add_argument('cluster', help='cluster name')
    parser.add_argument("-p", "--project", help="project name")
    parser.add_argument("--full", action="store_true", help="get full(extended) ops config")
    parser.add_argument("--plain", action="store_true", help="get plain ops config")
    parser.add_argument('target_dir', nargs='?', help='target dir to place ops config')
    parser.add_argument('--config', required=True, help="tianji_clt user info config file")


def ParseDiff(subparsers):
    parser = subparsers.add_parser('diff', help='show local diff')
    parser.add_argument('--versions', required=False, nargs=2, help='two versions to get the diff between them')
    parser.add_argument('--config', required=True, help="tianji_clt user info config file")
    parser.add_argument('--dir', required=True, help="cluster local directory")


def ParseLog(subparsers):
    parser = subparsers.add_parser('log', help='show logs')
    parser.add_argument('num', type=int, default=10, nargs='?', help='number of logs to show')
    parser.add_argument('--config', required=True, help="tianji_clt user info config file")
    parser.add_argument('--dir', required=True, help="cluster local directory")


def ParseShow(subparsers):
    parser = subparsers.add_parser('show', help='show this version ops config changes')
    parser.add_argument('version', help='submitted version of this cluster')
    parser.add_argument('--config', required=True, help="tianji_clt user info config file")
    parser.add_argument('--dir', required=True, help="cluster local directory")


def ParseUpdate(subparsers):
    parser = subparsers.add_parser('update', help='update local repo')
    parser.add_argument('--config', required=True, help="tianji_clt user info config file")
    parser.add_argument("-p", "--project", help="project name")
    parser.add_argument("--plain", action="store_true", help="get plain ops config")
    parser.add_argument('--dir', required=True, help="cluster local directory")


def ParsePush(subparsers):
    parser = subparsers.add_parser('push', help='push local change to config store')
    parser.add_argument('-m', '--comment', required=True, help="commit descrpiton")
    parser.add_argument("-p", "--project", help="project name")
    parser.add_argument('--updateos', action="store_true", help="confirm to update os service")
    parser.add_argument('--config', required=True, help="tianji_clt user info config file")
    parser.add_argument('--dir', required=True, help="cluster local directory")
    parser.add_argument('--force_rolling', action="store_true", help="do rolling even if there is no local diff")


def ParseBranch(subparsers):
    parser = subparsers.add_parser('branch',
                                   help='branch related operation, {-a: list all clusters, -d: delete cluster -c: create cluster}')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-a', '--all', action="store_true", help="list all clusters")
    group.add_argument("-d", "--delete", help="delete cluster")
    group.add_argument("-c", "--create", help="create cluster")
    group.add_argument("--allocate", help="allocate machines")
    group.add_argument("--listug", help="list cluster usergroup")
    group.add_argument("--addug", help="add cluster usergroup")
    group.add_argument("--deleteug", help="delete cluster usergroup")
    group.add_argument("-r", "--reference", help="set reference service template")
    parser.add_argument("-p", "--project", help="project name")
    parser.add_argument("-s", "--service_name", help="service name")
    parser.add_argument("-t", "--template_name", help="service template name")
    parser.add_argument("-v", "--version", help="template version")
    parser.add_argument("--remove", action="store_true")
    parser.add_argument("-f", "--file", help="allocate requirement/quota info")
    parser.add_argument("--ug", help="user group name")
    parser.add_argument("--autogen", action="store_true", help="auto gen enc key")
    parser.add_argument('--config', required=True, help="tianji_clt user info config file")


def ParseCluster(subparsers):
    parser = subparsers.add_parser('cluster',
                                   help='cluster related operation, {-a: list all clusters, -c: create cluster, -d: delete cluster}')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-a', '--all', action="store_true", help="list all clusters")
    group.add_argument("-d", "--delete", help="delete cluster")
    group.add_argument("-c", "--create", help="create cluster")
    group.add_argument("--allocate", help="allocate machines")
    group.add_argument("--listug", help="list cluster usergroup")
    group.add_argument("--addug", help="add cluster usergroup")
    group.add_argument("--deleteug", help="delete cluster usergroup")
    group.add_argument("-r", "--reference", help="set reference service template")
    parser.add_argument("-p", "--project", help="project name")
    parser.add_argument("-s", "--service_name", help="service name")
    parser.add_argument("-t", "--template_name", help="service template name")
    parser.add_argument("-v", "--version", help="template version")
    parser.add_argument("--remove", action="store_true")
    parser.add_argument("-f", "--file", help="allocate requirement/quota info")
    parser.add_argument("--ug", help="user group name")
    parser.add_argument("--autogen", action="store_true", help="auto gen enc key")
    parser.add_argument('--config', required=True, help="tianji_clt user info config file")


def ParseInstance(subparsers):
    parser = subparsers.add_parser('instance',
                                   help='serviceinstance related operation, {-l: list all serviceinstances, -a: add serviceinstance, -d: delete serviceinstance}')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-l', '--all', action="store_true", help="list all serviceinstances")
    group.add_argument("-a", "--add", help="add serviceinstance")
    group.add_argument("-d", "--delete", help="delete serviceinstance")
    parser.add_argument("-c", "--cluster", metavar='cluster', required=True, help="cluster name")
    parser.add_argument("-p", "--project", help="project name")
    parser.add_argument('--config', required=True, help="tianji_clt user info config file")


def ParseProject(subparsers):
    parser = subparsers.add_parser('project',
                                   help='project related operation, {-a: list all projects; -c: create project}')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-a', '--all', action="store_true", help="list all projects")
    group.add_argument("-c", "--create", help="create project")
    group.add_argument("-d", "--delete", help="delete project")
    parser.add_argument('--config', required=True, help="tianji_clt user info config file")


def ParseService(subparsers):
    parser = subparsers.add_parser('service',
                                   help='service related operation, {-a: list all services; -c: create service}')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-a', '--all', action="store_true", help="list all services")
    group.add_argument("-c", "--create", help="create service")
    parser.add_argument('--sys', action="store_true", help="whether to be sys service")
    parser.add_argument('--config', required=True, help="tianji_clt user info config file")


def ParseServiceTemplate(subparsers):
    parser = subparsers.add_parser('service_template', help='service_tmplate related operation')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-a", "--all", action="store_true",
                       help="--service_name --config, list all templates of this service")
    group.add_argument("-c", "--create", action="store_true", help="--service_name --template_name --dir --config -m")
    group.add_argument("-u", "--update", action="store_true", help="--service_name --template_name --dir --config -m")
    group.add_argument("-d", "--delete", action="store_true", help="--service_name --template_name --config")
    group.add_argument("-g", "--get", action="store_true", help="--service_name --template_name --dir --config")
    group.add_argument("-l", "--log", action="store_true", help="--service_name --template_name --config")
    group.add_argument("-r", "--reference", action="store_true",
                       help="--cluster_name --service_name --template_name --config")
    group.add_argument("--show_reference", action="store_true",
                       help="--service_name --template_name --config, get reference cluster for service template")
    parser.add_argument("-s", "--service_name")
    parser.add_argument("--cluster_name")
    parser.add_argument("-t", "--template_name")
    parser.add_argument("-v", "--version", help="template version")
    parser.add_argument("--remove", action="store_true")
    parser.add_argument("--plain", action="store_true", help="get plain ops config")
    parser.add_argument("--dir", help="tianji service tmplate local saved dir")
    parser.add_argument('num', type=int, default=10, nargs='?', help='number of logs to show')
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")
    parser.add_argument("-m", "--comment", help="commit descrpiton")
    parser.add_argument("--group", help="template group")
    parser.add_argument("--desc", help="descrpiton")


def ParseServiceSchema(subparsers):
    parser = subparsers.add_parser('service_schema', help='service_schema related operation')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-a", "--all", action="store_true",
                       help="--service_name --config, list all schemas of this service")
    group.add_argument("-c", "--create", action="store_true", help="--service_name --schema_name -f --config -m")
    group.add_argument("-u", "--update", action="store_true", help="--service_name --schema_name -f --config -m")
    group.add_argument("-d", "--delete", action="store_true", help="--service_name --schema_name --config")
    group.add_argument("-g", "--get", action="store_true", help="--service_name --schema_name -o --config")
    group.add_argument("-l", "--log", action="store_true", help="--service_name --schema_name --config")
    parser.add_argument("-s", "--service_name")
    parser.add_argument("-sh", "--schema_name")
    parser.add_argument("-v", "--version", help="schema version")
    parser.add_argument('num', type=int, default=10, nargs='?', help='number of logs to show')
    parser.add_argument("-f", "--file", help="tianji service schema local saved file")
    parser.add_argument("-o", "--output", help="tianji service schema local saved file")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")
    parser.add_argument("-m", "--comment", help="commit descrpiton")
    parser.add_argument("--desc", help="descrpiton")


def ParseServiceValidator(subparsers):
    parser = subparsers.add_parser('service_validator', help='service_validator related operation')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-a", "--all", action="store_true",
                       help="--service_name --config, list all validators of this service")
    group.add_argument("-c", "--create", action="store_true", help="--service_name --validator_name -f --config -m")
    group.add_argument("-u", "--update", action="store_true", help="--service_name --validator_name -f --config -m")
    group.add_argument("-d", "--delete", action="store_true", help="--service_name --validator_name --config")
    group.add_argument("-g", "--get", action="store_true", help="--service_name --validator_name -o --config")
    group.add_argument("-l", "--log", action="store_true", help="--service_name --validator_name --config")
    parser.add_argument("-s", "--service_name")
    parser.add_argument("-sh", "--validator_name")
    parser.add_argument("-v", "--version", help="validator version")
    parser.add_argument('num', type=int, default=10, nargs='?', help='number of logs to show')
    parser.add_argument("-f", "--file", help="tianji service validator local saved file")
    parser.add_argument("-o", "--output", help="tianji service validator local saved file")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")
    parser.add_argument("-m", "--comment", help="commit descrpiton")
    parser.add_argument("--desc", help="descrpiton")


def ParseCheck(subparsers):
    parser = subparsers.add_parser('check', help='check rolling/serverrole status')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-a", "--all", metavar='version', help="check overall status, use 'head' for latest rolling")
    group.add_argument("-r", "--rolling", metavar='version', help="check rolling, use 'head' for latest rolling")
    group.add_argument("-srlist", "--allsr", action="store_true", help="list all serverroles")
    group.add_argument("-sr", "--serverrole", nargs=2, metavar=('hostname', 'serverrole'),
                       help="get serverrole needs 2 params [hostname, serverrole]")
    parser.add_argument("-c", "--cluster", metavar='cluster', required=True, help="cluster name")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")


def ParseMachineInfo(subparsers):
    parser = subparsers.add_parser('machineinfo', help='get machine info')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-m", "--machines", help="hostnames, divided by ','")
    group.add_argument("-f", "--file", help="each hostname in one line")
    parser.add_argument("-a", "--attr", metavar='attributes',
                        help="attributes to return, default is 'cluster,ip,state,actionname,actionstatus'")
    parser.add_argument('--batch', action="store_true", help="batch get machineinfo")
    parser.add_argument('--local', action="store_true", help="get local machineinfo")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")


def ParseRollings(subparsers):
    parser = subparsers.add_parser('rollings', help='operation tianji rolling jobs')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-l", "--list", action="store_true", help="list all rolling jobs")
    group.add_argument("-s", "--show", metavar='show job of version',
                       help="show rolling job detail, use 'head' for head version")
    group.add_argument("--terminate", metavar='termitate job of version',
                       help="terminate rolling job, use 'head' for head version")
    group.add_argument("--pause", metavar='pause job of version', help="pause rolling job, use 'head' for head version")
    group.add_argument("--resume", metavar='resume job of version',
                       help="resume rolling job, use 'head' for head version")
    group.add_argument("--retry", metavar='retry job of version', help="retry rolling job, use 'head' for head version")
    parser.add_argument("-c", "--cluster", metavar='cluster', required=True, help="cluster name")
    parser.add_argument("-sv", "--subversion", metavar='sub version', help="sub version")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")


def ParseRollingStatus(subparsers):
    parser = subparsers.add_parser('rollingstatus', help='current deploy status')
    parser.add_argument("-c", "--cluster", metavar='cluster', required=True, help="cluster name")
    parser.add_argument("-v", "--version", metavar='version', help="get serverrole status")
    parser.add_argument("-j", "--json", action="store_true", help="output in json format")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")


def ParseServiceStatus(subparsers):
    parser = subparsers.add_parser('servicestatus', help='current service status')
    parser.add_argument("-c", "--cluster", metavar='cluster', required=True, help="cluster name")
    parser.add_argument("-s", "--services", help="push services, divided by ','")
    parser.add_argument("-j", "--json", action="store_true", help="output in json format")
    parser.add_argument("-v", "--version", metavar='version', help="get serverrole status")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")


def ParseMachineStatus(subparsers):
    parser = subparsers.add_parser('machinestatus', help='current machine status')
    parser.add_argument("-c", "--cluster", metavar='cluster', required=True, help="cluster name")
    parser.add_argument("-j", "--json", action="store_true", help="output in json format")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")


def ParseClearLocal(subparsers):
    parser = subparsers.add_parser('clearlocal', help='remove all tianji services and data on localhost')
    parser.add_argument("-b", "--bootconfig", help="tianji seed bootstrap config file")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")


def ParseCreateDockerCluster(subparsers):
    parser = subparsers.add_parser('createdockercluster', help='init cluster.conf, sys.conf and tianji client')
    parser.add_argument("-c", "--cluster", required=True, help="cluster name")
    parser.add_argument("-m", "--machines", required=True, help="cluster machines, divided by ','")
    parser.add_argument("-e", "--env", required=True, help="docker starting environment")
    parser.add_argument("-b", "--build", required=True, help="buildid for tjclient")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")


def ParseDeployServices(subparsers):
    parser = subparsers.add_parser('deployservices', help='push selected services')
    parser.add_argument("-c", "--cluster", metavar='cluster', required=True, help="cluster name")
    parser.add_argument("-s", "--services", required=True, help="push services, divided by ','")
    parser.add_argument("-p", "--project", help="project name")
    parser.add_argument('--dir', required=True, help="docker services directory")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")
    parser.add_argument('--force', action="store_true", help="do rolling no matter whethe has actual diff")


def ParseReplaceMachine(subparsers):
    parser = subparsers.add_parser('replacemachine', help='replace machine A with B')
    parser.add_argument("-c", "--cluster", metavar='cluster', required=True, help="cluster name")
    parser.add_argument("-s", "--source", required=True, help="source machine A")
    parser.add_argument("-d", "--dest", required=True, help="dest machine B")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")


def ParseUpgradeServerRoles(subparsers):
    parser = subparsers.add_parser('upgradeserverroles', help='push selected serverroles')
    parser.add_argument("-c", "--cluster", metavar='cluster', required=True, help="cluster name")
    parser.add_argument("-s", "--services", required=True, help="push services, divided by ','")
    parser.add_argument("-sr", "--serverroles", required=True, help="push serverroles, divided by ','")
    parser.add_argument("-p", "--project", help="project name")
    parser.add_argument('--dir', required=True, help="docker services directory")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")


def ParseAdmin(subparsers):
    parser = subparsers.add_parser('admin', help='add/remove/list admins')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-l", "--list", action="store_true", help="list admins")
    group.add_argument("-a", "--add", help="add admin users, divided by ','")
    group.add_argument("-d", "--delete", help="delete admin users, divided by ','")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")


def ParseListUser(subparsers):
    parser = subparsers.add_parser('listuser', help='list users for project/service')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-p", "--project", help="project name")
    group.add_argument("-s", "--service", help="service name")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")


def ParseAddUser(subparsers):
    parser = subparsers.add_parser('adduser', help='add users for project/service')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-p", "--project", help="project name")
    group.add_argument("-s", "--service", help="service name")
    parser.add_argument("-u", "--users", required=True, help="users to add, divided by ','")
    parser.add_argument('--owner', action="store_true", help="whether to be owner")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")


def ParseRemoveUser(subparsers):
    parser = subparsers.add_parser('removeuser', help='remove users for project/service')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-p", "--project", help="project name")
    group.add_argument("-s", "--service", help="service name")
    parser.add_argument("-u", "--users", required=True, help="users to add, divided by ','")
    parser.add_argument('--owner', action="store_true", help="whether to be owner")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")


def ParseTurnOn(subparsers):
    parser = subparsers.add_parser('turnon', help='turn on apiserver')
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")


def ParseAuthMode(subparsers):
    parser = subparsers.add_parser('authmode', help='operate authorization for apiserver')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--get", action="store_true", help="get mode")
    group.add_argument("--enable", action="store_true", help="set mode enable")
    group.add_argument("--testmode", action="store_true", help="set mode test")
    parser.add_argument("--addr", help="umm address")
    parser.add_argument("--id", help="umm id")
    parser.add_argument("--secret", help="umm secret")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")


def ParseBootStrap(subparsers):
    parser = subparsers.add_parser("bootstrap", help="bootstrap tianji")
    parser.add_argument("-f", "--file", help="bootstrap config file")


def ParseFetch(subparsers):
    parser = subparsers.add_parser("fetch", help="fetch log from host")
    parser.add_argument('type', help='support log/name types')
    parser.add_argument("-p", "--project", help="project name")
    parser.add_argument("-c", "--cluster", required=True, help="cluster name")
    parser.add_argument("-m", "--machine", required=True, help="log fetcher machine")
    parser.add_argument("-s", "--service", required=True, help="service name")
    parser.add_argument("-sr", "--serverrole", required=True, help="serverrole name")
    parser.add_argument("-a", "--app", required=True, help="app name")
    parser.add_argument("--path", required=True, help="log file path")
    parser.add_argument("--pattern", help="log file name pattern")
    parser.add_argument("-n", "--num", help="fetch num")
    parser.add_argument("-b", "--bytes", help="fetch bytes")
    parser.add_argument("-d", "--delimiter", help="log content delimiter")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")


def ParseDocker(subparsers):
    parser = subparsers.add_parser("docker", help="docker log from host")
    parser.add_argument('operation', help='support inspect/logs operations')
    parser.add_argument("-p", "--project", help="project name")
    parser.add_argument("-c", "--cluster", help="cluster name")
    parser.add_argument("-m", "--machine", required=True, help="log fetcher machine")
    parser.add_argument("-s", "--service", required=True, help="service name")
    parser.add_argument("-sr", "--serverrole", required=True, help="serverrole name")
    parser.add_argument("-a", "--app", required=True, help="app name")
    parser.add_argument('-n', "--num", type=int, default=50, help='number of logs to show')
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")


def ParseClusterMonitor(subparsers):
    parser = subparsers.add_parser("clustermonitor", help="get cluster monitor")
    parser.add_argument("-p", "--project", required=True, help="project name")
    parser.add_argument("-c", "--cluster", required=True, help="cluster name")
    parser.add_argument("-sr", "--serverrole", required=True, help="serverrole name")
    parser.add_argument("-m", "--monitor", required=True, help="monitor name")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")


def ParseMachine(subparsers):
    parser = subparsers.add_parser("machine", help="machine operations")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-a', '--all', action="store_true", help="list all machines")
    parser.add_argument("-p", "--project", required=True, help="project name")
    parser.add_argument("-c", "--cluster", required=True, help="cluster name")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")


def ParseBucket(subparsers):
    parser = subparsers.add_parser("bucket", help="machine bucket operations")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-a", "--all", action="store_true", help="list all machine buckets")
    group.add_argument("-s", "--show", help="get machine bucket")
    group.add_argument("-c", "--create", help="create machine bucket")
    group.add_argument("-d", "--delete", help="delete machine bucket")
    group.add_argument("-u", "--update", help="update machine bucket")
    group.add_argument("-m", "--migrate", help="migrate machine bucket")
    parser.add_argument("-p", "--project", required=True, help="project name")
    parser.add_argument("-f", "--file", help="update machines info")
    parser.add_argument("--toproject", help="migrate to project")
    parser.add_argument("--tobucket", help="migrate to machine bucket")
    parser.add_argument("--machines", help="migrate machines")
    parser.add_argument("--detail", help="migrate machines")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")


def ParseQuota(subparsers):
    parser = subparsers.add_parser("quota", help="quota operations")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-s', '--show', help="show cluster quotas")
    group.add_argument('-u', '--update', help="update cluster quotas")
    parser.add_argument("-p", "--project", required=True, help="project name")
    parser.add_argument("-f", "--file", help="update quota info")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")


def ParseUsergroup(subparsers):
    parser = subparsers.add_parser("usergroup", help="usergroup operations")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-a', '--all', action="store_true", help="list user groups")
    group.add_argument('-c', '--create', help="create user group")
    group.add_argument('-d', '--delete', help="delete user group")
    group.add_argument('-s', '--show', help="get user group")
    group.add_argument('-u', '--update', help="update user group")
    parser.add_argument("-p", "--project", required=True, help="project name")
    parser.add_argument('--owners', help="owners to be added")
    parser.add_argument('--users', help="users to be added")
    parser.add_argument('--deleteusers', help="users to be deleted")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")


def ParseDoAction(subparsers):
    parser = subparsers.add_parser("doaction", help="take action on serverrole")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--byproject', action="store_true", help="call from project")
    group.add_argument('--byservice', action="store_true", help="call from service")
    group.add_argument('--bymachine', action="store_true", help="call from service")
    parser.add_argument("-p", "--project", help="project name")
    parser.add_argument("-c", "--cluster", required=True, help="cluster name")
    parser.add_argument("-s", "--service", help="service name")
    parser.add_argument("-sr", "--serverrole", help="serverrole name")
    parser.add_argument("--action", required=True, help="do action")
    parser.add_argument("--status", default="", help="action status")
    parser.add_argument("--delete", action="store_true", help="delete action")
    parser.add_argument("-m", "--machine", required=True, help="do action on machine, use 'Machines' for all machines")
    parser.add_argument("-d", "--description", help="action description")
    parser.add_argument("--dimpact", default="", help="action dimpact")
    parser.add_argument("--simpact", default="", help="action simpact")
    parser.add_argument("--appimpact", default="", help="action appimpact")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")


def ParseMachineSRInfo(subparsers):
    parser = subparsers.add_parser("machinesrinfo", help="get machine serverrole info")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--byproject', action="store_true", help="call from project")
    group.add_argument('--byservice', action="store_true", help="call from service")
    parser.add_argument("-p", "--project", help="project name")
    parser.add_argument("-c", "--cluster", required=True, help="cluster name")
    parser.add_argument("-s", "--service", required=True, help="service name")
    parser.add_argument("-sr", "--serverrole", required=True, help="serverrole name")
    parser.add_argument("-m", "--machine", required=True, help="do action on machine, use 'Machines' for all machines")
    parser.add_argument('--monitor', action="store_true", help="get machine monitor info")
    parser.add_argument("--config", required=True, help="tianji_clt user info config file")


def ParseArgs(argList):
    allparser = argparse.ArgumentParser()
    subparsers = allparser.add_subparsers()
    ParseCheckout(subparsers)
    ParseDiff(subparsers)
    ParseLog(subparsers)
    ParseShow(subparsers)
    ParseUpdate(subparsers)
    ParsePush(subparsers)
    ParseBranch(subparsers)
    ParseCluster(subparsers)
    ParseInstance(subparsers)
    ParseProject(subparsers)
    ParseService(subparsers)
    ParseServiceTemplate(subparsers)
    ParseServiceSchema(subparsers)
    ParseServiceValidator(subparsers)
    ParseCheck(subparsers)
    ParseMachineInfo(subparsers)
    ParseRollings(subparsers)
    ParseRollingStatus(subparsers)
    ParseServiceStatus(subparsers)
    ParseMachineStatus(subparsers)
    ParseClearLocal(subparsers)
    ParseCreateDockerCluster(subparsers)
    ParseDeployServices(subparsers)
    ParseReplaceMachine(subparsers)
    ParseUpgradeServerRoles(subparsers)
    ParseAdmin(subparsers)
    ParseListUser(subparsers)
    ParseAddUser(subparsers)
    ParseRemoveUser(subparsers)
    ParseTurnOn(subparsers)
    ParseAuthMode(subparsers)
    ParseBootStrap(subparsers)
    ParseFetch(subparsers)
    ParseDocker(subparsers)
    ParseClusterMonitor(subparsers)
    ParseMachine(subparsers)
    ParseBucket(subparsers)
    ParseQuota(subparsers)
    ParseUsergroup(subparsers)
    ParseDoAction(subparsers)
    ParseMachineSRInfo(subparsers)
    return allparser.parse_args(argList)


def main(args):
    tjGit = TianjiGit()
    ar = ParseArgs(args)
    if args[0] == "checkout":
        tjGit.Checkout(ar)
    elif args[0] == "update":
        tjGit.Update(ar)
    elif args[0] == "push":
        tjGit.Push(ar)
    elif args[0] == "log":
        tjGit.Log(ar)
    elif args[0] == "show":
        tjGit.Show(ar)
    elif args[0] == "diff":
        tjGit.Diff(ar)
    elif args[0] == "branch":
        tjGit.Branch(ar)
    elif args[0] == "cluster":
        tjGit.Branch(ar)
    elif args[0] == "instance":
        tjGit.Instance(ar)
    elif args[0] == "project":
        tjGit.Project(ar)
    elif args[0] == "service":
        tjGit.Service(ar)
    elif args[0] == "service_template":
        tjGit.ServiceTemplate(ar)
    elif args[0] == "service_schema":
        tjGit.ServiceSchema(ar)
    elif args[0] == "service_validator":
        tjGit.ServiceValidator(ar)
    elif args[0] == "check":
        tjGit.Check(ar)
    elif args[0] == "machineinfo":
        tjGit.MachineInfo(ar)
    elif args[0] == "rollings":
        tjGit.Rollings(ar)
    elif args[0] == "rollingstatus":
        tjGit.RollingStatus(ar)
    elif args[0] == "servicestatus":
        tjGit.ServiceStatus(ar)
    elif args[0] == "machinestatus":
        tjGit.MachineStatus(ar)
    elif args[0] == "clearlocal":
        tjGit.ClearLocal(ar)
    elif args[0] == "createdockercluster":
        tjGit.CreateDockerCluster(ar)
    elif args[0] == "deployservices":
        tjGit.DeployServices(ar)
    elif args[0] == "replacemachine":
        tjGit.ReplaceMachine(ar)
    elif args[0] == "upgradeserverroles":
        tjGit.UpgradeServerRoles(ar)
    elif args[0] == "admin":
        tjGit.Admin(ar)
    elif args[0] == "listuser":
        tjGit.ListUser(ar)
    elif args[0] == "adduser":
        tjGit.AddUser(ar)
    elif args[0] == "removeuser":
        tjGit.RemoveUser(ar)
    elif args[0] == "turnon":
        tjGit.TurnOn(ar)
    elif args[0] == "authmode":
        tjGit.AuthMode(ar)
    elif args[0] == "bootstrap":
        tjGit.BootStrap(ar)
    elif args[0] == "fetch":
        tjGit.Fetch(ar)
    elif args[0] == "docker":
        tjGit.Docker(ar)
    elif args[0] == "clustermonitor":
        tjGit.ClusterMonitor(ar)
    elif args[0] == "machine":
        tjGit.Machine(ar)
    elif args[0] == "bucket":
        tjGit.Bucket(ar)
    elif args[0] == "quota":
        tjGit.Quota(ar)
    elif args[0] == "usergroup":
        tjGit.Usergroup(ar)
    elif args[0] == "doaction":
        tjGit.DoAction(ar)
    elif args[0] == "machinesrinfo":
        tjGit.MachineSRInfo(ar)


if __name__ == "__main__":

    #    tjSDK = TianjiSDK()
    #    tjSDK.initTjSDK(["http://rt1b07021.tbc:8085"], "tianji", "", "", "/tmp/tianij_conf")
    #    print tjSDK.DeleteClusterAndUsergroup("tianji", "zytest1")
    #    machines = ["rs3b07029.et2sqa","rs3b07011.et2sqa"]
    #    attr = "attr=state"
    #    print tjSDK.BatchGetMachineInfo(machines, attr)
    #    exit(1)
    #    print tjSDK.DeleteClusterAndUsergroup("tianji", "zytest1")
    #    tjSDK.initTjSDK(["http://rt1b07021.tbc:8085"], "tianji", "", "", "/tmp/tianij_conf")
    #    serverrole_ip_map = {}
    #    serverrole_ip_map["test1#"] = {}
    #    serverrole_ip_map["test1#"]["machines"] = ["10.97.252.47"]
    #    serverrole_ip_map["test1#"]["version"] = "tianteng#123456"
    #    serverrole_ip_map["test2#"] = {}
    #    serverrole_ip_map["test2#"]["machines"] = ["10.97.252.47"]
    #    serverrole_ip_map["test2#"]["version"] = "tianteng#123456"
    #    print tjSDK.DeployServerRole("zytest", "ops", serverrole_ip_map)
    #    exit(1)
    #    print tjSDK.ListProjects()
    #    print tjSDK.ListServices()
    #    print tjSDK.ListBucketMachineNumber("project_e2e")
    #    exit(1)
    #    result = tjSDK.UpdateMachine("project_e2e", "zytest1", {'Add': ["abc"], 'Delete': []})
    #    print result
    #    cluster_quota = {"default":{"MaxMachineCount":100}}
    #    user_group = {"Owners":["owner1", "owner2"], "Users":["owner1", "user2"]}
    #    machine_spec = {"default":[{"requires":{"hostname":".*"},"count":1}]}
    #
    #    #delete cluster
    #    exit(1)
    #    time.sleep(30)
    #    print tjSDK.CheckDeployStatus("project_e2e", "zytest")
    #
    #    #create and deploy cluster
    #    deployConf = {}
    #    deployConf["aaa"] = "abc"
    #    ret = tjSDK.DeployCluster("tianji", "zytest1", cluster_quota, user_group, machine_spec, deployConf)
    #    if ret["code"] != 200:
    #        print ret
    #        exit(1)
    #    print ret
    #    version = ret["data"]["version"]
    #
    #    #check submit status and get cluster version
    #    print tjSDK.CheckSubmitStatus("project_e2e", "zytest")
    #    print ret
    #    if ret["code"] != 200:
    #        print ret
    #        exit(1)
    #    version = ret["data"]["version"]
    #
    #    #check deploy status
    #    print tjSDK.CheckDeployStatus("project_e2e", "zytest", version)
    #
    #    print tjSDK.GetQuota("project_e2e", "zytest")
    #    print tjSDK.GetUsergroup("project_e2e", "zytest_usergroup")
    #    print tjSDK.ListClusterMachines("project_e2e", "zytest")
    #    exit(1)
    #    print tjSDK.ListMachineBuckets("project_e2e")
    #    print tjSDK.ListBucketMachines("project_e2e", "default")
    #    usergroup_json = {"Delete":["owner2"]}
    #    print tjSDK.UpdateUsergroup("project_e2e", "zytest_usergroup", usergroup_json)
    #    print tjSDK.GetUsergroup("project_e2e", "zytest_usergroup")
    #    exit(1)

    try:
        main(sys.argv[1:])
    except ReturnErrorException, e:
        print str(e)
        sys.exit(1)
    except Exception, e:
        traceback.print_exc()
        raise Exception, "tianji_clt failed, exception: %s" % (e)
        sys.exit(1)
